#include <FileTypesAndCreators.h>

#ifdef ftWindowsWaveFile
# error
#else
# define ftWindowsWaveFile					FOUR_CHAR_CODE('.WAV')
#endif

#define ftRawData							FOUR_CHAR_CODE('TEXT')
#define ftHyperXDBCustom					FOUR_CHAR_CODE('tU16')

#include "macintosh_cseries.h"

#define MAXIMUM_OPEN_DOCUMENTS_COUNT		32

#include "myDocument.h"

#include "shell.h"
#include "interface.h"
#include "screen.h"
#include "myFiles.h"

#include <string.h>

/* -------- constants */

#define SLOT_USED	0x8000

const OSType		docTypes[] = {ftRawData, ftHyperXDBCustom, ftSoundFile, AIFFID, AIFCID};

/* -------- globals */

StandardFileReply	selectedFile;

int					frontDoc = NONE;		/* nothing open initially */

documentListRec		openDocuments = {0,0,{0,0,"\puntitled"}};

/* -------- prototypes */

WindowRef NewDocumentWindow(documentRec*);
void CloseDocumentWindow(documentRec*);

void reposition_window_coordinates(void);

/* -------- private prototypes */

OSErr WriteDocument(documentRec*, FSSpecPtr customFSSpec);
OSErr LoadDocument(documentRec *doc);

/* -------- code */

// if we cant open anymore of em
boolean noMoreDocs()
{
	return (openDocuments.count>=MAXIMUM_OPEN_DOCUMENTS_COUNT);
}

documentRec *NewDoc(documentRec *doc, boolean new_window)
{
	WindowRef		window;

	if (doc==NULL)
		doc = openDocuments.index+openDocuments.count;
	if (noMoreDocs())
		goto no;
	assert(!(doc->flags & SLOT_USED));
	memset(doc,0,sizeof(*doc));
	doc->flags |= SLOT_USED;
	if (new_window)
	{
		assert(doc->window==NULL);
		if ((window=NewDocumentWindow(doc))==NULL)
			no: return NULL;
		doc->window = window;
	}
	++openDocuments.count;
	// whatever
	return doc;
}

/*
// returns true if a file has been opened
//	(should only apply to when get is true)
boolean OpenDoc(boolean get)
{
	documentRec		*doc;
	OSErr			err;

	// only if we have no more file slots...
	if (noMoreDocs() && !askSave(frontDoc))
		return;
	if (get)
	{
		if (QuickTime_version>=0x2508000)
			StandardGetFilePreview(NULL,sizeof(docTypes)>>2,(OSType*)&docTypes,&selectedFile);
		else
			StandardGetFile(NULL,sizeof(docTypes)>>2,(OSType*)&docTypes,&selectedFile);
		if (!selectedFile.sfGood)
			goto bad;
	}
	if ((doc=NewDoc(NULL,TRUE))==NULL)
	{
		alert_user(infoError,strERRORS,couldNotCreateNewDoc,QDError());
		goto bad;
	}
	doc->file_owner = *(FSSpec*)&selectedFile.sfFile;
	if (err=OpenDocumentFile(&doc))
	{
		if (err!=opWrErr)// || doc==NULL)
			goto error;
		FullSelectWindow(doc->window,true);
		goto end;
	}
	if (err=LoadDocument(doc))
	{
	error:
		alert_user(infoError,strERRORS,fileSystemError,err);
	bad:
		return 0;
	}
end:
	return 1;
}
*/

boolean WCloseDoc(WindowRef window)
{
	int				index;

	if ((index=WFindDoc(window))==NONE)
		return FALSE;
	CloseDoc(index);
}

int WFindDoc(WindowRef window)
{
	documentRec		*doc = openDocuments.index;
	int				count, x;

	if ((x=openDocuments.count)<=0)
		goto end;
	do
	{
		if (doc->window==window)
			return openDocuments.count-x;
		++doc;
	} while(--x);
end:
	return NONE;
}

int RFindDoc(SInt16 ref)
{
	documentRec		*doc = openDocuments.index;
	int				count, x;

	if ((x=openDocuments.count)<=0)
		goto end;
	do
	{
		if (doc->file_ref==ref) return openDocuments.count-x;
		++doc;
	} while(--x);
end:
	return NONE;
}

// we are being called to try and close one or all of our documents
boolean CloseDoc(int which)
{
	documentRec		*doc;
	int				x = 1;

	if (!openDocuments.count)
		goto end;
	if (which==kAllDocuments)
		which = (x=openDocuments.count)-1;	// start from last
	else if (which==kFrontDocument)
		which = frontDoc;

	doc = openDocuments.index+which;

	do
	{
		assert(doc->flags&SLOT_USED);
		if (askSave(which))
		{
			CloseDocumentWindow(doc);
			CloseDocumentFile(doc);
			doc->flags &= ~SLOT_USED;
			//QLRemoveEntryW(openDocuments,which);
		} else return FALSE;
		--doc, --which;
	} while(--x);
end:
	return TRUE;
}

// returns true if we can dispose of the document in memory
boolean askSave(int which)
{
	documentRec		*doc;

	if (which<0)
		return 0;
	/* only if somethings open, and its been modified */

	doc = openDocuments.index+which;
	assert(doc->flags & SLOT_USED);

	if (doc->flags & SLOT_DIRTY)
	{
		SysBeep(20);
		ParamText(doc->file_owner.name,empty,empty,empty);
		switch(Alert(129,NULL))
		{
		case 3:		/* Cancel */
			return 0;
		case 1:		/* Yes */
			SaveDoc(which,false);
		case 2:		/* No */
			;		/* we will fall into return 1 */
		}
	}
	return 1;
}

boolean SaveDoc(int which, boolean dialog)
{
	documentRec		*doc;
	int				x = sizeof(documentRec);
	OSErr			err = 0;

	if (which==kAllDocuments)
		which = (x=openDocuments.count)-1;
	else
	{
		x = 1;
		if (which==kFrontDocument)
			which = frontDoc;
	}
	doc = openDocuments.index+which;
	do
	{
		if (dialog)
		{
			StandardPutFile(NULL,doc->file_owner.name,&selectedFile);
			if (!selectedFile.sfGood)
				return 0;
			if (err=WriteDocument(doc,(FSSpec*)&selectedFile.sfFile))
				alert_user(infoError,strERRORS,fileSystemError,err);
		} else WriteDocument(doc,NULL);
		--doc;
	} while(--x);
}

/* -------- windows */

short		current_x = 12, current_y = 26, accum_space = 0;

WindowRef NewDocumentWindow(documentRec *doc)
{
	Rect			arect;

	ControlRef		ctl;
	Ptr				wins;
	WindowRef		window, curPort;

	if ((wins=NewPtr(sizeof(CWindowRecord)))==NULL)
		goto end;
	assert(doc->flags&SLOT_USED);
	if ((window=GetNewCWindow(documentWindowID,wins,dock_window))==NULL)
	{
		dprintf("no window!");
		goto dw;
	}
	// adjust coordinates for this window and the ones that come after it
	MoveWindow(window,current_x+accum_space*12,current_y+accum_space*10,FALSE);
	reposition_window_coordinates();
	*(SInt32*)&arect = 0, arect.right = 480, arect.bottom = 320;
	SizeWindow(window,arect.right,arect.bottom,FALSE);

	// set up graphics
	GetPort(&curPort);
	SetPort(window);
	ClipRect(&arect);

	// add controls
	arect.left = arect.right-16, ++arect.right;
	arect.top = 64, arect.bottom -= 17;
	if ((ctl=NewControl(window,&arect,empty,TRUE,0,0,0,scrollBarProc,0))==NULL)
		goto dispose;
	arect.left = 0, arect.right -= 17;
	arect.top = arect.bottom, arect.bottom += 16;
	if ((ctl=NewControl(window,&arect,empty,TRUE,0,0,0,scrollBarProc,0))==NULL)
		goto dispose;

	SetPort(curPort);
	return window;
dispose:
	SetPort(curPort);
	KillControls(window);
	DisposeWindow(window);
dw:
	DisposePtr(wins);
end:
	return NULL;
}

void reposition_window_coordinates()
{
	if (++accum_space>8)
	{
		accum_space = 0;
		if ((current_x+=8)>63)
			if (current_x>66)
				current_x = 10;
			else
				current_x = 12;
		if ((current_y+=6)>40)
			current_y = 26;
	}
}

void CloseDocumentWindow(documentRec *doc)
{
	WindowRef		window = doc->window;

	assert(window);
	KillControls(window);
	DisposeWindow(window);
}

void UpdateDocWindow(int which)
{
	
}

/* ------- explicit sh** */

OSErr OpenDocumentFile(documentRec **_doc)
{
	SInt64			size;

	documentRec		*doc;
	SInt16			refNum = sizeof(documentRec), index;
	OSErr			err = 0;

	if (_doc==(documentRec**)NULL || (doc=*_doc)==NULL || doc->file_ref>0) goto end;
	if ((err=EvaluateDocumentFile(doc,&size))!=0)
		goto open_err;
	if (doc->flags & _uses_64_bit_file_positioning)
	{
		FSRef		ref;

		if (err=FSpMakeFSRef(&doc->file_owner,&ref))
			goto end;
		err = FSOpenFork(&ref,0,NULL,fsRdWrPerm,&refNum);
	} else
		err = FSpOpenDF(&doc->file_owner,fsRdWrPerm,&refNum);
	if (err)
	{
	open_err:
		if (err==opWrErr)
		{
			if ((index=RFindDoc(refNum))<0)
				*_doc = NULL;
			else
				*_doc = openDocuments.index+index;
		}
		goto end;
	}
	doc->file_ref = refNum;
end:
	return err;
}

/* CloseDocumentFile is in file_audio.c */

OSErr WriteDocument(documentRec *doc, FSSpecPtr customFSSpec)
{
	OSErr			err = 0;

	assert(!(doc->flags&_uses_64_bit_file_positioning)==!(FileSystem_attrs&gestaltFSSupports2TBVols));
	if (customFSSpec)
	{
		FSpDelete(customFSSpec); /* ignore the error */
		if (err=FSpCreate(customFSSpec,AppCreatorType,doc->file_type,0))
			goto end;
	}
	switch(doc->file_type)
	{
	case ftSoundFile:
		break;
	case ftWindowsWaveFile:
		break;
	case AIFCID:
	case AIFFID:
		err = WriteDocument_AIFF(doc);
		break;
	}
end:
	return err;
}

OSErr LoadDocument(documentRec *doc)
{
	OSErr			err = 0;

	assert(!(doc->flags&_uses_64_bit_file_positioning)==!(FileSystem_attrs&gestaltFSSupports2TBVols));
	switch(doc->file_type)
	{
	/* import */
	case ftRawData:
	case ftHyperXDBCustom:
		break;

	case ftSoundFile:
		break;
	case ftWindowsWaveFile:
		break;
	case AIFCID:
	case AIFFID:
		err = ReadDocument_AIFF(doc);
	}
}

/* universal */

#include "format_recognition.h"

SInt32 identify_format(file_data_preview *file_preview, format_id_list *formats)
{
	SInt64			eof, file_offset, lltemp;
	SInt32			*aptr = (SInt32*)formats, list_offset, x, y, cdata[6];
	byte			*bptr, *cptr;
	scanned_block	*block;
	Boolean			skip, last_case, current_case;

	if (file_preview==NULL)
		return paramErr;
	eof = *(SInt64*)&file_preview->eof;
	if (!file_preview->block_count)
		return -2;
	if (*aptr++)
		return -1;				/* version zero only */
	if ((x=*aptr++)<0)
		return 0;		/* empty */
	++x;
	do
	{
		list_offset = (long)formats-(long)aptr;
		aptr += 2;
		aptr = (SInt32*)(1+strlen((char*)aptr)+(byte*)aptr);
		if ((y=*aptr++)<0)				/* empty */
		{
			list_offset = 0;
			continue;
		}
		++y, skip = false, current_case = true;
	/*	assert(!*(byte*)(aptr+1));	*/
		do
		{
			UInt32		spot_data;
			short_int	count, size, filter, data_length;

			SetInt64(file_offset,(char)-(*aptr<0),*aptr++);
			if (Int64Hi(file_offset)<0)
				SAdd64(file_offset,eof);
			spot_data = *aptr++;
			size = spot_size(spot_data);
			data_length = spot_length(spot_data);
			if (skip) size = 0; else
			{
				count = file_preview->block_count, block = file_preview->scan;
				do
				{
					lltemp = file_offset;
					SSub64(lltemp,block->offset);
					assert(block->length>0);
					if (!Int64Hi(lltemp) && Int64Lo(lltemp)<block->length-size)
					{	/* we have found our block */
						cptr = 8-size+(byte*)&cdata;
						*(Float64*)cptr = *(Float64*)(Int64Lo(lltemp)+(byte*)block->data);
						if ((filter=spot_filter(spot_data))==0)
							bptr = (byte*)aptr;
						else
						{
							*(Float64*)(8+cptr) = *(Float64*)aptr;
							bptr = size+(byte*)aptr;
							switch(filter)
							{
							case kSpotFilter_BNOR:
							case kSpotFilter_BOR:
								cdata[0] |= cdata[2], cdata[1] |= cdata[3];
								break;
							case kSpotFilter_BNAND:
							case kSpotFilter_BAND:
								cdata[0] &= cdata[2], cdata[1] &= cdata[3];
								break;
							case kSpotFilter_BXNOR:
							case kSpotFilter_BEXOR:
								cdata[0] ^= cdata[2], cdata[1] ^= cdata[3];
							case kSpotFilter_BNOT:
								break;
							case kSpotFilter_none:
								break;
							default:
								dprintf("unknown filter");
							}
							if (filter & kSpotFilter_BNOT)
								cdata[0] = ~cdata[0], cdata[1] = ~cdata[1];

							if (!(spot_data & 0x8000))
								set_bh(cdata,7-size,0);		/* unsigned - zero extend */
							else							/* signed	- sign extend */
								set_bh(cdata,7-size,-(*cptr<0));

							switch(spot_func(spot_data))
							{
							case kSpotFunc_Equal:
								if (!Int64Eq(*(SInt64*)cdata,*(SInt64*)bptr)) current_case = false;
								break;
							case kSpotFunc_LessThan:
								if (S64Ge(*(SInt64*)cdata,*(SInt64*)bptr)) current_case = false;
								break;
							case kSpotFunc_GreaterThan:
								if (S64Le(*(SInt64*)cdata,*(SInt64*)bptr)) current_case = false;
								break;
							case kSpotFunc_OneOf:
								if (data_length>=size)
								{
									short_int	i = data_length/size;

									set_bh(cdata,7-size,0);
									do
									{
										*(Float64*)(8+cptr) = *(Float64*)bptr;
										set_bh(cdata+2,7-size,0);
										if (!Int64Eq(*(SInt64*)cdata,*(SInt64*)bptr))
										{
											current_case = false;
											break;
										}
										bptr += size;
									} while(--i);
								}
								break;
							case kSpotFunc_NotEqual:
								if (Int64Eq(*(SInt64*)cdata,*(SInt64*)bptr))
									current_case = false;
								break;
							case kSpotFunc_GreaterOrEqual:
								if (S64Lt(*(SInt64*)cdata,*(SInt64*)bptr))
									current_case = false;
								break;
							case kSpotFunc_LessOrEqual:
								if (S64Gt(*(SInt64*)cdata,*(SInt64*)bptr))
									current_case = false;
								break;
							case kSpotFunc_NotOneOf:
								if (data_length>=size)
								{
									short_int	i = data_length/size;

									set_bh(cdata,7-size,0);
									do
									{
										*(Float64*)(8+cptr) = *(Float64*)bptr;
										set_bh(cdata+2,7-size,0);
										if (Int64Eq(*(SInt64*)cdata,*(SInt64*)bptr))
										{
											current_case = false;
											break;
										}
										bptr += size;
									} while(--i);
								}
								break;

							case kSpotFunc_NearOffset:
								SAdd64(*(SInt64*)cdata,file_offset);
							case kSpotFunc_AbsoluteOffset:
								if (U64Ge(*(UInt64*)cdata,*(UInt64*)&eof))
									current_case = false;
								break;
							default:
								dprintf("unknown function");
							}
						}
						goto next;
					}
					++block;
				} while(--count);
				dprintf("couldnt find a scanned block within with the desired range");
				skip = true;
			next:
				;
			}
			aptr = (SInt32*)(data_length+(byte*)aptr);
		} while(--y);
	} while(--x);
	list_offset = 0;
end:
	return list_offset;
}
