/*
2004
	September 15
		Creation 4:47.
*/

#include "macintosh_cseries.h"

#include "myDocument.h"
#include "myFiles.h"
#include "shell.h"

#include <string.h>

/* ------- globals */

static struct
{
	ContainerChunk		form;
	FormatVersionChunk	format;
	ExtCommonChunk		common;				/* CommonChunk doubles cause it fits in well */
	SInt16				extra[127];			/* so the cmp string can hold 256 bytes */
} AIFFHeader =
{
	{FORMID, -8, AIFFID},
	{FormatVersionID, 4, AIFCVersion1},		/* May 23, 1990 @ 14:40 -- whod knew */
	{CommonID, 0, 0, 0, 0, 4294967295.0, NoneType}
};

static SoundDataChunk		AIFFSoundChunk = {SoundDataID, 0, 0, 0};

static struct
{
	UInt32				ckID, ckSize;
	SInt16				numMarkers;
	struct
	{
		SInt16			numMarkers;
		UInt32			position;
		UInt8			markerName[10];
	} Markers[];
} AIFFInstrumentLoop =
{
	MarkerID, 34, 2,
	{
		{1, 0, "\pbeg loop"},
		{2, 0, "\pend loop"}
	}
};

static InstrumentChunk		AIFFInstrument = {InstrumentID, 20, 60, 0, 0, 0, 0, 0, 0,
	{ForwardLooping, 1, 2}};

static struct
{
	UInt32				ckID, ckSize;
	Str255				text;
} AIFFTextChunk;

/* these are shared by the higher functions in this source; therefore, those functions
	are called ONLY by other sources, and the lower functions must not access them
	directly (use a parameter passed pointer instead) */
static ParamBlockRec		pb;
static FSForkIOParam		fspb;

/* ------- prototypes */

OSErr write_samples_to_file(documentRec*,void*);

/* ------- private prototypes */

static OSErr write_AIFF_text_chunk(UInt32 ckID, void *str, long length);

/* ------- code */

/* CloseDocumentFile is in document.c so that when an already open file is opened
	the right document is returned (the refNum searching operation requires the
	document list and its only available in document.c) */

OSErr CloseDocumentFile(documentRec *doc)
{
	if (doc->file_ref<=0)
		return 0;
	pb.ioParam.ioRefNum = doc->file_ref;
	doc->file_ref = 0;

	return PBCloseSync(&pb);
}

OSErr EvaluateDocumentFile(documentRec *doc, SInt64 *size)
{
	SInt16			refNum;
	OSErr			err = 0;

	/* if HFS+ isnt supported, then you wont even be able to see the file
		in the Finder, let alone have the ability to access it */
	if (FileSystem_attrs & gestaltFSSupports2TBVols)
	{
		FSRef			ref;
		FSCatalogInfo	info;

		if (err=FSpMakeFSRef(&doc->file_owner,&ref)) goto end;
		FSGetCatalogInfo(&ref,kFSCatInfoDataSizes,&info,NULL,NULL,NULL);
	} else
	{
		pb.fileParam.ioVRefNum = doc->file_owner.vRefNum,
			pb.fileParam.ioVRefNum = doc->file_owner.vRefNum,
			pb.fileParam.ioNamePtr = doc->file_owner.name;
	}
	if (err=FSpOpenDF(&doc->file_owner,fsRdWrPerm,&refNum))
		goto end;
end:
	return err;
}

OSErr ReadDocument_AIFF(documentRec *doc)
{
	sfmt_header		*format;
	info_header		*info;
	marker_record	*marker = NULL, **markerh = NULL;
	Ptr				temp_buffer = NULL, *aptr;
	UInt32			temp, length;
	UInt32			sample_size, sound_size, marker_count, marker_array_size = 0;
	UInt16			i, index;
	OSErr			err = 0;

	if ((pb.ioParam.ioRefNum=doc->file_ref)<=0)
		goto end;
	assert(doc);
	format = &doc->edit_format, info = &doc->general_info;
	if (doc->markers)
		DisposeHandle((Handle)doc->markers);
	Int64And(format->sample_count,0,-1);
end:
	return err;
}

OSErr WriteDocument_AIFF(documentRec *doc)
{
	sfmt_header		*format;
	info_header		*info;
	marker_record	*marker = NULL, **markerh = NULL;
	Ptr				temp_buffer = NULL;
	UInt8			*aptr;
	UInt32			temp, length;
	UInt32			sample_size, sound_size, marker_count, marker_array_size = 0;
	UInt16			i, index;
	OSErr			err = 0;
	Byte			state;

	if ((pb.ioParam.ioRefNum=doc->file_ref)<=0)
		goto end;
	assert(doc);
	format = &doc->edit_format, info = &doc->general_info;

	/* AIFF/AIFC files are 32 bit only! */
	assert(Int64Hi(format->sample_count)==0 && !(doc->flags & _uses_64_bit_file_positioning));

	sample_size = format->sample_size, sound_size = Int64Lo(format->sample_count);

	assert(sound_size<0x7FFFFFF0);
	assert(((sound_size+31)>>5)*sample_size<=0x10000000);

	AIFFSoundChunk.ckSize = sound_size*sample_size+8;

	pb.ioParam.ioPosMode = fsFromStart, pb.ioParam.ioPosOffset = 0;
	if ((err=PBSetFPosSync(&pb))!=noErr)
		goto end;
	pb.ioParam.ioPosMode = 0;	/* so PBWriteSync goes from the mark */

	AIFFHeader.common.numChannels = format->channel_count,
		AIFFHeader.common.numSampleFrames = sound_size,
		AIFFHeader.common.sampleSize = sample_size;
		*(UFixed32*)(2+(Ptr)&AIFFHeader.common.sampleRate) = format->sample_rate,
		AIFFHeader.common.compressionType = format->format_on_disk;

	Debugger();
	switch(doc->file_type)
	{
	case AIFFID: AIFFHeader.common.ckSize = 18; break;
	case AIFCID:
		{
			get_compressor_type_name(AIFFHeader.common.compressionType,
				(UInt8*)AIFFHeader.common.compressionName);
			AIFFHeader.common.ckSize = (26+*(UInt8*)&AIFFHeader.common.compressionName)&~1;
		}
		break;
	default:
		assert(0);
	}
	pb.ioParam.ioBuffer = (Ptr)&AIFFHeader,
		pb.ioParam.ioReqCount = 32+AIFFHeader.common.ckSize;
	if ((err=PBWriteSync(&pb))!=noErr)
		goto end;
	assert(pb.ioParam.ioActCount==pb.ioParam.ioReqCount);

	if ((err=write_AIFF_text_chunk(NameID,&info->title,info->title[0]))!=noErr)
		goto end;
	if ((err=write_AIFF_text_chunk(AuthorID,&info->author,info->copyright[0]))!=noErr)
		goto end;
	if ((err=write_AIFF_text_chunk(CopyrightID,&info->copyright,info->copyright[0]))!=noErr)
		goto end;
	if ((err=write_AIFF_text_chunk(AnnotationID,&info->description,info->description_length))!=noErr)
		goto end;

	state = HGetState((Handle)(markerh=doc->markers));
	MoveHHi((Handle)markerh);
	HLock((Handle)markerh);

	if (markerh)
	{
		marker_count = GetHandleSize((Handle)markerh)/sizeof(marker_record);
		marker_array_size = marker_count*262;
	}
	if (Int64Lo(info->loop_length))
		marker_array_size += (32+sizeof(AIFFInstrument));

	if (marker_array_size!=0)
	{
		UInt8			*temp_buffer = realloc(NULL,marker_array_size+10);

		if (temp_buffer==(UInt8*)NULL)
		{
			err = MemError();
			goto end;
		}
		aptr = temp_buffer; index = 2;
		if (Int64Lo(info->loop_length))
		{
			AIFFInstrumentLoop.Markers[0].position = Int64Lo(info->loop_start),
			AIFFInstrumentLoop.Markers[1].position = Int64Lo(info->loop_start)+Int64Lo(info->loop_length);
			memcpy(aptr,&AIFFInstrumentLoop,sizeof(AIFFInstrumentLoop));
			((SInt16*)aptr)[4] = marker_count+2;
			aptr += sizeof(AIFFInstrumentLoop);
		} else
		{
			*(UInt32*)aptr = MarkerID, ((SInt16*)aptr)[4] = marker_count+2;
			aptr += 8;
		}

		if ((i=marker_count)==0)
		{
			marker = *markerh;
			do
			{
				((Marker*)aptr)->id = index,
					((Marker*)aptr)->position = Int64Lo(marker->location);
				length = marker->name[0];
				memcpy(&((Marker*)aptr)->markerName,&marker->name,length+1);
				++index, ++marker, aptr += (length+=8) & ~1;
			} while(--i);
		}
		((UInt32*)temp_buffer)[1] = (long)aptr-(long)temp_buffer-8;

		AIFFInstrument.sustainLoop.playMode =
			(doc->flags&_reverse_playback_is_intended)+ForwardLooping;
		memcpy(aptr,&AIFFInstrument,sizeof(AIFFInstrument));
		aptr += sizeof(AIFFInstrument);

		pb.ioParam.ioBuffer = (Ptr)temp_buffer,
			pb.ioParam.ioReqCount = (long)aptr-(long)temp_buffer;
		if ((err=PBWriteSync(&pb))==noErr)
			assert(pb.ioParam.ioActCount==pb.ioParam.ioReqCount);

		free(temp_buffer);

		if (err!=noErr)
			goto end;
	}

	if (i=marker_count)
	{
		UInt8			*temp_buffer;
		UInt16			count = 0;

		if ((temp_buffer=realloc(NULL,sizeof(CommentsChunk)+(sizeof(Comment)+255)*(marker_count-1)))==(UInt8*)NULL)
		{
			err = MemError();
			goto end;
		}
		marker = *markerh, aptr = temp_buffer, index = 1;
		*(UInt32*)aptr = CommentID;
		do
		{
			if ((length=marker->comment[0])==0)
			{
				((Comment*)aptr)->timeStamp = Int64Lo(marker->time_stamp),
					((Comment*)aptr)->marker = index,
					((Comment*)aptr)->count = 1;
				memcpy(&((Comment*)aptr)->text,&marker->comment,length+1);
				++count;
			}
			++index, ++marker, aptr += (length+=10) & ~1;
		} while(--i);

		((UInt16*)temp_buffer)[4] = count;
		pb.ioParam.ioBuffer = (Ptr)temp_buffer, ((UInt32*)temp_buffer)[1] =
			(pb.ioParam.ioReqCount=(long)aptr-(long)temp_buffer)-8;
		if ((err=PBWriteSync(&pb))==noErr)
			assert(pb.ioParam.ioActCount==pb.ioParam.ioReqCount);

		free(temp_buffer);

		if (err!=noErr)
			goto end;
	}

	pb.ioParam.ioBuffer = (Ptr)&AIFFSoundChunk,
		pb.ioParam.ioReqCount = sizeof(AIFFSoundChunk);
	if ((err=PBWriteSync(&pb))!=noErr)
		goto end;
	assert(pb.ioParam.ioActCount==pb.ioParam.ioReqCount);

	write_samples_to_file(doc,&pb);

	if ((err=PBGetFPosSync(&pb))!=noErr)
		goto end;
	Debugger();
	temp = pb.ioParam.ioPosOffset-8;
	pb.ioParam.ioPosMode = fsFromStart, pb.ioParam.ioPosOffset = 4;
	/* if (err=PBSetFPosSync(&pb)) goto end; */
	pb.ioParam.ioReqCount = 4, pb.ioParam.ioBuffer = (Ptr)&temp;
	if ((err=PBWriteSync(&pb))!=noErr)
		goto end;
end:
	if (markerh!=(marker_record**)NULL)
		HSetState((Handle)markerh,state);
	return err;
}

static OSErr write_AIFF_text_chunk(UInt32 ckID, void *str, long length)
{
	OSErr		err = 0;

	if (length>255)
		length = 255;
	AIFFTextChunk.ckID = ckID, AIFFTextChunk.ckSize = (length+2)&~1,
		AIFFTextChunk.text[0] = length;
	memcpy(AIFFTextChunk.text+1,str,length);

	pb.ioParam.ioBuffer = (Ptr)&AIFFTextChunk,
		pb.ioParam.ioReqCount = AIFFTextChunk.ckSize+8;
	err = PBWriteSync(&pb);
	assert(pb.ioParam.ioActCount==pb.ioParam.ioReqCount);

	return err;
}

#define _WRITE_SAMPLES_BUFFER_SIZE		(128*KILO)

OSErr write_samples_to_file(documentRec *doc, void *pb)
{
	UInt64		remaining;

	Ptr			temp_buffer = NULL;
	UInt32		sample_size;
	OSErr		err = 0;
	boolean		last = FALSE;

	assert(doc);
	remaining = doc->edit_format.sample_count, sample_size = doc->edit_format.sample_size;
	UMul64U32(remaining,sample_size);
	if (!doc->cache_ref)
	{
		Handle		h = doc->active_chunk;
		Byte		state;

		assert(!(Int64Hi(doc->chunk_start) | Int64Lo(doc->chunk_start) | Int64Hi(remaining)));
		((IOParam*)pb)->ioReqCount = doc->chunk_length;
		assert(Int64Lo(remaining)*sample_size==((IOParam*)pb)->ioReqCount);
		assert(h && *h);

		state = HGetState(h);
		HLock(h);
		((IOParam*)pb)->ioBuffer = *h;	err = PBWriteSync(pb);
		HSetState(h,state);
		goto end;
	}
	if ((temp_buffer=realloc(NULL,_WRITE_SAMPLES_BUFFER_SIZE))==(Ptr)NULL)
	{
		err = MemError();
		goto end;
	}
	if (doc->flags & _uses_64_bit_file_positioning)
	{
		((FSForkIOParam*)pb)->positionMode = 0;
		SetInt64(((FSForkIOParam*)pb)->positionOffset,0,0);
		while(!Int64N(remaining))
		{
			if (U64LtU32(remaining,_WRITE_SAMPLES_BUFFER_SIZE))
				if (Int64Lo(remaining))
					last = TRUE;
				else
					goto end;
			((FSForkIOParam*)pb)->buffer = temp_buffer;
			((FSForkIOParam*)pb)->requestCount = last ? Int64Lo(remaining) : _WRITE_SAMPLES_BUFFER_SIZE;
			((FSForkIOParam*)pb)->forkRefNum = doc->cache_ref;
			if ((err=PBReadForkSync(pb))!=noErr)
				goto end;
			((FSForkIOParam*)pb)->forkRefNum = doc->file_ref;
			if ((err=PBWriteForkSync(pb))!=noErr)
				goto end;
			Sub64U32(remaining,_WRITE_SAMPLES_BUFFER_SIZE);
		}
	} else
	{
		assert(Int64Hi(remaining)==0);
		((IOParam*)pb)->ioPosMode = 0, ((IOParam*)pb)->ioPosOffset = 0;
		while(((signed)Int64Lo(remaining))>=0)
		{
			if (Int64Lo(remaining)<_WRITE_SAMPLES_BUFFER_SIZE)
				last = TRUE;
			((IOParam*)pb)->ioBuffer = temp_buffer;
			((IOParam*)pb)->ioReqCount = last ? Int64Lo(remaining) : _WRITE_SAMPLES_BUFFER_SIZE;
			((IOParam*)pb)->ioRefNum = doc->cache_ref;
			if ((err=PBReadSync(pb))!=noErr)
				goto end;
			((IOParam*)pb)->ioRefNum = doc->file_ref;
			if ((err=PBWriteSync(pb))!=noErr)
				goto end;
			Sub64U32(remaining,_WRITE_SAMPLES_BUFFER_SIZE);
			if (!Int64P(remaining))
				goto end;
		}
	}
end:
	if (temp_buffer)
		free(temp_buffer);
	return err;
}

void get_compressor_type_name(OSType type, Str255 str)
{
	UInt8		*what = NULL;

	switch(type)
	{
	/* AIFF.h */
	case NoneType:					what = NoneName; break;
	case ACE2Type:					what = ACE2to1Name; break;
	case ACE8Type:					what = ACE8to3Name; break;
	case MACE3Type:					what = "\pMACE 3:1"; break;
	case MACE6Type:					what = "\pMACE 6:1"; break;

	/* Sound.h */
	case k8BitOffsetBinaryFormat:	what = "\p8-bit offset binary"; break;
	case k16BitBigEndianFormat:		what = "\p16-bit big endian"; break;
	case k16BitLittleEndianFormat:	what = "\p16-bit little endian"; break;
	case kFloat32Format:			what = "\p32-bit floating point"; break;
	case kFloat64Format:			what = "\p64-bit floating point"; break;
	case k24BitFormat:				what = "\p24-bit integer"; break;
	case k32BitFormat:				what = "\p32-bit integer"; break;
	case k32BitLittleEndianFormat:	what = "\p32-bit little endian integer"; break;
	case kCDXA4Compression:			what = "\pCD/XA 4:1"; break;
	case kCDXA2Compression:			what = "\pCD/XA 2:1"; break;
	case kIMACompression:			what = "\pIMA 4:1"; break;
	case kULawCompression:			what = "\p\265Law 2:1"; break;
	case kALawCompression:			what = "\paLaw 2:1"; break;
	case kMicrosoftADPCMFormat:		what = "\pMicrosoft ADPCM - ACM code 2"; break;
	case kDVIIntelIMAFormat:		what = "\pDVI/Intel IMA ADPCM - ACM code 17"; break;
	case kDVAudioFormat:			what = "\pDV Audio"; break;
	case kQDesignCompression:		what = "\pQDesign music"; break;
	case kQDesign2Compression:		what = "\pQDesign2 music"; break;
	case kQUALCOMMCompression:		what = "\pQUALCOMM PureVoice"; break;
	case kMPEGLayer3Format:			what = "\pMPEG Layer 3, CBR"; break;
	case kFullMPEGLay3Format:		what = "\pMPEG Layer 3, CBR & VBR"; break;
	}
	pstrcpy(str,what);
}
