/*
 iocustom.c
  custom runtime routines for MyCSeries version 0.1.13

  Version 0.1.13

  Created: December 25, 2004
  Revised: February 16, 2006

  Written (mostly) by A.P.

  This is a sample file for the remaining runtime routines on MacOS Classic 68K.
  Programmers compiling with Symantec C for PowerPC shouldn't use this.
  The appropriate MyCSeries library file should be linked with this when compiled.

  This will _not_ work with MacTraps -- replace MacTraps with Interface.o.
*/

/*
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <cseries.h>
#include <cs_math.h>
#include <cs_strings.h>
#include <cs_system.h>
#include <cs_byte_swapping.h>

#include <SizeTDef.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <ctype.h>
#include <errno.h>

/* constants */

/*** redefine this to your application signature, such as the shell name ***/
#ifndef APPLICATION_SIGNATURE
# define APPLICATION_SIGNATURE          "AudioWarp"
#endif

static const char    _log_date[] = "%s-- %d-%m-%Y %H:%M:%S\xd";

/* globals */

#ifdef THINK_C
# include <QuickDraw.h>
# ifdef env68k
   QDGlobals    qd;
# endif
#endif

/* these may be translated into other languages */
const char      _assertStr[] = "assertion failure",
                    _haltStr[] = "halt!",
                    _warnStr[] = "warning signaled",
                    _pauseStr[] = "pause";

char            temporary[384]; /* 384 bytes to avoid overflow */

static const char    opening_str[] = "Opening log for " APPLICATION_SIGNATURE;
static const char    closing_str[] = "Closing log";

boolean         sys_debugger_available = FALSE, use_sys_debugger = FALSE;

/* ------- debugging code */

#ifdef mac
# include <Folders.h>
# include <MacMemory.h>

/* their fault, not mine */
 /* to prevent "dprintf" from being defined as hex inline */
# undef SIXWORDINLINE
# define SIXWORDINLINE(a,b,c,d,e,f)

# include <Debugging.h>
#endif

#ifdef env68k
# pragma parameter __D0 Get68KInterruptLevel()
  extern short Get68KInterruptLevel(void) = {0x40C0,0xE048,0x240,7};
# undef ATTASKLEVEL0
# define ATTASKLEVEL0                   (Get68KInterruptLevel()==0)
#elif envppc
# ifndef ATTASKLEVEL0
#  define 
# endif
#else
# define ATTASKLEVEL0                   TRUE        /* ?? */
#endif

#undef exit
 void exit(int);

static UInt8        _dbgTemp[512];      /* debug messaging temporary */

static FILE         *debugging_log = NULL;

/*** it is recommended that the user add their own system debugger support here ***/

void initialize_debugger(boolean sysOnly)
{
   close_debugging_log(); /* just in case */

#ifdef mac
   sys_debugger_available = ISLOWLEVELDEBUGGERCALLABLE() ? TRUE : FALSE;
#else
   use_sys_debugger = FALSE;
#endif

   if (sysOnly)
   {
      use_sys_debugger = sys_debugger_available;
   } else
   {
      if ((debugging_log=fopen("debugging_log","rwb"))==NULL) return;

      /* we are appending to the log, not starting it */
      fseek(debugging_log,0,EOF);
      /* append the date */
      fprintf(debugging_log,_log_date, opening_str);
   }
}

void close_debugging_log()
{
   if (debugging_log!=NULL)
   {
      fprintf(debugging_log,_log_date, closing_str);
      fclose(debugging_log);
      debugging_log = NULL;
   }
}

void debugger_log(const void *s, int z)
{
   if (debugging_log==NULL)
   {
      return;
   }
   fprintf(debugging_log,_log_date, (char*)empty);
   fwrite(s,z,1,debugging_log);
}

void _debugMsg(const char *str, const char *file, int line, int level)
{
#ifdef mac
   psprintf(_dbgTemp,"file \xD2%s\xD3, line #%d: %s",file,line,str);
#else
   psprintf(_dbgTemp,"file \"%s\", line #%d: %s",file,line,str);
#endif

   if (DEBUG && sys_debugger_available)
   {
#ifdef mac
      DebugStr((UInt8*)_dbgTemp);
#endif
   } else
      debugger_log(&_dbgTemp[1],_dbgTemp[0]);

   if (level<0)
   {
      exit(1);
#ifdef mac
      DebugStr("\pFailed to exit.  You must manually terminate.");
#endif
   }
}

void dprintf(const char *fmt, ...)
{
   char         *s;
   int          len;

   if ((len=vsprintf((char*)&_dbgTemp[1],fmt,(void*)(1+(int*)&fmt)))>255) len = 255;

   s = memchr((char*)_dbgTemp,';',_dbgTemp[0]=len);

   if (debugging_log==NULL || !ATTASKLEVEL0 || s==NULL)
   {
      if (sys_debugger_available)
      {
#ifdef mac
          DebugStr(_dbgTemp);
#endif
      } else
      {
         /* otherwise, we are screwed */
      }
   } else debugger_log(&_dbgTemp[1],_dbgTemp[0]);
}

/* ------- io stuff */

#ifdef THINK_C

  void exit(int x)
  {
#pragma unused (x)
     ExitToShell();
  }

  /* make sure that this prototype is defined */
  extern int cs_exit(int);

 /* this is Symantec C's way of handling "startup" */
  extern int main(int, char**);

# if _USE_START_ && !_USE_APPLE_LIBRARIES_
   void __start(void)
   {
      char  *mainName = "main";

      main(1,&mainName);

      cs_exit(1);
   }
# endif

# if defined(env68k) && !TARGET_RT_MAC_68881

/*
-- from HyperXDB

function asin x
  if abs(x).5 then return atan(x/sqrt(1-x*x)) else return atan(x/sqrt(2-2*x-(1-x)^2))
end asin
function acos x
  return 2*atan(sqrt((1-x)/(1+x)))
end acos
function cot x
  return 1/tan(x)
end cot
function sinh x
  put exp1(x)into x
  return(x/(x+1)+x)/2
end sinh
function cosh x
  put exp(x)into x
  return(x+1/x)*.5
end cosh
function tanh x
  put exp(x)into x
  return(x-1/x)/(x+1/x)
end tanh
*/

   double_t asin(double_t x)
   {
      double_t     temp = double_t_one-x;

      if (fabs(x)<=0.5) return atan(x/sqrt(double_t_one-x*x));
      temp *= temp;
      return atan(x/sqrt(2-2*x-temp));
   }

   double_t acos(double_t x) { return 2*atan(sqrt((double_t_one-x)/(double_t_one+x))); }

# endif

  double_t cot(double_t x) { return sin(x)/cos(x); }
#endif

/* mem additions */

/* faster (and more reliable at not moving the handle around) */
#if TARGET_OS_MAC
# define bigblockcpy(d,s,z)             BlockMove(s,d,z)
#else
# define bigblockcpy(d,s,z)             memcpy(d,s,z)
#endif

/* this version may create fragmentation - use platform_handles for resizing instead */
void *rmalloc(void *p, size_t z)
{
   if (p==NULL)
   {
      platform_handle   h = (platform_handle)NULL;
#if TARGET_OS_MAC
      OSErr             err;
#endif
      boolean           t = FALSE;

      /* create a temporary buffer, preferably outside our heap */
#if TARGET_OS_MAC
      h = TempNewHandle(z,&err);
#endif
      if (h==(platform_handle)NULL) h = new_platform_handle(z,FALSE); else t = TRUE;
      if (h==(platform_handle)NULL) return NULL; /* failure */

      /* copy to the temp buffer so the data is saved before reallocating */
      bigblockcpy(*h,p,z);

      if ((p=realloc(p,z))!=NULL)
         bigblockcpy(p,*h,z); /* only if realloc succeeded */

      if (t)
      {
#if TARGET_OS_MAC
         TempDisposeHandle(h,&err);
#else
         assert(0);
#endif
      } else
      {
         dispose_platform_handle(h);
      }
   }

   return p;
}

/* printf */

UInt8 *psprintf(void *d, const char *fmt, ...)
{
   int          len;

   if ((len=vsprintf((char*)(1+(UInt8*)d),fmt,(void*)(1+(int*)&fmt)))>255) len = 255;
   *(char*)d = len;

   return((UInt8*)d);
}

char *csprintf(char *d, const char *fmt, ...)
{
   vsprintf(d,fmt,(void*)(1+(int*)&fmt));

   return(d);
}

static const char   form_chars[] = "+-.0123456789Zbhlz%PXcdiopsuxEGefgn";

enum
{ /* modifier form chars */
   _fc_plus,
   _fc_minus,
   _fc_period,
   _fc_zero,
   _fc_one,
   _fc_two,
   _fc_three,
   _fc_four,
   _fc_five,
   _fc_six,
   _fc_seven,
   _fc_eight,
   _fc_nine,
   _fc_big_endian, /* special type */
   _fc_byte, /* special type */
   _fc_half, /* special type - 16 bits (short) */
   _fc_long,
   _fc_little_endian, /* special type */

  /* printing form chars */
   _fc_escape,
   _fc_Pascal_string,
   _fc_Hexadecimal,
   _fc_char,
   _fc_decimal,
   _fc_int,
   _fc_octal,
   _fc_pointer,
   _fc_string,
   _fc_unsigned,
   _fc_hexadecimal,
   _fc_Exponential,
   _fc_Ranging_notation,
   _fc_exponential,
   _fc_floating,
   _fc_ranging_notation,
   _fc_notation
};

#define _CHAR_INDEX(s,c)                _char_index(s,c,sizeof(s))

static int _char_index(const char *s, int c, int z)
{
   char         *r;

   if ((r=memchr(s,c,z))==NULL) return NONE;
   return (long)r-(long)s;
}

enum
{ /* dsprintf flags */
   _pf_long     = 0x1
};

UInt8 *dsprintf(void *_d, const char *fmt, const void *_s)
{
   /* to store the valid escape stuff before calling sprintf */
   char         fbuf[16];
   /* to store large data for endian conversion */
   char         dbuf[MAX(16,sizeof(long double))];

   const char   *t = fmt, *e, *p;
   UInt8        *d = (UInt8*)_d, *s = (UInt8*)_d, c;
   int          len = strlen(t), size, index, endian, value;
   int          f, z;
   boolean      flags = 0;

   e = len+fmt;

   while(t<e && (t=memchr(p=t,'%',len))!=NULL)
   {
      memcpy(d,p,z=(long)t-(long)p); /* copy the string in between */
      d += z;        /* increment the destination */
      z = 4;         /* default data size is four bytes */
      p = t;         /* this is so we can increment `t' without feeling guilty */
      endian = NONE; /* clear the endian specific */
      f = -1;        /* set up our valid escape (turns non-negative at first) */
      flags = 0;     /* and clear the flags */

      switch(_CHAR_INDEX(form_chars,c=*++t))
      {
      case _fc_half:
      case _fc_long:
      case _fc_plus:
      case _fc_minus:
      case _fc_period:
      case _fc_zero:
      case _fc_one:
      case _fc_two:
      case _fc_three:
      case _fc_four:
      case _fc_five:
      case _fc_six:
      case _fc_seven:
      case _fc_eight:
      case _fc_nine:
         while((unsigned)(index=_CHAR_INDEX(form_chars,c=*t++))<_fc_escape)
         {
            switch(index)
            {
            case NONE:
               break;
            case _fc_big_endian:
            case _fc_little_endian:
               endian = index;
               break;
            case _fc_half: z = 2; break;
            case _fc_byte: z = 1; break;
            case _fc_long:
               if (flags & _pf_long) z = 8; /* two `l's mean "long long" */
               flags |= _pf_long;
            default:
               if (f<=9) fbuf[++f] = c;
               break;
            }
         }
         if (f>9) f = 9;
         if ((fbuf[++f]=c)==0) goto end; /* we could get termination at any time */

         fbuf[++f] = 0; /* and terminate our temporary format */

         if (index<=_fc_hexadecimal)
         {
            /* reflect whether the value is signed */
            index = index==_fc_hexadecimal ? NONE : index==_fc_decimal || index==_fc_int;

            if (z==1)
               value = index ? *(SInt8*)d : *(UInt8*)d;
            else if (z==2)
            {
               value = *(UInt16*)d;
               if (endian==_fc_big_endian)
                  value = index ? BES16_TO_NATIVE(value) : BEU16_TO_NATIVE(value);
               else if (endian==_fc_little_endian)
                  value = index ? LES16_TO_NATIVE(value) : LEU16_TO_NATIVE(value);
               else if (index)
                  value = (SInt16)value;
            } else if (z==4)
            {
               value = *(UInt32*)d;
               if (endian==_fc_big_endian)
                  value = BEU32_TO_NATIVE(value);
               else if (endian==_fc_little_endian)
                  value = LEU32_TO_NATIVE(value);
            } else if (z==8)
            {
               d += sprintf((char*)d,fbuf,value);
               s += 8;
               continue;
            }
            d += sprintf((char*)d,fbuf,value);
         } else
         {
            /* floating-point */
            if (flags & _pf_long)
            { /* 80, 96 or 128 bits (unreliable) */
               memcpy(dbuf,s,z=sizeof(long double));
               if (endian!=NONE)
                  if ((TARGET_RT_LITTLE_ENDIAN==1)!=(endian==_fc_little_endian))
                     reverse_bytes(z,dbuf);
               d += sprintf((char*)d,fbuf,(long double*)dbuf);
            } else
            { /* 64-bit double (always - for compatibility) */
               memcpy(dbuf,s,z=8);
               if (endian!=NONE)
                  if ((TARGET_RT_LITTLE_ENDIAN==1)!=(endian==_fc_little_endian))
                     reverse_bytes(z,dbuf);
               d += sprintf((char*)d,fbuf,(double*)dbuf);
            }
         }
         s += z;
         continue;

      case _fc_decimal:
      case _fc_unsigned:
      case _fc_octal:
      case _fc_hexadecimal:
         /* nothing special; just let vsprintf handle it */
         break;

      case _fc_escape:
         break;
      default:
         if (c==0) goto end;
      }
      fbuf[0] = c, fbuf[1] = 0;
      d += vsprintf((char*)d,fbuf,(void*)&s);
      s += z;
   }
end:
   if (t==NULL) t = p;
   if (t<e && t!=NULL) memcpy(d,t,(long)e-(long)t);

   return((UInt8*)_d);
}
