/* stolen from printf.c, Copyright 1995 Symantec */

#include <Endian.h>
#if TARGET_CPU_68K
#undef Endian16_Swap
#undef Endian32_Swap
#endif

/*typedef char *va_list;*/
#include "tstdio.h"
#include "stdarg.h"
#include "string.h"
#include "ansi_private.h"

#include <math.h>
#include <THINK SANE.h>

#define BUFLEN			512
#define TRUE			1
#define FALSE			0

static struct format
{
	unsigned 		leftJustify		: 1;
	unsigned 		forceSign		: 1;
	unsigned 		altForm			: 1;
	unsigned 		zeroPad			: 1;
	unsigned 		havePrecision	: 1;
	unsigned 		hSize			: 1;
	unsigned 		lSize			: 1;
	unsigned 		LSize			: 1;
	unsigned 		altEndian		: 1;	/* added for crap */
	char			sign;
	char			exponent;
	int				fieldWidth;
	int				precision;
} default_format;


struct decrec
{
	decimal		dec;
	/* following fields arent used by SANE */
	short		min, dot, max;
};

static void ftod(int, int, struct decrec *, long double);


int lprintf(FILE *fp, register const char *fmt, register va_list arg)
{
	register int				i, j, nwritten = 0;
	register unsigned char		c;
	register unsigned long		n;
	register char				*s;

	long double			x;
	char				buf[BUFLEN], *digits, *t;
	struct format		F;
	struct decrec		D;

	while(c=*fmt++)
	{
		if (c!='%')
			goto copy1;
		F = default_format;

			/*  decode flags  */

		for(;;)
		{
			c = *fmt++;
			if (c=='-')
				F.leftJustify = TRUE;
			else if (c=='+')
				F.forceSign = TRUE;
			else if (c==' ')
				F.sign = ' ';
			else if (c=='#')
				F.altForm = TRUE;
			else if (c=='0')
				F.zeroPad = TRUE;
			else
				break;
		}

			/*  decode field width  */

		if (c=='*')
		{
			if ((F.fieldWidth=va_arg(arg,int))<0)
				F.leftJustify = TRUE, F.fieldWidth = -F.fieldWidth;
			c = *fmt++;
		} else
			for(; c>='0' && c<='9'; c=*fmt++)
				F.fieldWidth = (10*F.fieldWidth)+(c-'0');

			/*  decode precision  */

		if (c=='.')
		{
			if ((c=*fmt++)=='*')
			{
				F.precision = va_arg(arg,int);
				c = *fmt++;
			} else
				for(; c>='0' && c<='9'; c=*fmt++)
					F.precision = (10*F.precision)+(c-'0');

			if (F.precision>=0)
				F.havePrecision = TRUE;
		}
		if (!c) goto done;

			/*  perform appropriate conversion  */

		s = &buf[BUFLEN];
		if (F.leftJustify)
			F.zeroPad = FALSE;
conv:	switch(c)
		{
				/*  ':' endian modifier - added 21-5-2005 */
			case ':':
				F.altEndian = TRUE, c = *fmt++;
				goto conv;
				/*  'h' size modifier  */
			case 'h':
				F.hSize = TRUE, c = *fmt++;
				goto conv;

				/*  'l' size modifier  */
			case 'l':
				F.lSize = TRUE, c = *fmt++;
				goto conv;

				/*  'L' size modifier  */
			case 'L':
				F.LSize = TRUE, c = *fmt++;
				goto conv;

				/*  decimal (signed)  */
			case 'd':
			case 'i':
				if (!F.altEndian)
				{
					n = va_arg(arg,int);
					if (F.hSize) n = (short)n;
				} else
					if (F.hSize)
						n = Endian16_Swap(va_arg(arg,int));
					else
						n = Endian32_Swap(va_arg(arg,int));

				if ((long)n<0)
					n = -n, F.sign = '-';
				else if (F.forceSign)
					F.sign = '+';
				goto decimal;

				/*  decimal (unsigned)  */
			case 'u':
				if (!F.altEndian)
				{
					n = va_arg(arg,unsigned int);
					if (F.hSize) n = (short)n;
				} else
					if (F.hSize)
						n = Endian16_Swap(va_arg(arg,unsigned int));
					else
						n = Endian32_Swap(va_arg(arg,unsigned int));
				F.sign = 0;
				goto decimal;

				/*  decimal (common code)  */
			decimal:
				if (!F.havePrecision)
				{
					if (F.zeroPad)
					{
						F.precision = F.fieldWidth;
						if (F.sign)
							--F.precision;
					}
					if (F.precision<1)
						F.precision = 1;
				}
#if !__powerc && __option(thinkc) && __option(mc68020)
				asm
				{
					moveq		#0,i
					tst.l		n
					beq.s		@3
					addq.l		#1,i
					divul.l		#10,d0:n
					beq.s		@2
			@1		addi.b		#'0',d0
					move.b		d0,-(s)
					addq.l		#1,i
					divul.l		#10,d0:n
					bne.s		@1
			@2		addi.b		#'0',d0
					move.b		d0,-(s)
			@3	}
#else
				for(i=0; n; n/=10,i++)
					*--s = n%10 + '0';
#endif
				for(; i<F.precision; i++)
					*--s = '0';
				if (F.sign)
					*--s = F.sign, ++i;
				break;

				/*  octal (unsigned)  */

			case 'o':
				if (!F.altEndian)
				{
					n = va_arg(arg,unsigned int);
					if (F.hSize) n = (short)n;
				} else
					if (F.hSize)
						n = Endian16_Swap(va_arg(arg,unsigned int));
					else
						n = Endian32_Swap(va_arg(arg,unsigned int));
				if (!F.havePrecision)
				{
					if (F.zeroPad)
						F.precision = F.fieldWidth;
					if (F.precision<1)
						F.precision = 1;
				}
				for(i=0; n; n /= 8,i++)
					*--s = n % 8 + '0';
				if (F.altForm && i && *s!='0')
				{
					*--s = '0';
					i++;
				}
				for(; i<F.precision; i++)
					*--s = '0';
				break;

				/*  hexadecimal (unsigned)  */

			case 'p':
				F.havePrecision = F.lSize = TRUE, F.precision = 8;
				/* ... */
			case 'X':
				digits = "0123456789ABCDEF";
				goto hexadecimal;
			case 'x':
				digits = "0123456789abcdef";
				/* ... */
			hexadecimal:
				if (!F.altEndian)
				{
					n = va_arg(arg,unsigned int);
					if (F.hSize) n = (short)n;
				} else
					if (F.hSize)
						n = Endian16_Swap(va_arg(arg,unsigned int));
					else
						n = Endian32_Swap(va_arg(arg,unsigned int));
				if (!F.havePrecision)
				{
					if (F.zeroPad)
					{
						F.precision = F.fieldWidth;
						if (F.altForm)
							F.precision -= 2;
					}
					if (F.precision<1)
						F.precision = 1;
				}
				for(i=0; n; n/=16, i++)
					*--s = digits[n % 16];
				for(; i < F.precision; i++)
					*--s = '0';
				if (F.altForm)
				{
					*--s = c;
					*--s = '0', i += 2;
				}
				break;

#ifndef _NOFLOATING_
				/*  fixed-point  */

			case 'f':
				if (F.LSize)
					x = va_arg(arg,long double);
				else
					x = va_arg(arg,double);
				if (!F.havePrecision)
					F.precision = 6;
				ftod(1,F.precision,&D,x);
				D.dot = D.dec.exp+D.dec.sig.length;
				if ((D.min=D.dot)>1)
					D.min = 1;
				D.max = D.dot+F.precision;
				if (D.max-D.min>508)
					memcpy(&D.dec.sig,"\6>>>>>>",7);
				goto floating;

				/*  floating-point  */

			case 'e':
			case 'E':
				if (F.LSize)
					x = va_arg(arg,long double);
				else
					x = va_arg(arg,double);
				if (!F.havePrecision)
					F.precision = 6;
				F.exponent = c;
				ftod(0,D.max = F.precision+1,&D,x);
				D.min = D.dot = 1;
				D.dec.exp += D.dec.sig.length-1;
				goto floating;

				/*  "general" notation  */

			case 'g':
			case 'G':
				if (F.LSize)
					x = va_arg(arg,long double);
				else
					x = va_arg(arg,double);
				if (!F.havePrecision)
					F.precision = 6;
				else if (F.precision==0)
					F.precision = 1;
				F.exponent = c-2;
				ftod(0,D.max=F.precision,&D,x);
				D.min = D.dot = 1;
				D.dec.exp += D.dec.sig.length - 1;
				if (D.dec.exp >= -4 && D.dec.exp<F.precision)
				{
					F.exponent = 0;
					if ((D.dot += D.dec.exp) < 1)
						D.min = D.dot;
				}
				if (!F.altForm && D.max>D.dec.sig.length)
					if ((D.max=D.dec.sig.length)<D.dot)
						D.max = D.dot;
				goto floating;

					/*  floating (common code)  */

			floating:
				if (D.dec.sig.text[0] > '9')
				{
					F.exponent = FALSE;
					D.dot = 0, D.min = 1, D.max = D.dec.sig.length;
				}
				i = 0;
				if (F.exponent)
				{
					n = D.dec.exp;
					if (D.dec.exp<0) n = -n;
					for(; n; n /= 10, i++)
						*--s = n % 10 + '0';
					for(; i<2; i++)
						*--s = '0';
					*--s = D.dec.exp < 0 ? '-' : '+';
					*--s = F.exponent, i += 2;
				}
				j = D.max;
				if (j==D.dot && !F.altForm)
					++D.dot;
				do
				{
					if (j==D.dot)
					{
						*--s = '.', ++i;
					}
					if (j > 0 && j <= D.dec.sig.length) *--s = D.dec.sig.text[j-1];
					else *--s = '0';
				} while(--j >= D.min);
				i += D.max - j;
				if (D.dec.sgn)
					F.sign = '-';
				else if (F.forceSign)
					F.sign = '+';
				if (F.zeroPad)
				{
					j = F.fieldWidth;
					if (F.sign)
						--j;
					for(; i < j; i++)
						*--s = '0';
				}
				if (F.sign)
				{
					*--s = F.sign;
					i++;
				}
				break;
#endif /* _NOFLOATING_ */

				/*  character  */

			case 'c':
				*--s = va_arg(arg,int), i = 1;
				break;

				/*  string  */
/* A.P.: added this little f**ker; 22-3-2004 2:28 */
			case 'P':
				F.altForm = TRUE;
			case 's':
				s = va_arg(arg,char*);
				if (F.altForm)
				{
					i = (unsigned char)*s++;
					if (F.havePrecision && i>F.precision)
						i = F.precision;
				} else
				{
					if (!F.havePrecision)
						i = strlen(s);
					else if (t=(char*)memchr(s,'\0',F.precision))
						i = t-s;
					else
						i = F.precision;
				}
				break;

				/*  store # bytes written so far  */

			case 'n':
				s = (char*)va_arg(arg,void*);
				if (F.hSize)
					*(short*) s = nwritten;
				else if (F.lSize)
					*(long*) s = nwritten;
				else
					*(int*) s = nwritten;
				continue;

				/*  oops - unknown conversion, abort (if not %)  */

			default:
				if (c!='%')
					goto done;
			copy1:
				if (putc(c,fp)<0)
					return(EOF);
				++nwritten;
				continue;
		}

			/*  pad on the left  */

		if (i < F.fieldWidth && !F.leftJustify)
			do
			{
				if (putc(' ',fp)<0)
					return(EOF);
				++nwritten;
			} while(i<--F.fieldWidth);

			/*  write the converted result  */

		if (fwrite(s,1,i,fp)!=i)
			return(EOF);
		nwritten += i;

			/*  pad on the right  */

		for(; i<F.fieldWidth; i++)
		{
			if (putc(' ',fp)<0)
				return(EOF);
			++nwritten;
		}
	}
done:
	return(nwritten);
}
