/*
 main.c

All dates in international format (DD-MM-YYYY)

A.P.:
2004
	September 5
		Math stuff (especially the prax and F4b4 math stuff) was moved to
			newmath.c.
	10PM
		Thinking about stuff.  Reformatting sh**.  Whatever.
	September 6
	5AM
		Testing out RMS with sine waves.
	September 8
	10AM
		Testing out the newer sine function.  Its expected to give bad results.
	September 10
		Testing out the newer sine function, again.  This time is expected to
			pass in all respects.
	September 11-17
		Worked on interface and structure finishes.
	(Some day in October)
		It takes about two seconds (not timed) to generate two seconds of a
			middle C sine wave at a sample rate of 46,200 per second with the
			newest sine function (USine16).
	December 30
	1PM-2PM
		Minor changes.  A few more additions.
2005
	March 5
		The new isqrt hex inline (optimized for 68040) averages about 4.7s each
		execution!  Time tests performed on a 33MHz 68LC040.
2006
	February 15 3:08
		Got CLib, IntEnv and MacRuntime compiled in here!  And all it took was
		weeks of manual optimizations.... **arghh**.
*/

#include <Traps.h>
#include <QuickDraw.h>
#include <Math64.h>

#include <math.h>
#include <string.h>

#include "macintosh_cseries.h"

#include "shell.h"
#include "interface.h"
#include "screen.h"

#include "ae.h"
#include "myDocument.h"
#include "my32bqd.h"

#include "IASScript.h"

EXTERN_API_C( long double )
UInt64ToLongDouble(UInt64 value);

/* constants */

#define MINIMUM_MEMORY_HEAP					(long)(1.8*MEG+.5)

const UInt8		_out_of_memory[] = "\pSorry, not enough memory to continue.";

/* globals */

static EventRecord		currentEvent;

Rect			overhead_box = {35,7,71,509};
RgnHandle		gpRgn;

boolean			gDone = FALSE, gSuspended = FALSE, currently_dragging = FALSE;

long			QuickTime_version = -1, SoundManager_version = -1, FileSystem_attrs = 0;

struct click_profile	click_profile_list[NUMBER_OF_CLICK_PROFILES] =
	{0, {-32768,-32768}, NONE, NONE};

int				overhead_text_length = 0, scrollbar_width = DEFAULT_SCROLLBAR_WIDTH;
char			overhead_text[512];

Str255			currentSoundName;

/* privates */

static void initialize_application_heap(void);
static void CleanUp(void);
static void evaluate_system(void);

static void HandleMouseDown(EventRecord *evt, boolean halted);
static boolean HandleKeyDown(boolean halted);
static void HandleGoAway(WindowRef window);
static void HandleSelectWindow(WindowRef window);

static void Resume(void);
static void Suspend(void);

/* prototypes */

boolean ShowControls(WindowRef window, boolean show);

/* code */

main()
{
	initialize_application_heap();
	cs_initialize(NULL);

#if DEBUG
	initialize_debugger(TRUE);
#endif

	SetCursor(*GetCursor(watchCursor));

	evaluate_system();

	install_apple_event_handlers();

	if (allocate_ias_memory(0)!=_ias_no_error)
		alert_user(fatalError,strERRORS,outOfMemory,-1);

	initialize_screen();
	initialize_menubar();

	SetCursor(&qd.arrow);

	do HandleEvent(NULL,NULL,0,0); while(!gDone);

	DisposeRgn(gpRgn);

	free_ias_memory(TRUE);

	close_debugging_log();
	exit(0);
}

static void initialize_application_heap()
{
	MaxApplZone();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();

	InitGraf((Ptr)&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0L);
	InitCursor();

	GetDateTime((unsigned long*)&qd.randSeed);

	gpRgn = NewRgn();
	assert(gpRgn);
}

static void evaluate_system()
{
	long		result = -1, value;
	OSErr		err;

	if (NGetTrapAddress(_Gestalt,OSTrap)==NGetTrapAddress(_Unimplemented,ToolTrap)
			|| (err=Gestalt(gestaltSystemVersion,&result))!=noErr)
		alert_user(fatalError,strERRORS,systemTooOldError,result);

	if (result<=0x74F)
		alert_user(fatalError,strERRORS,oldSystemVersionError,result);

	err = Gestalt(gestaltNativeCPUtype,&result);
	if (err || result<=gestaltCPU68010)
		alert_user(fatalError,strERRORS,oldProcessorError,result);

	if ((value=FreeMem())<65536)/*MINIMUM_MEMORY_HEAP)*/
		alert_user(fatalError,strERRORS,insufficientMemoryAtLaunch,value>>=10);

	if (!Gestalt(gestaltFSAttr,&result))
		FileSystem_attrs = result;
	if (!Gestalt(gestaltQuickTimeVersion,&result))
		QuickTime_version = result;
	if (!Gestalt(gestaltSoundAttr,&result))
	{
		*(NumVersion*)&result = SndSoundManagerVersion();
		SoundManager_version = result;
	}
}

/* runtime */

boolean HandleEvent(DialogRef dialog, EventRecord *event, boolean halted, boolean external)
{
	WindowRef	whichwind;
	boolean		update = FALSE;

	if (event)
		currentEvent = *event;
	else
		WaitNextEvent(everyEvent,&currentEvent,30,NULL);
	/*SystemTask();*/
	switch(currentEvent.what)
	{
	case mouseDown:
		if (!external)
			HandleMouseDown(&currentEvent,halted);
		break;
	case keyDown:
	case autoKey:
		update = HandleKeyDown(halted);
		break;
	case updateEvt:
		whichwind = (WindowRef)currentEvent.message;
		if ((UInt32)whichwind<=0)
			return;
		if (dialog!=whichwind)
		{
			BeginUpdate(whichwind);
			HandleUpdateWindow(whichwind);
			EndUpdate(whichwind);
		} else
			update = TRUE;
		break;
	case activateEvt:
		FullSelectWindow(GetNextWindow(dock_window),false);
		FullSelectWindow((WindowRef)currentEvent.message,true);
		break;
	case kHighLevelEvent:
		if (AEProcessAppleEvent(&currentEvent))
			;/*FatalErrorAE();	-- we dont want to quit cause of some small thing */
		break;
    	case osEvt:
		InitCursor();
		switch(((UInt32)currentEvent.message)>>24)
		{
		case suspendResumeMessage:
			if (currentEvent.message & resumeFlag) Resume(); else Suspend();
			break;
		}
	}
	if (!halted) global_idle_proc();
end:
	return update;
}

void FullSelectWindow(WindowRef window, boolean front)
{
	if (window==(WindowRef)NULL)
		return;
	HiliteWindow(window,front);
	ShowControls(window,front);
}

boolean ShowControls(WindowRef window, boolean show)
{
	ControlRef	ctl = GetControlListFromWindow(window);
	boolean		ne = !!ctl;

	if (show)
		while(ctl)
		{
			ShowControl(ctl);
			ctl = (*ctl)->nextControl;
		}
	else
		while(ctl)
		{
			HideControl(ctl);
			ctl = (*ctl)->nextControl;
		}

	return ne;
}

static void HandleMouseDown(EventRecord *evt, boolean halted)
{
	Rect		arect;
	Point		ml, newsize, evloc = evt->where;

	WindowRef	whichwind, tempwind;
	long		menuhit, oldwbsize, lnewsize;
	short		origval, clicktype = FindWindow(evloc,&whichwind);
	short		oldheight, wlistindex, maplv = 0;

	*(SInt32*)&arect = 0, *(1+(SInt32*)&arect) = 0x4000400;
	if (!halted)
	{
		currently_dragging = FALSE;
		if (clicktype!=inMenuBar) HandleSelectWindow(whichwind);
	}
	GetPort(&tempwind);

	BlockMove(click_profile_list,click_profile_list+1,sizeof(struct click_profile)*(NUMBER_OF_CLICK_PROFILES-1));
	click_profile_list[0].target_type = NONE, click_profile_list[0].where = evloc;

	switch(clicktype)
	{
	case inMenuBar:
		click_profile_list[0].target_type = kTargetMenubar,
			click_profile_list[0].target_index = 0;
		if (!halted)
		{
			menuhit = MenuSelect(evloc);
			HiliteMenu(0);
			click_profile_list[0].target_index = menuhit>>16;
			ProgramMenuChoice(menuhit);
		};
		break;
	case inDrag:
		click_profile_list[0].target_type = kTargetWindowTitlebar;
		if (!halted)
		{
			currently_dragging = TRUE;
		/* all screens!! */
			DragWindow(whichwind,evloc,&qd.screenBits.bounds);
		}
		break;
	case inGoAway:
		click_profile_list[0].target_type = kTargetWindowClosebox;
		if (!halted && TrackGoAway(whichwind,evloc))
			switch(((WindowPeek)whichwind)->refCon)
			{
			case kDockWindow:
				SysBeep(20);
				break;
			case kDocumentWindow:
				WCloseDoc(whichwind);
				break;
			case kPaletteWindow:
				ShowHideFloating(whichwind);
			}
		break;
	case inZoomIn:
	case inZoomOut:
		click_profile_list[0].target_type = kTargetWindowZoom;
		break;
	case inGrow:
		click_profile_list[0].target_type = kTargetWindowGrow;
		if (!halted)
		{
			assert(((WindowPeek)whichwind)->refCon==kDocumentWindow);
			*(SInt32*)&arect = 0x520070, *(1+(SInt32*)&arect) = 0x1B300F8;
			lnewsize = GrowWindow(whichwind,evloc,&arect);
			if (lnewsize)
			{
				oldheight = arect.bottom;
				SizeWindow(whichwind,(short)lnewsize,lnewsize>>16,false);
				SetRect(&arect,0,0,(short)lnewsize,lnewsize>>16);
				SetPort(whichwind);
				ClipRect(&arect);
				InvalRect(&arect);
			}
		}
		break;
	case inContent:
		click_profile_list[0].target_type = kTargetWindowSpace;
		if (!halted)
		{
			SetPort(whichwind);
			GlobalToLocal(&evloc);
		}
	}
	SetPort(tempwind);
}

static boolean HandleKeyDown(boolean halted)
{
	WindowRef	whichwind;
	long		temp;
	short		itemhit, achr;

	achr = (Byte)currentEvent.message;
	if (currentEvent.modifiers & cmdKey)
	{
		if (achr=='.')
			return -1;
		if (halted)
			goto end;
		if ((currentEvent.modifiers & shiftKey) && ((achr>'`') && (achr<='z')))
			achr |= 1<<8;
		if ((temp=MenuKey(achr))!=0)
		{
			ProgramMenuChoice(temp);
			temp = 0;
			wait(1);
			HiliteMenu(0);
		} else if (IsDialogEvent(&currentEvent) && DialogSelect(&currentEvent,&whichwind,&itemhit))
			;	/* last resort */
	} else if (!halted)
	{
		if (IsDialogEvent(&currentEvent) && DialogSelect(&currentEvent,&whichwind,&itemhit))
		{
			/* handle your floating dialog event */
			goto end;
		}
		switch(achr)
		{	/* toggle all palettes when tab key is pressed */
		case '\t':
			ShowHideAllFloating(0);
			break;
		default:
			;
		}
	}
end:
	return FALSE;
}

static void HandleSelectWindow(WindowRef window)
{
	WindowRef	nearFront;

	switch(((WindowPeek)window)->refCon)
	{
	case kDocumentWindow:
		/*nearFront = GetBottomPalette(1); */

		/*if ((long)nearFront>0) */
		{
			assert(dock_window);
			SendBehind(window,dock_window);
			break;
		}
	default:
		SelectWindow(window);
	}
}

void HandleUpdateWindow(WindowRef window)
{
	WindowRef	curPort;
	short		index, familyID;

	GetPort(&curPort);
	if ((long)window<=0) goto end;
	SetPort(window);
	switch(((WindowPeek)window)->refCon)
	{
	case kDocumentWindow:
		index = WFindDoc(window);
		assert(index>=0);
		UpdateDocWindow(index);
		break;
	case kDockWindow:
		EraseRect(&dock_window->portRect);
		RGBForeColor(cs_system_colors+_system_gray93);
		MoveTo(5,14);
		GetFNum("\pMonaco",&familyID);
		TextMode(srcOr);
		TextFont(familyID);
		TextSize(10);
		TextFace(0);
		if (currentSoundName[0]==0)
			DrawString("\p(no sound playing)");
		else
			DrawString(currentSoundName);
		TextSize(14);
		RGBForeColor(cs_system_colors+_system_gray87);
		TextFace(bold);
		TETextBox(overhead_text,overhead_text_length,&overhead_box,teJustLeft);
		break;
	case kPaletteWindow:
		break;
	}
end:
	SetPort(curPort);
}

void SetOverheadBoxText(const char *text)
{
	GrafPtr		old_port;

	assert(dock_window);
	GetPort(&old_port);
	SetPort(dock_window);

	overhead_text_length = strlen(text);
	memcpy(overhead_text,text,sizeof(overhead_text));
	InvalRect(&overhead_box);
	SetPort(old_port);
}

void SetCurrentSoundName(ConstStr255Param text)
{
	GrafPtr		old_port;

	assert(dock_window);
	GetPort(&old_port);
	SetPort(dock_window);

	if (text==(ConstStr255Param)NULL || text[0]==0)
		currentSoundName[0] = 0;
	else
	{
		psprintf(temporary,"Playing: %P", text);
		memcpy(currentSoundName,temporary,1+((UInt8*)temporary)[0]);
	}
	InvalRect(&dock_window->portRect);
	SetPort(old_port);
}

void handle_error(short err)
{
	char			*t;

	switch(err)
	{
	case -128: t = "User aborted."; break;
	case -108: t = "Out of memory."; break;
	case noErr: t = (char*)empty; break;
	case 1: t = "An error occured while encoding."; break;
	case 2: t = "Unsupported sound header."; break;
	case 3: t = "No sampled sound in 'snd '."; break;
	case 4: t = csprintf(temporary,"Resource mgr gave error #%d", ResError()); break;
	case 5: t = csprintf(temporary,"Could not get 'snd ' resource (#%d)", ResError()); break;
	case 6: t = "No sound commands in `snd '."; break;
	case 7: t = "Bad 'snd ' format."; break;
	case 8: t = "Data fork is empty."; break;
	case 9: t = "File is empty."; break;

	default: t = csprintf(temporary,"System error #%d", err);
	}
	SetOverheadBoxText(t);
}

/* interface */

void Resume()
{
	WindowRef	window;

	gSuspended = FALSE;
	if ((window=FrontWindow())!=NULL)
		;
	ShowHideAllFloating(kSHSuspendResume|kSHResume);
end:
	;
}

void Suspend()
{
	WindowRef	window;

	ShowHideAllFloating(kSHSuspendResume);
	gSuspended = TRUE;
	if (window=FrontWindow())
		;
end:
	;
}

void global_idle_proc()
{
}

void wait(SInt32 delay_ticks)
{
	UInt64		last_ms, ms;

	machine_milliseconds(last_ms);
	if (delay_ticks<0)
	{
		delay_ticks = -delay_ticks;

		HandleEvent(NULL,NULL,1,0);
		HandleEvent(NULL,NULL,1,0);
		HandleEvent(NULL,NULL,1,0);
	}
	Add64U32(last_ms,delay_ticks);

	while(U64Gt(last_ms,ms))
	{
		machine_milliseconds(ms);
		if (GetNextEvent(everyEvent,&currentEvent))
			HandleEvent(NULL,&currentEvent,1,0);
	}
}

/* menubar */

void CheckOnly(short m, short f, short l, short i, boolean c)
{
	MenuRef		mh = GetMenuHandle(m);
	short		x = 0;

	if (mh)
	{
		if (i<0)
			i = x;
		else
			x = CountMItems(mh);
		if (f<=1)
			if (l<0)
			{
				++x;
				if (c)
					while(--x)
						CheckItem(mh,x,x!=i);
				else
					while(--x)
						CheckItem(mh,x,x==i);
			} else
			{
				if (l>0 && l<=x)
					x = l;
				++x;
				if (c)
					while(--x)
						CheckItem(mh,x,x!=i);
				else
					while(--x)
						CheckItem(mh,x,x==i);
			}
		else if (f>1)
		{
			if (l>0 && l<=x) x = l;
			++x;
			if (c)
				while(--x>=f)
					CheckItem(mh,x,x!=i);
			else
				while(--x>=f)
					CheckItem(mh,x,x==i);
		}
	}
}

void initialize_menubar()
{
	ClearMenuBar();
	AppendResMenu(add_menu_resource(MBAppleID,0),'DRVR');
	add_menu_resource(MBFileID,0);
	add_menu_resource(MBEditID,0);

	ProgramMenuInit();

	DrawMenuBar();
}

MenuRef add_menu_resource(short id, short loc)
{
	MenuRef		mh;

	if (!(mh=GetMenu(id)))
	{
		if (!(id=ResError()))
			id = MemError();
		alert_user(fatalError,strERRORS,unexpectedError,id);
	}
	InsertMenu(mh,loc);
	return mh;
}

/* memory */

void *tmalloc(size_t size)
{
	void		*store = realloc(NULL,size);

	/* if (store==NULL) ... <dispose something and try again> */
	return store;
}
