/*
 mixbit_ssnd.c
  source file for "mixbit" sampled sound encoding

  Version 0.0.0.8

  Created: May 6, 2006
  Revised: June 1, 2006

 Copyright (C) 2006 by Adam Pickett

    This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.

    You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Friday, May 12, 2006 9:30 AM EDST - 11:59PM
A.P.: SUCCESS!!  I've gotten my "Kiss My Ass" ADPCM format to work!
  Implemented Apple Events and the only left to do is the (so-called)
	"mixbit standard" format
Saturday, May 13, 2006 12:18 AM EDST
  That now works.
Tuesday, May 16, 2006 7:22 AM EDST
  The KMT format has been neutralized to the IMA standard.  The XADPCM
    format will be on it's own.
Wednesday, May 17, 2006 2:35 AM EDST
  In-position clipping in the ADPCM formats should have been in place,
    rather than in the out-sample.
*/

#include <cseries.h>

#include <string.h>

#include "mixbit.h"
#include "mixbit_ssnd.h"

/* constants */

#define XADPCM_MAXIMUM_FALL_OFF			8

/* structures */

typedef struct ADPCM_data
{
	SInt32			current, previous, sample, boundary, leak;
	SInt32			adjustment[8];
	SInt32			step[];
} ADPCM_data;

/* globals */

static SInt16		_1_u7_dectb[256] =
{
       0,      1,      2,      3,      4,      5,      6,      7,
       8,      9,     10,     11,     12,     14,     16,     18,
      20,     22,     24,     26,     28,     30,     33,     36,
      39,     42,     45,     48,     51,     54,     57,     60,

      64,     68,     72,     76,     80,     84,     88,     92,
      96,    100,    105,    110,    115,    120,    125,    130,
     135,    140,    145,    150,    156,    162,    168,    174,
     180,    186,    192,    198,    204,    210,    217,    225,

     234,    244,    256,    384,    512,    640,    768,    896,
    8<<7,   9<<7,  10<<7,  11<<7,  12<<7,  14<<7,  16<<7,  18<<7,
   20<<7,  22<<7,  24<<7,  26<<7,  28<<7,  30<<7,  33<<7,  36<<7,
   39<<7,  42<<7,  45<<7,  48<<7,  51<<7,  54<<7,  57<<7,  60<<7,

   64<<7,  68<<7,  72<<7,  76<<7,  80<<7,  84<<7,  88<<7,  92<<7,
   96<<7, 100<<7, 105<<7, 110<<7, 115<<7, 120<<7, 125<<7, 130<<7,
  135<<7, 140<<7, 145<<7, 150<<7, 156<<7, 162<<7, 168<<7, 174<<7,
  180<<7, 186<<7, 192<<7, 198<<7, 204<<7, 210<<7, 217<<7, 225<<7,

  -32768,     -1,     -2,     -3,     -4,     -5,     -6,     -7,
      -8,     -9,    -10,    -11,    -12,    -14,    -16,    -18,
     -20,    -22,    -24,    -26,    -28,    -30,    -33,    -36,
     -39,    -42,    -45,    -48,    -51,    -54,    -57,    -60,

     -64,    -68,    -72,    -76,    -80,    -84,    -88,    -92,
     -96,   -100,   -105,   -110,   -115,   -120,   -125,   -130,
    -135,   -140,   -145,   -150,   -156,   -162,   -168,   -174,
    -180,   -186,   -192,   -198,   -204,   -210,   -217,   -225,

    -234,   -244,   -256,   -384,   -512,   -640,   -768,   -896,
   -8<<7,  -9<<7, -10<<7, -11<<7, -12<<7, -14<<7, -16<<7, -18<<7,
  -20<<7, -22<<7, -24<<7, -26<<7, -28<<7, -30<<7, -33<<7, -36<<7,
  -39<<7, -42<<7, -45<<7, -48<<7, -51<<7, -54<<7, -57<<7, -60<<7,

  -64<<7, -68<<7, -72<<7, -76<<7, -80<<7, -84<<7, -88<<7, -92<<7,
  -96<<7,-100<<7,-105<<7,-110<<7,-115<<7,-120<<7,-125<<7,-130<<7,
 -135<<7,-140<<7,-145<<7,-150<<7,-156<<7,-162<<7,-168<<7,-174<<7,
 -180<<7,-186<<7,-192<<7,-198<<7,-204<<7,-210<<7,-217<<7,-225<<7,
};

static ADPCM_data	_ima_adpcm[] =
{
	0,  0,  0,  88, 0,

	{  -1, -1, -1, -1,  2,  4,  6,  8},

	{ /* 6.78555, factor 1.1, not rounded; first entries {7, 8, 9, 10}, last {32767} */
	         7,     8,      9,     10,     11,     12,     13,     14,
	        16,    17,     19,     21,     23,     25,     28,     31,
	        34,    37,     41,     45,     50,     55,     60,     66,
	        73,    80,     88,     97,    107,    118,    130,    143,
	       157,   173,    190,    209,    230,    253,    279,    307,
	       337,   371,    408,    449,    494,    544,    598,    658,
	       724,   796,    876,    963,   1060,   1166,   1282,   1411,
	      1552,  1707,   1878,   2066,   2272,   2499,   2749,   3024,
	      3327,  3660,   4026,   4428,   4871,   5358,   5894,   6484,
	      7132,  7845,   8630,   9493,  10442,  11487,  12635,  13899,
	     15289, 16818,  18500,  20350,  22385,  24623,  27086,  29794,
	     32767
	},
};

static ADPCM_data	_kmt_adpcm[] =
{
	0,  0,  0, 63,  0,

	{  -2, -1, -1, -1,  0,  2,  4,  6},

	{ /* 9, factor gr**(1/4), not rounded; first entries {7, 15} */
            7,    15,      20,     23,     26,     29,     33,     37,
           42,    47,      54,     60,     68,     77,     87,     98,
          111,   125,     141,    159,    180,    203,    229,    258,
          291,   328,     370,    418,    471,    531,    599,    676,
          762,   860,     970,   1094,   1234,   1392,   1570,   1771,
         1997,   2252,   2540,   2865,   3232,   3645,   4111,   4636,
         5229,   5898,   6652,   7502,   8461,   9543,  10763,  12139,
        13691,  15441,  17415,  19641,  22152,  24984,  28178,  31780,
	},
};

static UInt8		_1_u7_enctl[250] =
{
    0,  1,  2,  3,  4,  5,  6,  7,
    8,  9, 10, 11, 12, 13, 14, 14,
   14, 15, 15, 16, 16, 17, 17, 18,
   18, 19, 19, 20, 20, 21, 21, 21,

   22, 22, 22, 23, 23, 23, 24, 24,
   24, 25, 25, 25, 26, 26, 26, 27,
   27, 27, 28, 28, 28, 29, 29, 29,
   30, 30, 30, 31, 31, 31, 32, 32,

   32, 32, 33, 33, 33, 33, 34, 34,
   34, 34, 35, 35, 35, 35, 36, 36,
   36, 36, 37, 37, 37, 37, 38, 38,
   38, 38, 39, 39, 39, 39, 40, 40,

   40, 40, 41, 41, 41, 41, 41, 42,
   42, 42, 42, 42, 43, 43, 43, 43,
   43, 44, 44, 44, 44, 44, 45, 45,
   45, 45, 45, 46, 46, 46, 46, 46,

   47, 47, 47, 47, 47, 48, 48, 48,
   48, 48, 49, 49, 49, 49, 49, 50,
   50, 50, 50, 50, 51, 51, 51, 51,
   51, 52, 52, 52, 52, 52, 52, 53,

   53, 53, 53, 53, 53, 54, 54, 54,
   54, 54, 54, 55, 55, 55, 55, 55,
   55, 56, 56, 56, 56, 56, 56, 57,
   57, 57, 57, 57, 57, 58, 58, 58,

   58, 58, 58, 59, 59, 59, 59, 59,
   59, 60, 60, 60, 60, 60, 60, 61,
   61, 61, 61, 61, 61, 61, 62, 62,
   62, 62, 62, 62, 62, 62, 63, 63,

   63, 63, 63, 63, 63, 63, 64, 64,
   64, 64, 64, 64, 64, 64, 64, 65,
   65, 65, 65, 65, 65, 65, 65, 65,
   65, 65,
};

/* macros */

#if TARGET_RT_LITTLE_ENDIAN
# define addS16(d,s,m)					do{d+=((*s++<<8)+*s++)<<(m);}while(0)
#else
# define addS16(d,s,m)					do{d+=(*(SInt16*)s)<<(m); s+=2;}while(0)
#endif
#if TARGET_RT_LITTLE_ENDIAN
# define readU24(d,s)					do{d=(*s++<<16)|(*s++<<8)|*s++;}while(0)
#else
# define readU24(d,s)					do{d=(*(UInt32*)(-1+s))&0x00FFFFFF; s+=3;}while(0)
#endif

/* private prototypes */

static SInt32 decode_adpcm4_sample(ADPCM_data *h, UInt8 v);
static UInt8 encode_adpcm4_sample(ADPCM_data *h, int q);
static UInt8 encode_adpcm_adjustment(ADPCM_data *h, int q);

static const UInt8 *decode_xadpcm5(ADPCM_data *h, const UInt8 *s, SInt16 *d, int c, int o);
static SInt32 decode_adpcm5_sample(ADPCM_data *h, UInt8 v);
static UInt8 encode_xadpcm5_sample(ADPCM_data *h, int q1, int q2);

static UInt8 *pack_xadpcm5(UInt8 *d, const UInt8 *s, int c);

/* code */

const UInt8 *mxb_decode_ssnd16(const UInt8 *_s,
	SInt16 **_d,
	SInt16 *e,
	int c,
	short _o,
	boolean *ovfl)
{
	ADPCM_data		*h;

	const UInt8		*s;
	SInt16			*d;
	int				t, o = _o<<1;
	UInt8			v, m, b = 0;
	boolean			dummy;

	if ((s=_s)==(const UInt8*)NULL || _d==(SInt16**)NULL || (d=*_d)==(SInt16*)NULL
			|| e==(SInt16*)NULL)
		return NULL;
	if (ovfl==(boolean*)NULL)
		ovfl = &dummy;
	*ovfl = 0;

	/* the end must be precalculated in case we bail due to ending boundary */
	v = *s++; m = (v>>6) & 1; _s = s+c;
	if (v & 128)
		goto variable;
	switch(v & 191)
	{ /* fixed size formats */
	case _mixbit_format_U8_PCM:
		do
		{
			if (d>=e)
				{*ovfl = 1; break;}
			*d++ = ((*s++<<7)-16384)<<m; d = (SInt16*)(o+(byte*)d);
		} while(--c);
		break;
	case _mixbit_format_S16_PCM:
		_s = s+(c<<1);
		do
		{
			if (d>=e)
				{*ovfl = 1; break;}
			addS16(*d++,s,m); d = (SInt16*)(o+(byte*)d);
		} while(--c);
		break;
	case _mixbit_format_1_U7_DPCM:
		do
		{
			if (d>=e)
				{*ovfl = 1; break;}
			if ((t+=_1_u7_dectb[*s++]<<m)<-32768)
				*d++ += -32768;
			else if (t>32767)
				*d++ += 32767;
			else
				*d++ += t;
			d = (SInt16*)(o+(byte*)d);
		} while(--c);
		break;
	case _mixbit_format_IMA_ADPCM:
//		memset(h,0,12);
		h = _ima_adpcm;
		if (d>=e)
			{*ovfl = 1; break;}
		t = *s++; *d++ = h->sample = t = ((SInt16)((t<<8)+*s++))<<m;
		/* get the initial adjustment and bail if it is out of range */
		if ((h->current=*s++)>h->boundary)
			return NULL;
		if (*s++!=0)
			return NULL;
		v = c; _s = s+((c+1)>>1);
		if ((c>>=1)!=0)
			do
			{
				if (d>=e)
					{*ovfl = 1; break;}
				if ((t+=decode_adpcm4_sample(h,*s>>4)<<m)<-32768)
					t = -32768;
				else if (t>32767)
					t = 32767;
				*d++ += t; d = (SInt16*)(o+(byte*)d);
				if (d>=e)
					{*ovfl = 1; break;}
				if ((t+=decode_adpcm4_sample(h,*s++ & 0xF)<<m)<-32768)
					t = -32768;
				else if (t>32767)
					t = 32767;
				*d++ += t; d = (SInt16*)(o+(byte*)d);
			} while(--c);
		if (v & 1)
			if (d>=e)
				*ovfl = 1;
			else if ((t+=decode_adpcm4_sample(h,*s++>>4)<<m)<-32768)
				t = -32768;
			else if (t>32767)
				t = 32767;
			*d++ += t;
		break;

	default:
		return NULL;
	}
	goto update;
variable:
	readU24(t,s); _s = s+t; t = 0;
	switch(v & 191)
	{
	case _mixbit_format_standard:
		do
		{
			if (d>=e)
				{*ovfl = 1; break;}
			if ((SInt8)(v=*s++)>=0)
				if ((t+=(v-64)<<m)<-32768)
					*d++ += -32768;
				else if (t>32767)
					*d++ += 32767;
				else
					*d++ += t;
			else
				*d++ += t=(((v<<8)+*s++)-49152)<<m;
			d = (SInt16*)(o+(byte*)d);
		} while(--c);
		break;
	case _mixbit_format_KMT_XADPCM:
		h = _kmt_adpcm;
//		if ((h->current=*s++)>h->boundary)
//			return NULL;
		t = *s++; h->sample = t = (SInt16)((t<<8)+(*s & 0xC0));
		h->current = *s++ & 0x3F;
		d = decode_xadpcm5(h,s,d,c,o);
		break;
	default:
		return NULL;
	}
update:
	*_d = d;
	return _s;
}

static const UInt8 *decode_xadpcm5(ADPCM_data *h,
	const UInt8 *s,
	SInt16 *d,
	int _c,
	int o)
{
	const UInt8		*r = s, *g; /* repeat bits, sample nybbles, sign bits */
	int				c = _c, i = 0;
	UInt8			br, bs, bg, gv, sv, pv;

	/* scan to the sign and sample bits */
	pv = gv = *s++, bg = 0x80, i = 0;
	do
	{
		++i;
		if ((bg>>=1)==0)
			bg = 0x80, gv = *s++;
		if ((gv & bg)==0 && --c==0)
			break;
		pv = gv;
	} while(--c);

	bs = (bs & 0xF0) ? 0 : 4;
	if (bg & 0x80)
		g = s+(i>>1); /* no more packed bits before the samples */
	do
	{
		v = (*s>>bg);
		if ((bg^=4)!=0)
			v = *s++;
		if (*r & bs)
			v |= 0x80;
		if (d>=e)
			{*ovfl = 1; break;}
		if ((t+=decode_adpcm5_sample(h,v)<<m)<-32768)
			t = -32768;
		else if (t>32767)
			t = 32767;
		*d++ += t; d = (SInt16*)(o+(byte*)d);

		if (*r & br)
		{
			if (d>=e)
				{*ovfl = 1; break;}
			if ((t+=decode_adpcm5_sample(h,v)<<m)<-32768)
				t = -32768;
			else if (t>32767)
				t = 32767;
			*d++ += t; d = (SInt16*)(o+(byte*)d);
		}
		if ((br>>=1)==0)
			{++r; br = 0x80;}
		if ((bs>>=1)==0)
			{++s; bs = 0x80;}
	} while(--c);

	return d;
}

UInt8 *mxb_encode_samples16(const SInt16 *_s,
	UInt8 *_d,
	int _c,
	int v,
	short _o)
{
	ADPCM_data		*h;

	const SInt16	*s = _s;
	int				c = _c, t, u = 16384, o = _o<<1;
	int				p1 = INT_MIN, p2 = INT_MIN;
	UInt8			*d = _d, m, b;

	if (c<=0)
		return NULL;
	m = (v>>6) & 1;
	switch(v & 447)
	{
	case _mixbit_format_U8_PCM | 256:
		do
		{
			t = 16384+(*s++>>m); s = (SInt16*)(o+(byte*)s);
			if ((t>>=7)<0)
				if (p2<=0)
					*d++ = 0;
				else
					*d++ = t=-t;
			else if (t>255)
				if (p2<=0)
					*d++ = 0xFF;
				else
					*d++ = t=~t;
			else
			{
				if (p2>0)
					*d++ = ((signed)d[-1])>>7;
				else
					*d++ = t;
				p2 = 0;
				continue;
			}
			++p2;
		} while(--c);
		break;
	case _mixbit_format_U8_PCM:
		do
		{
			t = 16384+(*s++>>m); s = (SInt16*)(o+(byte*)s);
			if ((t>>=7)<0)
				*d++ = 0;
			else if (t>255)
				*d++ = 0xFF;
			else
				*d++ = t;
		} while(--c);
		break;
	case _mixbit_format_S16_PCM | 256:
	case _mixbit_format_S16_PCM:
#if TARGET_RT_LITTLE_ENDIAN
		do
		{
			t = *s++>>m; s = (SInt16*)(o+(byte*)s);
			*d++ = (t>>8) & 0xFF; *d++ = t & 0xFF;
		} while(--c);
#else
		do
		{
			t = *s++>>m; s = (SInt16*)(o+(byte*)s);
			*(SInt16*)d = t; d += 2;
		} while(--c);
#endif
		break;
	case _mixbit_format_1_U7_DPCM | 256:
	case _mixbit_format_1_U7_DPCM:
		do
		{
			t = 16384+(*s++>>m); s = (SInt16*)(o+(byte*)s);
			p1 = t-u; u = t;
			if ((t=ABS(p1))<=249)
				t = ((((unsigned)p1)>>8) & 0x80) | _1_u7_enctl[t];
			else if (t<=29699)
				t = ((((unsigned)p1)>>8) & 0x80) | _1_u7_enctl[(t+64)>>7]+0x40;
			else if (t<=30783)
				t = ((((unsigned)p1)>>8) & 0x80) | 63;
			else
				t = 128;
			*d++ = t;
			u += p1-_1_u7_dectb[t]; /* error correction */
		} while(--c);
		break;
	case _mixbit_format_standard | 256:
		do
		{
			t = 16384+(*s++>>m); s = (SInt16*)(o+(byte*)s);
			p1 = t-u; u = t;
			if ((p1+=64)>=0 && p1<=127)
				{*d++ = p1; continue;}
			if (t<0)
				if (p2++==0)
					t = 0;
				else
					t = -t;
			else if (t>=32767)
				if (p2++==0)
					t = 32767;
				else
					{t = -t; t += 32768;}
			*d++ = (t>>8) | 128; *d++ = t; p2 = 0;
		} while(--c);
		break;
	case _mixbit_format_standard:
		do
		{
			t = 16384+(*s++>>m); s = (SInt16*)(o+(byte*)s);
			p1 = t-u; u = t;
			if ((p1+=64)>=0 && p1<=127)
				*d++ = p1;
			else if (t<0)
				{*d++ = 0x80; *d++ = 0;}
			else if (t>=32767)
				{*d++ = 0xFF; *d++ = 0xFF;}
			else
				{*d++ = (t>>8) | 128; *d++ = t;}
		} while(--c);
		break;
	case _mixbit_format_IMA_ADPCM | 256:
	case _mixbit_format_IMA_ADPCM:
		h = _ima_adpcm;
		memset(h,0,12);
		p1 = c;
		h->sample = t = *s++>>m; s = (SInt16*)(o+(byte*)s);
		*d++ = t>>8; *d++ = t;
		if ((c>>=1)!=0)
		{
			*d++ = h->current = encode_adpcm_adjustment(h,(*s>>m)-h->sample);
			*d++ = 0;
			do
			{
				v = encode_adpcm4_sample(h,*s++>>m)<<4; s = (SInt16*)(o+(byte*)s);
				*d++ = v | encode_adpcm4_sample(h,*s++>>m); s = (SInt16*)(o+(byte*)s);
			} while(--c);
		} else
			{*d++ = h->current = 1; *d++ = 0;}
		if (p1 & 1)
			{*d++ = encode_adpcm4_sample(h,*s>>m)<<4;}
		break;
	case _mixbit_format_KMT_XADPCM | 256:
	case _mixbit_format_KMT_XADPCM:
		h = _kmt_adpcm;
		memset(h,0,12);

		t = (*s>>m) & ~64;
		*d++ = t>>8;
		b = encode_adpcm_adjustment(h,t-(*s>>m))+2;
		*d++ = MIN(b,63) | t;
		_d = d; u = INT_MAX, b = 0, h->leak = 8;
		do
		{
			h->sample = (_d[-2]<<8) | _d[-1] & ~64, h->current = _d[-1] & 63;
			do
			{
				t = *s++>>m; s = (SInt16*)(o+(byte*)s);
				t = encode_xadpcm5_sample(h,t,c>0 ? *s>>m : t);
				if (p2>=0)
					if (p2==p1)
						{*d++ = p2 | 0x10; p1 = -1;}
					else
						*d++ = p2;
				p2 = p1; p1 = t;
			} while(--c);
			if (p2>=0)
			{
				*d++ = p2;
				if (p1>=0)
					if (p1==p2)
						d[-1] |= 0x20;
					else
						*d++ = p1;
			}
			if (++h->leak>12)
				break;
			t = (long)d-(long)_d;
			if (t<u)
				u = t, b ^= 1;

			s = _s; d = _d+b*(2+(c=_c));
			memset(h,0,12);
			p1 = p2 = INT_MIN;
		} while(1);
		memmove(_d+2+_c,_d+b*(2+_c),u);
		d = pack_xadpcm5(_d,_d+2+_c,u);
		break;

	default:
		return NULL;
	}

	return d;
}

UInt8 *mxb_encode_samples8u(const UInt8 *_s,
	UInt8 *_d,
	int _c,
	int v,
	short _o)
{
	ADPCM_data		*h;

	const UInt8		*s = _s;
	int				c = _c, t, u = 16384, o = _o+1;
	int				p1 = INT_MIN, p2 = INT_MIN;
	UInt8			*d = _d, m, b = 0;

	if (c<=0)
		return NULL;
	m = (v>>6) & 1;
	switch(v & 447)
	{
	case _mixbit_format_U8_PCM | 256:
		do
		{
			t = *s; t = 16384+(((t<<8)+t-32768)>>m); s += o;
			if ((t>>=7)<0)
				if (p2<=0)
					*d++ = 0;
				else
					*d++ = t=-t;
			else if (t>255)
				if (p2<=0)
					*d++ = 0xFF;
				else
					*d++ = t=~t;
			else
			{
				if (p2>0)
					*d++ = ((signed)d[-1])>>7;
				else
					*d++ = t;
				p2 = 0;
				continue;
			}
			++p2;
		} while(--c);
		break;
	case _mixbit_format_U8_PCM:
		do
		{
			t = *s; t = 16384+(((t<<8)+t-32768)>>m); s += o;
			if ((t>>=7)<0)
				*d++ = 0;
			else if (t>255)
				*d++ = 0xFF;
			else
				*d++ = t;
		} while(--c);
		break;
	case _mixbit_format_S16_PCM | 256:
	case _mixbit_format_S16_PCM:
		do
		{
			t = *s; t = ((t<<8)+t-32768)>>m; s += o;
#if TARGET_RT_LITTLE_ENDIAN
			*d++ = (t>>8) & 0xFF; *d++ = t & 0xFF;
#else
			*(SInt16*)d = t; d += 2;
#endif
		} while(--c);
		break;
	case _mixbit_format_1_U7_DPCM | 256:
	case _mixbit_format_1_U7_DPCM:
		do
		{
			t = *s; t = 16384+(((t<<8)+t-32768)>>m); s += o;
			p1 = t-u; u = t;
			if ((t=ABS(p1))<=249)
				t = ((((unsigned)p1)>>8) & 0x80) | _1_u7_enctl[t];
			else if (t<=29699)
				t = ((((unsigned)p1)>>8) & 0x80) | _1_u7_enctl[(t+64)>>7]+0x40;
			else if (t<=30783)
				t = ((((unsigned)p1)>>8) & 0x80) | 63;
			else
				t = 128;
			*d++ = t;
			u += _1_u7_dectb[t]-p1; /* error correction */
		} while(--c);
		break;
	case _mixbit_format_standard | 256:
		do
		{
			t = *s; t = 16384+(((t<<8)+t-32768)>>m); s += o;
			p1 = t-u; u = t;
			if ((p1+=64)>=0 && p1<=127)
				{*d++ = p1; continue;}
			if (t<0)
				if (p2++==0)
					t = 0;
				else
					t = -t;
			else if (t>=32767)
				if (p2++==0)
					t = 32767;
				else
					{t = -t; t += 32768;}
			*d++ = (t>>8) | 128; *d++ = t; p2 = 0;
		} while(--c);
		break;
	case _mixbit_format_standard:
		do
		{
			t = *s; t = 16384+(((t<<8)+t-32768)>>m); s += o;
			p1 = t-u; u = t;
			if ((p1+=64)>=0 && p1<=127)
				*d++ = p1;
			else if (t<0)
				{*d++ = 0x80; *d++ = 0;}
			else if (t>=32767)
				{*d++ = 0xFF; *d++ = 0xFF;}
			else
				{*d++ = (t>>8) | 128; *d++ = t;}
		} while(--c);
		break;
	case _mixbit_format_IMA_ADPCM | 256:
	case _mixbit_format_IMA_ADPCM:
		h = _ima_adpcm;
		memset(h,0,12);
		p1 = c; t = *s; h->sample = t = ((t<<8)+t-32768)>>m;
		*d++ = t>>m+8; *d++ = t>>m;
		if ((c>>=1)!=0)
		{
			*d++ = h->current = encode_adpcm_adjustment(h,((*s*257-32768)>>m)-h->sample);
			*d++ = 0;
			do
			{
				t = *s; s += o; v = encode_adpcm4_sample(h,t=((t<<8)+t-32768)>>m)<<4;
				t = *s; s += o; *d++ = v | encode_adpcm4_sample(h,t=((t<<8)+t-32768)>>m);
			} while(--c);
		} else
			{*d++ = h->current = 1; *d++ = 0;}
		if (p1 & 1)
			{t = *s; *d++ = encode_adpcm4_sample(h,t=((t<<8)+t-32768)>>m)<<4;}
		break;
	case _mixbit_format_KMT_XADPCM | 256:
	case _mixbit_format_KMT_XADPCM:
		h = _kmt_adpcm;
		memset(h,0,12);

		t = ((*s*257-32768)>>m) & ~64;
		*d++ = t>>8;
		b = encode_adpcm_adjustment(h,t-(*s>>m))+2;
		*d++ = MIN(b,63) | t;
		_d = d; u = INT_MAX, b = 0, h->leak = 8;
		do
		{
			h->sample = (_d[-2]<<8) | _d[-1] & ~64, h->current = _d[-1] & 63;
			do
			{
				t = *s; s += o; t = ((t<<8)+t-32768)>>m;
				t = encode_xadpcm5_sample(h,t,c>0 ? (*s*257-32768)>>m : t);
				if (p2>=0)
					if (p2==p1)
						{*d++ = p2 | 0x10; p1 = -1;}
					else
						*d++ = p2;
				p2 = p1; p1 = t;
			} while(--c);
			if (p2>=0)
			{
				*d++ = p2;
				if (p1>=0)
					if (p1==p2)
						d[-1] |= 0x20;
					else
						*d++ = p1;
			}
			if (++h->leak>12)
				break;
			t = (long)d-(long)_d;
			if (t<u)
				u = t, b ^= 1;

			s = _s; d = _d+b*(2+(c=_c));
			memset(h,0,12);
			p1 = p2 = INT_MIN;
		} while(1);
		memmove(_d+2+_c,_d+b*(2+_c),u);
		d = pack_xadpcm5(_d,_d+2+_c,u);
		break;

	default:
		return NULL;
	}

	return d;
}

ufixed get_ssnd_maximum_size(int v)
{
	static ufixed	fixed_sizes[] = {65536, 131072, 65536, 32768};
	static ufixed	variable_sizes[] = {131072, 0, 131072};

	if ((v&=191)>=128)
	{
		if ((unsigned)(v-=128)>sizeof(variable_sizes)/4)
			return 0;
		else
			return variable_sizes[v];
	} else if (v>sizeof(fixed_sizes)/4)
		return 0;
	else
		return fixed_sizes[v];
}

static SInt32 decode_adpcm4_sample(ADPCM_data *h, UInt8 v)
{
	int				p = h->step[h->current];
	int				t = p>>3;

	if ((h->current+=h->adjustment[v & 7])<0)
		h->current = 0;
	else if (h->current>h->boundary)
		h->current = h->boundary;

	if (v & 1)
		t += p>>2;
	if (v & 2)
		t += p>>1;
	if (v & 4)
		t += p;
	if (v & 8)
		t = -t;

	return t;
}

static UInt8 encode_adpcm4_sample(ADPCM_data *h, int q)
{
	int				p = h->step[h->previous=h->current];
	int				t = p>>3;
	UInt8			v;

	if ((q-=h->sample)<0)
		v = 8, q = -q;
	else
		v = 0;
	if (q>p)
		v |= 4, t += p, q -= p;
	if (q>(p>>=1))
		v |= 2, t += p, q -= p;
	if (q>(p>>=1))
		v |= 1, t += p, q -= p;

	if (v & 8)
		t = h->sample-t;
	else
		t += h->sample;

	if ((h->sample=t)<-32768)
		h->sample = -32768;
	else if (t>32767)
		h->sample = 32767;

	if ((h->current+=h->adjustment[v & 7])<0)
		h->current = 0;
	else if (h->current>h->boundary)
		h->current = h->boundary;

	return v;
}

static SInt32 decode_adpcm5_sample(ADPCM_data *h, UInt8 v)
{
	int				p = h->step[h->current];
	int				t = p>>4;

	if ((h->current+=h->adjustment[v & 7]+(v & 8))<0)
		h->current = 0;
	else if (h->current>h->boundary)
		h->current = h->boundary;

	if (v & 1)
		t += p>>3;
	if (v & 2)
		t += p>>2;
	if (v & 4)
		t += p>>1;
	if (v & 8)
		t += p;
	if (v & 0x80)
		t = -t;

	return t;
}

static UInt8 encode_xadpcm5_sample(ADPCM_data *h, int q1, int q2)
{
	int				p = h->step[h->previous=h->current];
	int				t = p>>4;
	UInt8			v;

	q2 -= q1; q1 -= h->sample;
	if ((h->current & 1) && ABS(q1)==1)
		q1 = 0;
	if (q1<0)
		v = 0x80, q1 = -q1, q2 = -q2;
	else
		v = 0;
	if (q2>0)
	{
		if (h->current<=XADPCM_MAXIMUM_FALL_OFF)
		{
			q1 += q2>>h->leak-2;
			v |= 0x41, t += p>>3, q1 -= p>>3;
			if (q1>=p)
				v |= 8, t += p, q1 -= p;
			if (q1>=(p>>=1))
				v |= 4, t += p, q1 -= p;
			if (q1>=(p>>=1))
				v |= 2, t += p, q1 -= p;
		} else
		{
			q1 += q2>>h->leak;
			if (q1>=p)
				v |= 8, t += p, q1 -= p;
			if (q1>=(p>>=1))
				v |= 4, t += p, q1 -= p;
			if (q1>=(p>>=1))
				v |= 2, t += p, q1 -= p;
			if (q1>=(p>>=1))
				v |= 1, t += p, q1 -= p;
		}
	} else
	{
		q1 += q2>>h->leak+2;
		if (h->current<=XADPCM_MAXIMUM_FALL_OFF)
		{
			v |= 0x41, t += p>>3, q1 -= p>>3;
			if (q1>p)
				v |= 8, t += p, q1 -= p;
			if (q1>(p>>=1))
				v |= 4, t += p, q1 -= p;
			if (q1>(p>>=1))
				v |= 2, t += p, q1 -= p;
		} else
		{
			if (q1>p)
				v |= 8, t += p, q1 -= p;
			if (q1>(p>>=1))
				v |= 4, t += p, q1 -= p;
			if (q1>(p>>=1))
				v |= 2, t += p, q1 -= p;
			if (q1>(p>>=1))
				v |= 1, t += p, q1 -= p;
		}
	}
	if (v & 0x80)
		t = h->sample-t;
	else
		t += h->sample;

	if ((h->sample=t)<-32768)
		h->sample = -32768;
	else if (t>32767)
		h->sample = 32767;

	if ((h->current+=h->adjustment[v & 7]+(v & 8))<0)
		h->current = 0;
	else if (h->current>h->boundary)
		h->current = h->boundary;

	return v;
}

static UInt8 encode_adpcm_adjustment(ADPCM_data *h, int q)
{
//	int				p = h->step[h->current];
	int				i = 0, n = h->boundary>>1;

	/* calculate a reasonable average for the next two samples
		and the next step magnitude */
//	p = (p+ABS(q1-h->sample)*5+ABS(q2-q1)*2)>>3;

	/* find the closest one underneath it */
	if (q>=h->step[n])
		i = n;
	if (n>0)
		do
		{
			if (h->step[i]>q)
				{--i; return MAX(i,0);}
			++i;
		} while(--n);

	return MIN(i,h->boundary);
}

static UInt8 *pack_xadpcm5(UInt8 *d, const UInt8 *_s, int _c)
{
	const UInt8		*s = _s;
	int				c, v;
	UInt8			b;

	/* repeat bits, packed samples (4 bit), remaining sign bits */
	c = _c, b = 0x80, v = *s++;
	do
	{
		if (v & b)
		
		if (b)
	} while(--c);
}
