/* A.P.: copied from the Old Durandal source 21-12-2004 7:38 */

#include "macintosh_cseries.h"

#include "interface.h"
#include "my32bqd.h"

/* constants */

static const char dgwqe[] = "DisposeGWorld: QDError() == %d";

/* macros */

/* LMSetMBarHeight is not supported under Carbon */
#define MenuBarHeight		*(short*)0xBAA
#define GrayRgn				*(RgnHandle*)0x9EE

#if TARGET_API_MAC_CARBON
#define DeviceList			GetDeviceList()
#define TheGDevice			GetGDevice()
#define MainDevice			GetMainDevice()
#define GetNextGDevice		GetNextDevice
#else
#define DeviceList			(*(GDHandle*)0x8A8)
#define TheGDevice			(*(GDHandle*)0xCC8)
#define MainDevice			(*(GDHandle*)0x8A4)
#define GetNextGDevice(x)	(*x)->gdNextGD
#endif

/* globals */

RgnHandle		saveGrayRgn = (RgnHandle)NULL;	/* GrayRgn save (in hiding the menubar) */
short			saveMenuBarHeight = 20;		/* set to the default, in case things go bad... */
Boolean			hasHidden = 0;				/* the menubar has been hidden at least once */

static UInt8	bit_depth_table[] =	{16, 1, 2, 3, 4, 5, 6, 7, 8,12,12,12,12,16,16,16,16,
					24,24,24,24,24,24,24,24,32,32,32,32,32,32,32,32,
					 1, 2, 3, 4, 5, 6, 7, 8, 9,10};
static UInt8	bit_true_table[] =	{16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,16,16,16,16,16,
					16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
					 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

/* prototypes */

void GetRawColor(RGBColor *color, UInt32 raw, int type);

/* code */

void myLockPixels(GWorldPtr p)
{
	PixMapHandle	h;

	assert(p);
	h = p->portPixMap;
#ifdef env68k
	assert((long)h>0);
#endif
	assert(LockPixels(h));
}

void myUnlockPixels(GWorldPtr p)
{
	PixMapHandle	h;

	assert(p);
	h = p->portPixMap;
#ifdef env68k
	assert((long)h>0);
#endif
	UnlockPixels(h);
}

PixMapHandle myGetGWorldPixMap(register GWorldPtr p)
{
	assert(p);
#ifdef env68k
	if ((short)(p->portVersion&0xC000)!=(short)0xC000)
		return (PixMapHandle)NULL;
	return p->portPixMap;
#else
	return GetGWorldPixMap(p);
#endif
}

void *myGetPixBaseAddr(register GWorldPtr p)
{
	assert(p);
#ifdef env68k
	/* we should be a color port goddamnit!!! */
	if ((short)(p->portVersion&0xC000)!=(short)0xC000)
	{
		assert(0);
		return NULL;	/* (((GrafPtr)p)->portBits).baseAddr */
	}
	return (*(p->portPixMap))->baseAddr;
#else
	return GetPixBaseAddr(p);
#endif
}

/*****	GrafWorld routines	*****/

OSErr myUpdateGWorld(GWorldPtr *p,
	short d,
	const Rect *r,
	CTabHandle ct,
	GDHandle gd,
	int kind)
{
	PixMapHandle	h = (PixMapHandle)NULL;
	CTabHandle		macintosh_ctab = (CTabHandle)NULL;
	long			err = 0;

	if (!p) return -1;
	vassert(!kind,"can only handle type zero!");

	if (*p)
	{
		h = (*p)->portPixMap;
#ifdef env68k
		assert((long)h>0);
#else
		assert(h);
#endif
		DisposeGWorld(*p);
		if (err=QDError()) dprintf(dgwqe,err);
		*p = NULL;
	}

	/*if (d<=0) d = (**(**TheGDevice).gdPMap).pixelSize; */
	if (d<=0) d = (**(**p).portPixMap).pixelSize;

	if ((((d*RECT_WIDTH(r)+16)*RECT_HEIGHT(r)+7)>>3)+96*KILO>FreeMem())
	{
		dprintf("myUpdateGWorld: not enough memory for projected size");
		return memFullErr;
	}

	if (d==8 && !ct)
	{
		/*ct = macintosh_ctab = build_macintosh_color_table(world_color_table,FALSE); */
		/*assert(macintosh_ctab);	-- an assertion failure should occur within build_macintosh_color_table */
	}
	err = NewGWorld(p,d,r,ct,gd,0);
	if (macintosh_ctab) DisposeHandle((Handle)macintosh_ctab);

	return err;
}


void myHideMenuBar(GDHandle gh, Boolean update)
{
#if TARGET_API_MAC_CARBON
	HideMenuBar();
#else
	RgnHandle	argn = (RgnHandle)NULL;

	if (!saveGrayRgn)
	{
		if (!(saveGrayRgn=NewRgn())) halt();
		if (!hasHidden) atexit(myShowMenuBar);
		CopyRgn(GrayRgn,saveGrayRgn);				/* save the GrayRgn */
		/* fill the requested graphics devices rectangle to the gray region */
		if (!(argn=NewRgn())) halt();
		RectRgn(argn,&(*gh)->gdRect);
		UnionRgn(argn,GrayRgn,GrayRgn);
		if (update)
		{
			XorRgn(saveGrayRgn,argn,argn);
			PaintBehind(NULL,argn);
		}
		DisposeRgn(argn);
		saveMenuBarHeight = MenuBarHeight;
		MenuBarHeight = 0;
	}
#endif
	hasHidden = true;
}

void myShowMenuBar()
{
#if TARGET_API_MAC_CARBON
	ShowMenuBar();
#else
	if (saveGrayRgn)
	{
		/* restore the GrayRgn and the menubar attributes */
		CopyRgn(saveGrayRgn,GrayRgn);
		DisposeRgn(saveGrayRgn);
		saveGrayRgn = (RgnHandle)NULL, MenuBarHeight = saveMenuBarHeight;
	}
#endif
}

/*
typedef struct
{
	UInt8		mode[2];
} _alpha_mode;

static _alpha_mode_table[] =
{
	{srcCopy, blend},
	{srcOr, blend},
	{subPin, subPin},
	{subPin, subPin},
	{subPin, subPin},
	{subPin, subPin},
	{subPin, subPin},
};
*/

void ARGB32ForeColor(ARGB32 _color, _alpha_channel_style style)
{
	static short	modes[] = {subPin, addOver, subOver};

	RGBColor		color;
	short			mode = srcCopy;
/*
	_alpha_mode		modes;

	assert(style>=0 && style<NUMBER_OF_ALHA_CHANNEL_STYLES);
	modes = _alpha_mode_table[style];
	if (modes[0]&0xF0)
	{
		short	mode = if ((UInt8)~_color.a);
	}
*/
	if (style<=_general_overlay)
	{
		if (!(UInt8)~_color.a)
		{
			mode = srcCopy;
			goto skip;
		}
		goto blend;
	} else if (style==_general_text)
	{
		if (!(UInt8)~_color.a)
		{
			mode = srcOr;
			goto skip;
		}
	blend:
		mode = blend;
	} else if (style<NUMBER_OF_ALHA_CHANNEL_STYLES)
		mode = modes[style-_general_subtraction];
	else assert(0);

	color.red = color.green = color.blue = _color.a*257;
	OpColor(&color);
skip:
	if (style==_general_text) TextMode(mode); else PenMode(mode);
color:
	color.red = _color.r*257, color.green = _color.g*257, color.blue = _color.b*257;
	RGBForeColor(&color);
}

void RGB32BackColor(ARGB32 _color)
{
	RGBColor	color;

	color.red = _color.r*257, color.green = _color.g*257, color.blue = _color.b*257;
	RGBBackColor(&color);
}

void SetRawForeColor(UInt32 raw, int type)
{
	RGBColor	color;

	GetRawColor(&color,raw,type);
	RGBForeColor(&color);
}

void SetRawBackColor(UInt32 raw, int type)
{
	RGBColor	color;

	GetRawColor(&color,raw,type);
	RGBBackColor(&color);
}

void SetRawOpacity(UInt32 raw, int type)
{
	RGBColor	color;

	GetRawColor(&color,raw,type);
	OpColor(&color);
}

#ifdef env68k
#	pragma parameter __D1 DER15(__D0)
extern SInt32 DER15(const UInt16) = {0xE9C0, 0x1405};
#	pragma parameter __D1 NEERG15(__D0)
extern SInt32 NEERG15(const UInt16) = {0xE9C0, 0x1545};
#	pragma parameter __D1 ELUB15(__D0)
extern SInt32 ELUB15(const UInt16) = {0xE9C0, 0x1685};

#	pragma parameter __D1 NEERG16(__D0)
extern SInt32 NEERG16(const UInt16) = {0xE9C0, 0x1546};
#else
	
#endif

#pragma parameter __D0 Endian32_Swap(__D0)
extern UInt32 Endian32_Swap(UInt32 value) = {0xE158, 0x4840, 0xE158};

#define FCOLOR16(r,g,b)			{0,2,{65535.0*(r)+.5,65535.0*(g)+.5,65535.0*(b)+.5}}
#define FBASIC16(r,g,b)			{0,1,{65535.0*(r)+.5,65535.0*(g)+.5,65535.0*(b)+.5}}
#define FGRAYC16(x)				{0,0,{65535.0*(x)+.5,65535.0*(x)+.5,65535.0*(x)+.5}}

typedef struct ListedRGBColor
{
	UInt8		flags, type;
	RGBColor	color;
} ListedRGBColor;

static ListedRGBColor general_colors[] =
{
	FGRAYC16(1.0),
	FGRAYC16(14.0/15.0),
	FGRAYC16(0.75),
	FGRAYC16(0.5),
	FGRAYC16(1.0/3.0),
	FGRAYC16(2.0/15.0),
	FGRAYC16(1.0/15.0),
	FGRAYC16(0.0),

	FBASIC16(0.991,2e-2,1e-4),
	FBASIC16(0.995,0.491,0.0),
	FBASIC16(0.995,0.989,0.0),
	FBASIC16(2e-3,0.924,7e-3),
	FBASIC16(0.0,0.955,0.965),
	FBASIC16(0.0,7e-3,0.985),
	FBASIC16(0.645,0.652,1.0),
	FBASIC16(0.925,0.0,0.915),

	FBASIC16(0.491,1e-2,1e-4),
	FBASIC16(1e-3,0.461,1e-4),
	FBASIC16(0.0,3e-3,0.495),
	FBASIC16(0.0,0.481,0.495),
	FBASIC16(0.509,0.495,0.0),
	FBASIC16(0.475,0.0,0.469),

	FCOLOR16(11.0/15.0,11/15.0,0.98),
	FCOLOR16(3.0/5.0,3.0/5.0,4.0/5.0),
	FCOLOR16(4.0/15.0,4.0/15.0,6.0/15.0),
	FCOLOR16(0.0,0.0,1.0/6.0),

	FCOLOR16(1e-4,0.951,0.75),
	FCOLOR16(0.98,0.875,0.75),
	FCOLOR16(0.45,0.34,0.211),
	FCOLOR16(0.72,0.75,0.895),
	FCOLOR16(0.1,0.255,0.395),
	FCOLOR16(0.99,0.75,0.815),
};

void GetRawColor(RGBColor *color, UInt32 _raw, int type)
{
	RGBColor	opcolor;
	UInt32		raw = _raw;
	boolean		opset;

	switch((UInt16)type)
	{
	case _ct_general:
		*color = general_colors[raw].color;
		break;
	case _ct_single:
		color->red = color->green = color->blue = 65535.0*(*(Float32*)&_raw);
		break;
	case _ct_LA32:
		raw = Endian32_Swap(raw);
	case _ct_AL32:
		color->red = color->green = color->blue = raw;
		opcolor.red = opcolor.green = opcolor.blue = raw>>=16;
		goto yep;
	case _ct_ARGB24:
		color->red = 257*((ARGB32*)&_raw)->r, color->green = 257*((ARGB32*)&_raw)->g, color->blue = 257*((ARGB32*)&_raw)->b;
		opcolor.red = opcolor.green = opcolor.blue = 257*((ARGB32*)&_raw)->a;
		goto yep;
	case _ct_ABGR24:
		color->red = 257*((ARGB32*)&_raw)->b, color->green = 257*((ARGB32*)&_raw)->g, color->blue = 257*((ARGB32*)&_raw)->r;
		opcolor.red = opcolor.green = opcolor.blue = 257*((ARGB32*)&_raw)->a;
		goto yep;
	case _ct_RGBA24:
		color->red = 257*((ARGB32*)&_raw)->a, color->green = 257*((ARGB32*)&_raw)->r, color->blue = 257*((ARGB32*)&_raw)->g;
		opcolor.red = opcolor.green = opcolor.blue = 257*((ARGB32*)&_raw)->b;
		goto yep;
	case _ct_BGRA24:
		color->red = 257*((ARGB32*)&_raw)->g, color->green = 257*((ARGB32*)&_raw)->r, color->blue = 257*((ARGB32*)&_raw)->a;
		opcolor.red = opcolor.green = opcolor.blue = 257*((ARGB32*)&_raw)->b;
		goto yep;
	case _ct_ARGB15:
		color->red = (65535*RED15(raw))/31, color->green = (65535*GREEN15(raw))/31, color->blue = (65535*BLUE15(raw))/31;
		opcolor.red = opcolor.green = opcolor.blue = 65535*(((UInt16)raw)>>15);
		goto yep;
	case _ct_ABGR15:
		color->red = (65535*BLUE15(raw))/31, color->green = (65535*GREEN15(raw))/31, color->blue = (65535*RED15(raw))/31;
		opcolor.red = opcolor.green = opcolor.blue = 65535*(((UInt16)raw)>>15);
		goto yep;
	case _ct_RGBA15:
		color->red = (65535*DER15(raw))/31, color->green = (65535*NEERG15(raw))/31, color->blue = (65535*ELUB15(raw))/31;
		opcolor.red = opcolor.green = opcolor.blue = 65535*(raw&1);
		goto yep;
	case _ct_BGRA15:
		color->red = (65535*ELUB15(raw))/31, color->green = (65535*NEERG15(raw))/31, color->blue = (65535*DER15(raw))/31;
		opcolor.red = opcolor.green = opcolor.blue = 65535*(raw&1);
		goto yep;
	case _ct_BGRBGR32:
		color->red = (65535*BLUE15(raw))/31, color->green = (65535*NEERG16(raw))/63, color->blue = (65535*DER15(raw))/31;
		raw >>= 16;
		opcolor.red = (65535*BLUE15(raw))/31, opcolor.green = (65535*NEERG16(raw))/63, opcolor.blue = (65535*DER15(raw))/31;
	yep:
		opset = TRUE;
		break;
	default:
		return;
	}
	if (_raw & _ct_aplha_is_opacity)
		OpColor(&opcolor);
}

/* custom offscreen stuff */

CTabHandle GenerateGrayCTab(SInt32 _count)
{
	Handle			ct = (Handle)NULL;
	CTabPtr			ctp;
	UInt16			*color;
	UInt32			accum, dmag, count;
	short			x, fixed_shift = 16, round;

	x = -1;
	if (!(count=_count) || count>2048)
		goto end;
	ct = NewHandle(8+count<<3);
	HLock(ct);
	ctp = *(CTabHandle)ct;
	color = (UInt16*)ctp->ctTable+1;
	ctp->ctSeed = GetCTSeed(), ctp->ctFlags = 0;
	*(Float64*)ctp->ctTable = 0;
	if (--count)
	{
		dmag = 0xFFFF0000UL / count, round = count>>1;
		do
		{
			accum += dmag;
			*color++ = ++x; *color++ = *color++ = *color++ = (accum+round)>>fixed_shift;
		} while(--count);
	} else ctp->ctSize = 0;
	HUnlock(ct);
end:
	return (CTabHandle)ct;
}

GDHandle NewOffscreenGDevice(
	SInt16 pixelSize,
	SInt16 version,
	SInt16 rowBytes,
	UInt16 cmpSpec,
	CTabHandle ct,
	const Rect *bounds,
	void *pixelData)
{
	Handle			h;
	GDPtr			gd;
	PixMapHandle	pmaph;
	PixMapPtr		pmap;
	Boolean			gray;

	if (pixelSize<=0 || pixelSize>56)
		return (GDHandle)NULL;

	ReserveMem(0x44);
	h = NewHandleClear(0x3E);
	if (h)
	{
		HLock(h);
		gd = (GDPtr)*h;
		if (!(pmaph=NewPixMap()))
		{
			DisposeHandle(h);
			h = (Handle)NULL;
		} else
		{
			gd->gdRect = *bounds, gd->gdPMap = pmaph;
			if (pixelSize>32)
			{
				gray = true;
				/*ct = GetCTable(pixelSize);*/
				pixelSize -= 32;
				/*if (!ct) */
				ct = GenerateGrayCTab(1<<pixelSize);
			} else gray = false;

			if (pixelSize>1)
				gd->gdFlags = 0x4001;
			else
				gd->gdFlags = 0x4000;
			if (cmpSpec & 0xFE00)
				gd->gdType = 2;

			if (!(rowBytes & 0x3FFF))
				rowBytes = CalculateRowBytes(bounds->right-bounds->left,pixelSize,127,-16);
			rowBytes |= 0x8000;
			pmap = (PixMapPtr)(Ptr)*pmaph;
			pmap->baseAddr = pixelData, pmap->pmVersion = version,
				pmap->rowBytes = rowBytes, pmap->bounds = *bounds, pmap->pmTable = ct,
				pmap->pixelType = gd->gdType<<3, pmap->pixelSize = pixelSize,
				pmap->cmpSize = cmpSpec & 255, pmap->cmpCount = cmpSpec>>8;

			if ((gd->gdITable=(ITabHandle)NewHandleClear(4)) && !(cmpSpec & 0xFE00))
				MakeITable(ct,gd->gdITable,0);
			pmap = (PixMapPtr)(Ptr)*pmaph;
			if (!pmap->pmTable)
				pmap->pmTable = (CTabHandle)NewHandleClear(20);
		}
	}
	return (GDHandle)h;
}

OSErr OpenOffscreenCPort(struct XGrafPort *xport,
	CTabHandle ct,
	const Rect *bounds,
	SInt32 depth,
	SInt32 flags)
{
	GrafPtr		old_port;
	PixMapPtr	pix;
	short		width, height, err = 0, mask;
	long		trim;

	if (!xport || !(width=bounds->right-bounds->left) || ((unsigned)width>16383) ||
			!(height=bounds->bottom-bounds->top) || ((unsigned)height>16383) ||
			((unsigned)depth>sizeof(bit_depth_table)))
		return paramErr;

	GetPort(&old_port);
	OpenCPort(&xport->port);
	if (err=QDError()) goto end;
	SetPort((GrafPtr)&xport->port);	/* just to be sure */

	pix = *xport->port.portPixMap;
	*(SInt32*)&pix->bounds = 0;
	pix->bounds.right = width, pix->bounds.bottom = height;
	pix->pixelType = bit_true_table[depth];

	if (flags & kXPortPixels16ByteAligned)
		trim = 127, mask = -16;
	else
		trim = 31, mask = -4;

	xport->pixelData = NewHandle((pix->rowBytes=CalculateRowBytes(width,pix->pixelSize,trim,mask))*height);
	if (err=MemError())
	{
		CloseCPort(&xport->port);
		goto end;
	}
	if (flags & kXPortPixelsLocked)
	{
		HLock(xport->pixelData);
		pix = *xport->port.portPixMap;
		pix->rowBytes |= 0x8000, pix->pmVersion = 1, pix->baseAddr = *xport->pixelData;
	} else
	{
		pix = *xport->port.portPixMap;
		pix->rowBytes |= 0x8000, pix->pmVersion = 2, pix->baseAddr = (Ptr)xport->pixelData;
	}
	MovePortTo(bounds->left,bounds->top);
	PortSize(width,height);
	ClipRect(bounds);
	CopyRgn(xport->port.clipRgn,xport->port.visRgn);
end:
	SetPort(old_port);
	return err;
}

void DisposeOffscreenCPort(struct XGrafPort *xport)
{
	PixMapPtr	pix;

	if (!xport) return;
	pix = *xport->port.portPixMap;
	if (xport->pixelData && *(UInt8*)&xport->pixelData<0x40)
	{
		DisposeHandle(xport->pixelData);
		xport->pixelData = (Handle)NULL;
	}
	CloseCPort(&xport->port);
}
