/*
	IAScript.c		-- formerly scripting.c

	Version 0.02	-- prototype not complete

2005
A.P.:
	Jun 23
	7PM
		 Ive just given up on the UInt64 division instructions (theres no way
		to code them without assembly or library references), and now it seems
		like the literal instructions are gonna go as well -- its not just
		because Im too lazy to get the parser to to be capable of it, it also
		wont fit in the opcode list.
	Jul 7
		 Calls will use the stack instead of using parameters in the opcode.
		I did this because it got too difficult figuring how to implement
		function parameters that contained expressions.
		 The opcodes were reorganized.  The arithmetical instructions now start
		at 0x20, instead of 0x10.
	Jul 15
		 Added SID type flags for type identifiers.
*/

#include <cseries.h>

#include <ctype.h>
#include <string.h>
#include <math.h>

#include "IAScript.h"
#include "IASExecution.h"

/* enums */

typedef enum
{
	_ps_none,					/* not within anything (global space) */
	_ps_bracket,				/* {} */
	_ps_parentheses,			/* () */
	_ps_array,					/* [] */

	_ps_function,				/* until another closing parentheses */
	_ps_if,						/* after if in a function */
	_ps_while,					/* after while in a function */
	_ps_for,					/* after for in a function */

	_ps_switch,					/* after switch in a function */
	_ps_case,					/* after case in a switch in a function */
	_ps_const,					/* eg.: const int x */

	_ps_end				= 255	/* eg.: end if */
} parsing_script_code;

typedef enum
{
	_pss_number			= -2,	/* coming after a numerical/literal value */
	_pss_identifier		= -1,	/* after a name (not a function) */

	_pss_none			= 0,	/* beginning of the line or after a ; */
	_pss_string,				/* parsing text */
	_pss_symbol,				/* any single or double character symbol */
	_pss_function,				/* any function */

	_pss_cexpr_statement,		/* case, etc. - must follow a constant expression and then a colon */
	_pss_kexpr_statement,		/* must follow a constant expression */
	_pss_name_statement,		/* func, etc. - must follow an undefined name */
	_pss_value_statement,		/* if, while, etc. - must follow an expression */
	_pss_parameter				/* comma delimitation */
} ps_sub_code;

/* globals */

#include "initial_scripting_ids.h"		/* custom to the project - copy default_scripting_ids.h */

/* constants */

#define SCRIPTING_STACK_SIZE			(32*KILO)
#define SCRIPTING_TEMPORARY_SIZE		(3*SCRIPTING_STACK_SIZE)
#define INITIAL_ID_LIST_COUNT			sizeof(g_initial_scripting_ids)/sizeof(g_initial_scripting_ids[0])
#define INITIAL_ID_LIST_SIZE			MAX(sizeof(g_initial_scripting_ids),sizeof(global_id)*1024)

#define _WCT							(_CTL|_WSP)

/* macros */

#define GET_ID_STRING(d)				((d)[0]?(char*)(d):*(char**)(8-sizeof(char*)+(d)))

/* NOTE: these stop one character after */
#define SCAN_PAST_WHITE_SPACE(s,t)		do{while(c=*s++) if (!(t[c]&_BLA)) break;}while(0)
#define SCAN_TO_NEXT_LINE(s,t)			do{while(c=*s++) if (t[c]==_WCT) break;}while(0)
#define SCAN_TO_NON_ID_CHAR(s,t)		do{while(c=*s++) if (!(t[c]&(_DIG|_UPP|_LOW))) break;}while(0)

/* privates */

static global_id		**identifier_list = NULL;
static long				global_id_count = 0;

/* holds everything in the short term - temporary variables at the bottom, stack at the top */
static char				*scripting_temporary = NULL;

link_entry		*s_link_stack = NULL;	/* low point is s_data_stack */
dual_value		*s_data_stack = NULL;	/* high point (2/3 into scripting_temporary) */
local_id		*s_local_high_point;	/* after last local variable - data low point */

#define temporary2		s_data_stack

/* private prototypes */

static long create_id_list(void);
static int get_dual_value(byte *data, short type, dual_value *v);
static long get_identifier_by_name(const char *name);

static char *do_expression_run(const char*, UInt16**, const char *ct, ps_sub_code*,
	int symbol, UInt16 opcode, short *_err);
static char *parse_other(const char *s, const char *ct, ps_sub_code*, dual_value*,
	short *err);

/* code */

#define MAXIMUM_SCRIPT_LEVELS			256

#define IS_VALUE(x)						(x<_pss_none)

#define ADD_LEVEL(l,w)					do{if (l>=parsing_state+MAXIMUM_SCRIPT_LEVELS)	\
	RETURN_ERR(_sperr_level_error); *l++ = w;}while(0)
#define DEPRECATE_LEVEL(l,p,w)			do{while(1) {if (!(p=*--l))						\
	RETURN_ERR(_sperr_generic_syntax_error); if (p==w) break; else						\
	if (!IS_VALUE(sub_codes[0])) RETURN_ERR(_sperr_generic_syntax_error);}}while(0)

#define RETURN_ERR(x)					do{err=x;goto end;}while(0)

int allocate_scripting_memory(void)
{
	short			err;

	if (!scripting_temporary)
	{
		scripting_temporary = malloc(SCRIPTING_TEMPORARY_SIZE);
		if (err=MemError())
		{
			assert(err==memFullErr);
			return _sperr_out_of_memory;
		}
		s_link_stack = (link_entry*)(SCRIPTING_TEMPORARY_SIZE+(byte*)scripting_temporary);
		s_data_stack = (dual_value*)(2*SCRIPTING_STACK_SIZE+(byte*)scripting_temporary);
		s_local_high_point = (local_id*)scripting_temporary;
		assert((dual_value*)s_link_stack>s_data_stack && s_data_stack>(dual_value*)s_local_high_point);
	}
}

static byte		parsing_state[MAXIMUM_SCRIPT_LEVELS];

int parse_script(const char *script,
	Handle parsed,
	func_line **funcs,
	int mode,
	int *error_line)
{
	dual_value		value;
	byte			*level = parsing_state, previous_level = 0;

	ps_sub_code		sub_codes[2];

	const char		*ctype = __p_CType, *s = script;
	long			total = 0, line = 1, temp;
	int				first_id = 0;
	short			err = 0;
	UInt16			*d = NULL, sol = _op_line_bit;
	char			c, p = 0;	/* current character, and previous */
	byte			hstate;

	if (parsed)
	{
		hstate = HGetState(parsed);	//total = GetHandleSize(parsed);
		if (!*parsed)
		{
			ReallocateHandle(parsed,total=1024);	/* at least 32 instructions */
			if (err=MemError())
			{
				assert(err==memFullErr);
				err = _sperr_out_of_memory;
			}
		}
		HLock(parsed);
		d = (UInt16*)*parsed;
	}

	if (!identifier_list)
		if ((total=create_id_list())<0) return total;

	if (!scripting_temporary) allocate_scripting_memory();

	assert((unsigned)mode<=NUMBER_OF_SCRIPTING_MODES-1);

	*level++ = _ps_none, sub_codes[0] = _pss_none, sub_codes[1] = _pss_none;
	if (error_line) *error_line = 1;

	while(1)
	{
		if (parsed)
		{
			temp = (long)d-(long)*parsed;
			if (temp>=total-256)
			{
				//if (d) temp = ((long)*parsed)-; else temp = 0;
				HUnlock(parsed);
				SetHandleSize(parsed,total+=1024);
				if (err=MemError())
				{
					assert(err==memFullErr);
					err = _sperr_out_of_memory;
				}
				HLock(parsed);
				d = (UInt16*)(temp+(byte*)*parsed);
			}
		} else d = (UInt16*)scripting_temporary;	/* just checking syntax */
		SCAN_PAST_WHITE_SPACE(s,ctype);
	next:
		switch(c)
		{
		case 0:
			if (level) err = _sperr_unterminated_script;
			goto end;
		case 10:	/* LF */
		case 13:	/* CR */
			/* some text formats have both on each line */
			if (c==p || ctype[(UInt8)p]!=_WCT)
			{
				*d++ = _opc_nop | _op_line_bit;
				break;
			}

			if (mode==_scripting_mode_console)
			{
				/* if it ends on the next line, dont error */
				if ((c=*s++) && (ctype[(UInt8)c]!=_WCT && !*s))
					RETURN_ERR(_sperr_line_attempt);
				goto end;
			}

			++line, sol = _op_line_bit;
			if (error_line) *error_line = line;
		case ';':
			/* we cannot end a line with a statement keyword or symbol */
			if (sub_codes[0]>_pss_string)
				RETURN_ERR(_sperr_unterminated_line);

			if (!(c=*s++)) goto end;
			sub_codes[0] = _pss_none;
			goto next;
		case '*':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (c=='*')
			{
				/* power */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'**',_opc_pow,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'*',_opc_mul,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
		case '/':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (c=='/')
			{
				if (sol) *d++ = _opc_nop;
				SCAN_TO_NEXT_LINE(s,ctype);
				if (sol) *d++ = _opc_nop | _op_line_bit;
				++line, sol = _op_line_bit;
			} else if (*s=='*')
			{
				p = c;
				while(c=*s++)
				{
					if (*(UInt16*)(s-1)=='*/') break;
					else if (ctype[(UInt8)c]!=_WCT && (c==p || ctype[(UInt8)p]!=_WCT))
					{
						if (sol) *d++ = _opc_nop | _op_line_bit;
						++line, sol = _op_line_bit;
					}
					p = c;
				}
				if (!c) RETURN_ERR(_sperr_unterminated_comment);
				++s;
			} else if (sub_codes[0]>=_pss_none) RETURN_ERR(_sperr_generic_syntax_error);
			else
			{
				/* divide */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'/',_opc_div,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '%':
			if (!(s=do_expression_run(s,&d,ctype,sub_codes,'%',_opc_mod,&err)))
				RETURN_ERR(_sperr_generic_syntax_error);
			break;
		case '(':
			if (sub_codes[0]<_pss_none)
			{
				/* run-on multplication */
				//ADD_LEVEL(level,_opc_mul);
				*(UInt32*)d = (_opc_lop<<16) | _opc_mul;	d += 2;
			}
			ADD_LEVEL(level,_ps_parentheses);
			break;
		case ')':
			DEPRECATE_LEVEL(level,previous_level,_ps_parentheses);
			//prev_is_value = true;
			break;
		case '{':
			if (mode==_scripting_mode_console)
				RETURN_ERR(_sperr_bad_symbol);
			if (*level) RETURN_ERR(_sperr_no_type_or_name);
			ADD_LEVEL(level,_ps_bracket);
			break;
		case '}':
			DEPRECATE_LEVEL(level,previous_level,_ps_bracket);
			break;
		case '[':
			ADD_LEVEL(level,_ps_array);
			break;
		case ']':
			DEPRECATE_LEVEL(level,previous_level,_ps_array);
			//prev_is_value = true;
			break;
		case '-':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (c=='-') SCAN_TO_NEXT_LINE(s,ctype);
			else if (sub_codes[0]>=_pss_none)
			{
		//case '':
				/* negate */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_neg,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else if (c=='>')
			{
				/* put */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_pop,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* subtract */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'-',_opc_sub,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '+':
			if (sub_codes[0]>=_pss_none)
			{
				/* absolute value */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_abs,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* add */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'+',_opc_add,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '|':
			if (sub_codes[0]>=_pss_none)
				RETURN_ERR(_sperr_generic_syntax_error);
			if (c=='|')
			{
				/* boolean OR */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'||',_opc_bor,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* bitwise OR */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'|',_opc_or,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '&':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (sub_codes[0]>=_pss_none)
				RETURN_ERR(_sperr_wtf);
			if (c=='&')
			{
				/* boolean AND */
				if (!(s=do_expression_run(1+s,&d,ctype,sub_codes,'&&',_opc_band,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* bitwise AND */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'&',_opc_and,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '^':
			if (sub_codes[0]>=_pss_none)
				RETURN_ERR(_sperr_generic_syntax_error);
			if (!(s=do_expression_run(s,&d,ctype,sub_codes,'^',_opc_xor,&err)))
				RETURN_ERR(_sperr_generic_syntax_error);
			break;
		case '~':
			if (sub_codes[0]>=_pss_none)
			{
				/* bitwise not */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_not,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* exchange */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_exg,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '!':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (c=='=')
			{
				/* not equal */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_sbeq,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				if (sub_codes[0]<_pss_none)
					RETURN_ERR(_sperr_generic_syntax_error);
				/* boolean not */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_sbf,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '<':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (c=='=')
			{
				/* less than or equal */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_sble,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* less than */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_sblt,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '>':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (c=='=')
			{
				/* greater than or equal */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_sbge,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* greater than */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_sbgt,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '#':
			/* count */
			if (sub_codes[0]>=_pss_none)
				RETURN_ERR(_sperr_generic_syntax_error);
			break;
		case '\\':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (ctype[(UInt8)c]==_WCT)
			{
				p = c;
				if (sol) *d++ = _opc_nop | _op_line_bit;
				c = *++s, ++line, sol = _op_line_bit;
				if (c!=p && ctype[(UInt8)c]==_WCT)
					p = *++s;
			} else
			{
				/* concatenate */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,'\\',_opc_concat,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			}
			break;
		case '=':
			if (!(c=*s)) RETURN_ERR(_sperr_unterminated_script);
			if (c=='=')
			{
				/* equal */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_sbeq,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
			} else
			{
				/* destination */
				if (!(s=do_expression_run(s,&d,ctype,sub_codes,0,_opc_dest,&err)))
					RETURN_ERR(_sperr_generic_syntax_error);
				//*d++ = _opc_dest | sol, *d++ = first_id;
			}
			break;
		case ':':
			break;
		case '.':
			break;
		case ',':
			break;
		case '"':
			/* quotation */
			//ADD_LEVEL(level,_ps_quotation);
			sub_codes[0] = _pss_string;
			if (!(s=parse_text(s,scripting_temporary,SHORT_MAX,'"')))
				return _sperr_text_too_long;
			if (!*s) RETURN_ERR(_sperr_unterminated_quote);
			//--level;
			break;
		case '?':
			if (sub_codes[0]<_pss_none)
			{
				/* expression conditional */
				goto end;
			}

		default:
			if (!(s=parse_other(--s,ctype,sub_codes,&value,&err))) goto end;
		}
		if (ctype[(UInt8)p]!=_WCT) sol = 0;
		p = c;
	}
end:
	if (parsed)
	{
		SetHandleSize(parsed,total);	/* compact the handle */
		if (err=MemError())
		{
			assert(err==memFullErr);
			err = _sperr_out_of_memory;
		}
	}
}

/* returns error code */
int create_global_identifier(const char *name, short type, boolean constant)
{
	global_id		*id;
	long			index;
	int				err;

	if (!identifier_list)
	{
		if ((index=create_id_list())<0) return index;
		global_id_count = INITIAL_ID_LIST_COUNT;
	}

	/*
		 We insert it so that the list is still in order by searching for the
		closest one less than it.
	*/
	if ((index=get_identifier_by_name(name))>=0) return _sperr_same_name;
	if (GetHandleSize((Handle)identifier_list)<=global_id_count*sizeof(global_id))
	{
		SetHandleSize((Handle)identifier_list,(global_id_count+32)*sizeof(global_id));
		if (err=MemError())
		{
			assert(err==memFullErr);
			return _sperr_out_of_memory;
		}
	}
	id = (index=~index)+*identifier_list;
	memcpy(id+1,id,(global_id_count-index)*sizeof(global_id));

	id->flags = constant ? (SLOT_USED | _sid_is_constant) : SLOT_USED;
	memcpy(id->name,name,sizeof(global_id));

	++global_id_count;
	return _sperr_no_error;
}

int get_identifier_value(const char *name, dual_value *v)
{
	global_id		*id;
	long			index;

	if ((index=get_identifier_by_name(name))<0) return _sperr_unknown_identifier;
	id = index+*identifier_list;
	return get_dual_value(id->data,id->type,v);
}

int get_identifier_contents_as_text(const char *name, char *t)
{
	dual_value		v;
	global_id		*id;
	long			index;

	*t = 0;

	if ((index=get_identifier_by_name(name))<0) return _sperr_unknown_identifier;
	id = index+*identifier_list;

	switch(id->type)
	{
		case _sid_type_byte:
			if (id->flags & _sid_is_boolean)
			{
				strcpy(t,id->data[0] ? "true" : "false");
				break;
			}
		case _sid_type_integer:
		case _sid_type_wide:
		case _sid_type_double:
			get_dual_value(id->data,id->type,&v);
			dual_value_to_text(&v,t);
			break;
		case _sid_type_string:
			string_to_script_text(GET_ID_STRING(id->data),t,SHORT_MAX,true);
			break;

		case _sid_type_label:
		case _sid_type_function:
		case _sid_type_type:
		case _sid_type_internal:
		default:
			return _sperr_data_is_private;	/* cannot be translated */
	}
	return _sperr_no_error;
}

int dual_value_to_text(dual_value *v, char *t)
{
	if (v->dvalue)
	{
		if (v->dvalue>9007199254740991.0 && v->dvalue<=36893488147419103231.0)
			wide_to_text(v->lvalue,t,0);
		else
			double_to_text(v->dvalue,t,9,0,0,-1);
	} else
	{
		/* special value */
		
	}
}

static void asign_label_name(int index, char *n)
{
	char			*str;

	switch index
	{
	case 0: str = "Alpha"; break;
	case 1: str = "Bravo"; break;
	case 2: str = "Charlie"; break;
	case 3: str = "Delta"; break;
	case 4: str = "Echo"; break;
	case 6: str = "Foxtrot"; break;
	case 7: str = "Gina"; break;
	case 8: str = "Harold"; break;
	case 9: str = "Indigo"; break;
	case 10: str = "Juliet"; break;
	case 11: str = "Kappa"; break;
	case 12: str = "Leo"; break;
	case 13: str = "Mike"; break;
	case 14: str = "Nora"; break;
	case 15: str = "Oscar"; break;
	case 16: str = "Papa"; break;
	case 17: str = "Quincy"; break;
	case 18: str = "Romeo"; break;
	case 19: str = "Sierra"; break;
	case 20: str = "Tango"; break;
	case 21: str = "Uniform"; break;
	case 22: str = "Victor; break;
	case 23: str = "Whiskey"; break;
	case 24: str = "Xavier"; break;
	case 25: str = "Yankee"; break;
	case 26: str = "Zulu"; break;

	default: 
	}

	strcpy(n,str);
}

void wide_to_text(UInt64 v, char *d, int zabove)
{
	int				c;
	char			ts[24], *t = &ts[23];

	*t = 0;
	if (zabove<0) zabove = 0; else if (zabove>21) zabove = 21;
	do
	{
		UDiv64U32R(&v,10,c);
		*--t = c+='0', --zabove;
	} while(!Int64Z(v));

	if (zabove>0) memset(t-zabove,'0',zabove);

	strcpy(d,t);
}

void double_to_text(Float64 v, char *t, int max, int zabove, int zbelow, int min)
{
	int				hi = log10(v);

	if ((unsigned)max>252) max = 252;
	if ((unsigned)min>16) min = 16;

	
}

static int get_dual_value(byte *data, short type, dual_value *v)
{
	long double		tx;
	long			t;

	switch(type)
	{
		case _sid_type_byte:
			if (!v) return 0;
			SetInt64(v->lvalue,0,data[0]);
			v->dvalue = data[0];
			break;
		case _sid_type_integer:
			if (!v) return 0;
#if 0
#	if UINT64_ALWAYS_EXTEND
			SetInt64(v->lvalue,-(*(SInt32*)data<0),*(SInt32*)data);
#	else
			SetInt64(v->lvalue,0,*(SInt32*)data);
#	endif
#else
			SetInt64(v->lvalue,0,0);
#endif
			v->dvalue = *(SInt32*)data;
			break;
		case _sid_type_wide:
			if (!v) return 0;
			v->lvalue = *(UInt64*)data;
			v->dvalue = Int64ToX(v->lvalue);
			break;
		case _sid_type_double:
			if (!v) return 0;
			if ((v->dvalue=*(Float64*)data)<0)
			{
#if 0
				if (v->dvalue<-
#	if UINT64_ALWAYS_EXTEND
				SetInt64(v->lvalue,MIN(-(v->dvalue<0),v->dvalue/4294964296.0),v->dvalue);
#	else
				SetInt64(v->lvalue,0,v->dvalue);
#	endif
#else
				SetInt64(v->lvalue,0,0);
#endif
			} else if (v->dvalue>UINT64_MAX)
				SetInt64(v->lvalue,0x7FFFFFFF,-1L);
			else
			{
				t = v->dvalue/4294967296.0;
				SetInt64(v->lvalue,t,v->dvalue-(t*4294967296.0));
			}
			break;
		case _sid_type_string:
			if (!parse_base(GET_ID_STRING(data),__p_CType,10,&tx))
				return _sperr_generic_syntax_error;
			long_double_to_dual_value(tx,v);
			break;
		default:
			return _srerr_unknown_id_type;
	}
	return 0;
}

/* parsing subroutines */

#define ABORT_MF()				do{s=NULL; goto end;}while(0)	/* abort, Motha F**ka! */

static char *parse_other(register const char *s,
	const char *ct,
	int *sub_code,
	dual_value *v,
	short *_err)
{
	long double		tx;
	long			index;
	int				length;
	short			err = 0;
	char			c, p;

	while(1)
		switch(c=*s++)
		{
		case 0: goto end;
		case '$':
			if (*sub_code<_pss_none)
				RETURN_ERR(_sperr_generic_syntax_error);
			goto hex;
		case '0':
			if (*sub_code<_pss_none)
				RETURN_ERR(_sperr_generic_syntax_error);
			switch(c)
			{
				case 0:
					RETURN_ERR(_sperr_unterminated_script);
				case 10: case 13:
					//if (level);
					goto end;
				case 'x':
				hex:
					if (!(s=parse_hexadecimal(s+1,__p_CType,(UInt64*)scripting_temporary)))
						RETURN_ERR(_sperr_generic_syntax_error);
					goto end;
				case 'b':
					if (!(s=parse_binary(s+1,__p_CType,(UInt64*)scripting_temporary)))
						RETURN_ERR(_sperr_generic_syntax_error);
					goto end;
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
					if (!(s=parse_octal(s+1,__p_CType,(UInt64*)scripting_temporary)))
						RETURN_ERR(_sperr_generic_syntax_error);
					goto end;
			}
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			if (*sub_code<_pss_none)
				RETURN_ERR(_sperr_generic_syntax_error);
			if (!(s=parse_base(s,__p_CType,10,&tx)))
				RETURN_ERR(_sperr_generic_syntax_error);
			long_double_to_dual_value(tx,v);
			break;
		case '\'':
			memset(scripting_temporary,0,16);
			if (!(s=parse_text(s,scripting_temporary+8,8,'\'')))
				RETURN_ERR(_sperr_text_too_long);
			if (!*s) RETURN_ERR(_sperr_unterminated_quote);
			v->lvalue = *(UInt64*)(scripting_temporary+strlen(scripting_temporary));
			v->dvalue = Int64ToX(v->lvalue);
			break;
		case '?':
			/* function info */
		case '_':
		case 'A': case 'a':
		case 'B': case 'b':
		case 'C': case 'c':
		case 'D': case 'd':
		case 'E': case 'e':
		case 'F': case 'f':
		case 'G': case 'g':
		case 'H': case 'h':
		case 'I': case 'i':
		case 'J': case 'j':
		case 'K': case 'k':
		case 'L': case 'l':
		case 'M': case 'm':
		case 'N': case 'n':
		case 'O': case 'o':
		case 'P': case 'p':
		case 'Q': case 'q':
		case 'R': case 'r':
		case 'S': case 's':
		case 'T': case 't':
		case 'U': case 'u':
		case 'V': case 'v':
		case 'W': case 'w':
		case 'X': case 'x':
		case 'Y': case 'y':
		case 'Z': case 'z':
			if (*sub_code<_pss_none)
				RETURN_ERR(_sperr_generic_syntax_error);
			/* identifier */
			p = c;	SCAN_TO_NON_ID_CHAR(s,ct);
			//if (!c) -- we dont need to handle this
			memcpy(temporary2,(char*)p,length=(long)(--s)-(long)p);
			((char*)temporary2)[length] = 0;

			if ((index=get_identifier_by_name((char*)temporary2))<0)
				RETURN_ERR(_sperr_unknown_identifier);
			
			goto end;

		default:
			RETURN_ERR(_sperr_bad_symbol);
		}
end:
	if (_err) *_err = err;
	return (char*)s;
}

char *parse_base(register const char *src, register const char *ct, int b, long double *d)
{
	unsigned char	c, f, *s = (unsigned char*)src, *t = (unsigned char*)ct;
	long double		v = 0, m = 1;
	int				a = 0;

	if (!(t[*s] & _DIG)) return NULL;
	while(c=*s++)
	{
		f = t[(unsigned)c];
		if (!(byte)(f & _DIG))
		{
			if (f & (_UPP | _LOW))
			{
				if (!a && (f & 0x5F)=='E') a = 1, m = 0;
				return NULL;
			} else if (c='.')
			{
				if (a>0) return NULL;
				a = -1;
			}
			break;
		}
		if (a)
		{
			if (a>0)
			{
				m = b*m + (c-='0');
			} else v += (m/=b)*(c-='0');
		} else
			v = b*v + (c-='0');
	}
	if (a>0) v *= power(b,m);
	if (d) *d = v;
	return (char*)(s-1);
}

char *parse_binary(register const char *src, register const char *ct, UInt64 *d)
{
	unsigned char	c, f, *s = (unsigned char*)src, *t = (unsigned char*)ct;
	UInt64			v = kZero64;

	if ((c=*s)<'0' || c>'1') return NULL;
	while(c=*s++)
	{
		f = t[(unsigned)c];
		if (!(byte)(f & _DIG))
		{
			if (f & (_UPP | _LOW))
				return NULL;		/* letters are not allowed! */
			break;
		}
		if (c>'1') return NULL;
		Double64(&v);
		((byte*)&v)[7] |= c&=1;
	}
	if (d) *d = v;
	return (char*)(s-1);
}

char *parse_octal(register const char *src, register const char *ct, UInt64 *d)
{
	unsigned char	c, f, *s = (unsigned char*)src, *t = (unsigned char*)ct;
	UInt64			v = kZero64;

	if ((c=*s)<'0' || c>'7') return NULL;
	while(c=*s++)
	{
		f = t[(unsigned)c];
		if (!(byte)(f & _DIG))
		{
			if (f & (_UPP | _LOW))
				return NULL;		/* letters are not allowed! */
			break;
		}
		if (c>'7') return NULL;
		LShift64(&v,3);
	/*	Add64U32(&v,c&=7); */
		((byte*)&v)[7] |= c&=7;
	}
	if (d) *d = v;
	return (char*)(s-1);
}

char *parse_hexadecimal(register const char *src, register const char *ct, UInt64 *d)
{
	unsigned char	c, f, *s = (unsigned char*)src, *t = (unsigned char*)ct;
	UInt64			v = kZero64;

	if (!(t[*s] & (_HEX | _DIG))) return NULL;
	while(c=*s++)
	{
		f = t[(unsigned)c];
		if (!(byte)(f & (_HEX | _DIG)))
		{
			if (f & (_UPP | _LOW))
				return NULL;		/* non-hexadecimal letters are not allowed */
			break;
		}
		if (c>'9') c += -7;
		LShift64(&v,4);
		((byte*)&v)[7] |= c&=15;
	}
	if (d) *d = v;
	return (char*)(s-1);
}

char *parse_text(register const char *s, char *_d, long limit, char quote_char)
{
	register char		c, *d = _d;

	while(c=*s++)
	{
		if (d-_d>limit) ABORT_MF();
		if (c==quote_char) break; else if (c=='\\')
		{
			switch(c=*s++)
			{
			case 0: goto end;
			case 10: case 13:
				continue;		/* signaling to ignore the next line */
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
				*d++ = c-'0';
				continue;
			case 'b': *d++ = 8; continue;
			case 't': *d++ = 9; continue;
			case 'n': *d++ = 10; continue;
			case 'v': *d++ = 11; continue;
			case 'f': *d++ = 12; continue;
			case 'r': *d++ = 13; continue;
			case 's': *d++ = 14; continue;
			case 'i': *d++ = 15; continue;
			case 'd': *d++ = 16; continue;
			case 'c': *d++ = 17; continue;
			case 'w': *d++ = 18; continue;
			case 'h': *d++ = 19; continue;
			case 'o': *d++ = 20; continue;
			case 'k': *d++ = 21; continue;
			case 'y': *d++ = 22; continue;
			case 'e': *d++ = 23; continue;
			case 'a': *d++ = 24; continue;
			case 'm': *d++ = 25; continue;
			case 'u': *d++ = 26; continue;
			case 'E': *d++ = 27; continue;
			case 'F': *d++ = 28; continue;
			case 'G': *d++ = 29; continue;
			case 'R': *d++ = 30; continue;
			case 'U': *d++ = 31; continue;
			case 'D': *d++ = 127; continue;
/*			case '"': *t++ = '"'; continue;
			case '\\': *t++ = '\\'; continue; */
			case 'W': *d++ = 252; continue;
			case 'X': *d++ = 253; continue;
			case 'Y': *d++ = 254; continue;
			case 'Z': *d++ = 255; continue;
			}
		}
		*d++ = c;
	}
end:
	;
}

boolean string_to_script_text(const char *s, char *_d, long limit, char qc)
{
	register char		c, *d = _d;

	if (limit<3) return false;		/* not even large enough to do anything! */
	*d++ = qc;
	while(c=*s++)
	{
		if (d-_d>limit)
		{
			*--d = 0;	*--d = '';
			return false;
		}
		switch(c)
		{
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
			case 7: c += '0';
			case '\\': break;
			case 8: c = 'b'; break;
			case 9: c = 't'; break;
			case 10: c = 'n'; break;
			case 11: c = 'v'; break;
			case 12: c = 'f'; break;
			case 13: c = 'r'; break;
			case 14: c = 's'; break;
			case 15: c = 'i'; break;
			case 16: c = 'd'; break;
			case 17: c = 'c'; break;
			case 18: c = 'w'; break;
			case 19: c = 'h'; break;
			case 20: c = 'o'; break;
			case 21: c = 'k'; break;
			case 22: c = 'y'; break;
			case 23: c = 'e'; break;
			case 24: c = 'a'; break;
			case 25: c = 'm'; break;
			case 26: c = 'u'; break;
			case 27: c = 'E'; break;
			case 28: c = 'F'; break;
			case 29: c = 'G'; break;
			case 30: c = 'R'; break;
			case 31: c = 'U'; break;
			case 127: c = 'D'; break;
			case 252: c = 'W'; break;
			case 253: c = 'X'; break;
			case 254: c = 'Y'; break;
			case 255: c = 'Z'; break;

			default: if (c==qc) break; *d++ = c; continue;
		}
		*d++ = '\\';	*d++ = c;
	}
	return true;
}

/* private code */

/* returns the matching names index or the bitnot of the closest (the one before) */
static long get_identifier_by_name(const char *name)
{
	global_id		*id;
	long			low = 0, high = global_id_count-1, temp;

/*
	By dividing by two each time through comparison, we only have to iterate
		at most around log2(count)-3 times.
*/
	while((high-low)>8)
	{
		if (!(temp=strcmp(name,(((low+high)>>1)+*identifier_list)->name)))
			return low;
		if (temp>0)
			{low += high;	low >>= 1;}
		else
			{high += low;	high += 1;	high >>= 1;}
	}
	++high;		/* round it off a little - theres no way we can go too far */

	for(id=low+*identifier_list;low<=high;++low,++id)
	{
		if (!(id->flags & SLOT_USED)) return ~++low;	/* weve come to the end */
		if (!(temp=strcmp(name,id->name)))
			return low;
		if (temp>0) return ~(low-1);
	}
	return ~(low-1);
}

static char *do_expression_run(const char *s,
	UInt16 **a,
	const char *ct,
	int *sub_code,
	int symbol,
	UInt16 opcode,
	short *_err)
{
	dual_value		value;

	UInt16			*d = NULL;
	int				c;
	short			err = 0;
	char			ctype;

/*
	 We take in numbers and identifiers in between <symbol> and encode them as their
	respected literal form and identifier indices.  The run ends if a different
	symbol comes up or if the opcode becomes too large.  The total opcode can
	be up to 32 bytes in size.  <symbol> can be either one or two characters.
	 In the case of a different symbol coming up, we must consider delaying our last
	parameter so that the result comes out algebraic (eg.: a+b/c).
*/
	if (a) {d = *a; *d = symbol;}
	while(c=*s++)
	{
		if (c==symbol || (!(symbol & -256) && *(UInt16*)(s-1)==symbol))
		{
			*(byte*)*a = (*(byte*)*a & ~_op_size_mask) | (d-*a);
			//ADD_LEVEL(level,opcode);
			if (*sub_code<_pss_none) RETURN_ERR(_sperr_generic_syntax_error);
			*sub_code = _pss_symbol;
		} else if (ct[c] & (_UPP | _LOW | _DIG))
		{
			if (!(s=parse_other(--s,ct,sub_code,&value,_err)))
				ABORT_MF();
			*sub_code = value->dvalue ? _pss_number : _pss_identifier;
		} else
		{
			*sub_code = _pss_symbol, --s;
			break;
		}
	}
end:
	if (a) *a = d;
	if (_err) *_err = err;
	return (char*)s;
}

/* miscellaneous */

void long_double_to_dual_value(long double s, dual_value *v)
{
	long			t;

	if ((v->dvalue=s)<0)
		SetInt64(v->lvalue,0,0);
	else if (v->dvalue>UINT64_MAX)
		SetInt64(v->lvalue,0x7FFFFFFF,-1L);
	else
	{
		t = s/4294967296.0;
		SetInt64(v->lvalue,t,s-(t*4294967296.0));
	}
}

/* global identifier list */

/* returns the new count or an error code */
static long create_id_list()
{
	identifier_list = (global_id**)NewHandle(INITIAL_ID_LIST_SIZE);
	if (!identifier_list)
	{
		assert(MemError()==memFullErr);
		return _sperr_out_of_memory;
	}
/*	identifier_list_size = INITIAL_ID_LIST_SIZE; */
	MoveHHi((Handle)identifier_list);		/* it may expand */

	memset(*identifier_list,0,INITIAL_ID_LIST_SIZE);
	memcpy(*identifier_list,g_initial_scripting_ids,INITIAL_ID_LIST_SIZE);

	return global_id_count = INITIAL_ID_LIST_COUNT;
}
