/*
	IAScript.h
		Immediate Augmenter Scripting

	Version: 0.02	- prototype not complete
	Revised: July 15, 2005

	Written by A.P.
*/

/* structures */

enum
{
/* scripting modes */
	_scripting_mode_normal			= 0,
	_scripting_mode_restricted,					/* can only change existing variables */
	_scripting_mode_console,					/* single line - cannot create functions */
	NUMBER_OF_SCRIPTING_MODES,

/* identifier type */
	_sid_type_byte					= 0,		/* non-zero conditional - only first byte used */
	_sid_type_boolean	= _sid_type_byte,
	_sid_type_integer,							/* 32 bit signed integer - only first four bytes used */
	_sid_type_wide,								/* 64 bit unsigned integer */
	_sid_type_double,							/* all eight bytes used */
	_sid_type_string,							/* either a small c-string or a pointer to allocated text */
	_sid_type_label,							/* data is func_line */
	_sid_type_function,							/* data is func_line */
	_sid_type_type,								/* data is _sid_type - only first byte used */
	_sid_type_internal,							/* data holds an internal value (private) */

/* parsing errors */
	_sperr_no_error					= 0,
	_sperr_out_of_memory			= -1,		/* could not create/expand handle */
	_sperr_level_error				= -2,		/* went too deep */
	_sperr_modify_constant			= -3,		/* tried to modify a constant id */
	_sperr_unknown_identifier		= -4,		/* all ids must be predefined! */
	_sperr_unknown_function			= -5,		/* a function or command with that name does not exist */
	_sperr_function_attempt			= -6,		/* tried to create a function in restricted or console mode */
	_sperr_variable_attempt			= -7,		/* tried to create a variable in restricted mode */
	_sperr_line_attempt				= -8,		/* more than one line in console mode */
	_sperr_generic_syntax_error		= -9,
	_sperr_bad_symbol				= -10,		/* a certain character cannot be used there in this mode */
	_sperr_same_name				= -11,		/* trying to create an identifier with the same name of another */
	_sperr_unterminated_comment		= -12,
	_sperr_unterminated_quote		= -13,
	_sperr_unterminated_script		= -14,		/* expected closing brackets and/or parenthesis */
	_sperr_unterminated_line		= -15,		/* expected closing parenthesis, etc. */
	_sperr_no_type_or_name			= -16,		/* opening brackets in global space without a name or type */
	_sperr_wtf						= -17,		/* this is not C! */
	_sperr_data_is_private			= -18,		/* contents of internal ids cannot be revealed */
	_sperr_text_too_long			= -19,		/* too much */

/* runtime errors */
	_srerr_bad_goto					= -100,		/* tried to step to a point outside the code */
	_srerr_illegal					= -101,		/* next instruction is bad */
	_srerr_divide_by_zero			= -102,		/* integer divide by zero was caught (divide by zero is allowable if the destination is a double) */
	_srerr_value_overflow			= -103,		/* an integer is being set to a value larger than its bounds */
	_srerr_non_value_math			= -104,		/* tried to do an illegal mathematical operation with a non-value (label, function, etc.) */
	_srerr_id_collision				= -105,		/* tried to a local id with an index that already exists */
	_srerr_unknown_id_type			= -106,		/* internal error - the ids type is invalid */
	_srerr_unknown_dual_value		= -107,		/* (data stack error) */
	_srerr_data_stack_too_deep		= -108,		/* data stack went too deep (too much going on) */
	_srerr_data_flip_side			= -109,		/* stack pointer went over one found in the link entry */
	_srerr_link_too_deep			= -110,		/* link stack went too deep (too much recursion?) */
	_srerr_link_destination			= -111,		/* entry pulled has bad destination */
	_srerr_link_opcode				= -112		/* entry pulled has bad opcode */
};

typedef struct func_line//(8)
{
	SInt32			line;						/* representing line number in script */
	SInt32			offset;						/* offset in parsed code */
} func_line;

typedef struct dual_value//(16)
{
	Float64			dvalue;
	UInt64			lvalue;				/* holds other data when dvalue is zero */
} dual_value;

/* prototypes */

int allocate_scripting_memory(void);	/* call early on to reduce heap fragmentation */

/*
	parse_script

  parses the script into a handle of opcodes (can be inconsistent with different versions
	of script.c - dont try to decipher it unless you know which version youre using)
	and returns a line count or error code (negative always means error)
  also returns the line of where the error code - or zero if no error - in error_line
  if funcs is not NULL, then it will be appended (otherwise a temporary handle will be used
	and destroyed)
  pass NULL to parsed for syntax checking (no code produced)
*/
int parse_script(const char *script, Handle parsed, func_line **funcs, int mode, int *error_line);

/*
	execute_script

  executes the script in the handle; step_count is the number of lines stepped before returning
	- uses and updates step_line (so make sure you pass the right number!)
  use a non-positive count to step count to step an unlimited number of times (a.k.a. run),
	or until a return to process command is reached (returns zero in step_line in that case)
  returns an error code
*/
int execute_script(Handle parsed, int mode, int *step_line, int step_count);

int iterate_scripting_ids(long *index, const char *name, int *type, boolean *constant,
	boolean *write_on_termination);

int create_global_identifier(const char *name, short type, boolean constant);
int get_identifier_value(const char *name, dual_value *v);
int get_identifier_contents_as_text(const char *name, char *t);

int set_identifier_value(const char *name, dual_value *v);
int set_identifier_text(const char *name, const char *t);

/* misc. */

/* for normal translation, pass the global ctype to ct
	returns NULL if invalid */
char *parse_base(const char *src, const char *ct, int base, long double *dst);
char *parse_binary(const char *s, const char *t, UInt64 *d);
char *parse_octal(const char *src, const char *ct, UInt64 *dst);
char *parse_hexadecimal(const char *src, const char *ct, UInt64 *dst);
char *parse_text(const char *s, char *t, long limit, char quote_char);

void long_double_to_dual_value(long double s, dual_value *v);

void wide_to_text(UInt64 v, char *t, int zabove);
void double_to_text(Float64 v, char *t, int max, int zabove, int zbelow, int min);
int dual_value_to_text(dual_value *v, char *t);			/* returns error code */

/* returns true if all is good */
boolean string_to_script_text(const char *s, char *d, long limit, char quote_char);
