/*
 scripting_test.c
  ... and then some

  Revised: June 1, 2006
*/

#include <cs_strings.h>

#include <ctype.h>
#include <string.h>

#include "interface.h"
#include "graphics_core.h"

#include "IASRuntime.h"
#include "mixbit.h"

#include "myDialogs.h"
#include "myDocument.h"
#include "mySound.h"

/* prototypes */

void ProgramMenuInit(void);
void ProgramMenuChoice(long menuHit);
void startup(boolean display);

/* constants */

static const char	format_list[] = "8 bit PCM16 bit PCM1_U7 DPCMIMA ADPCM 4:1"
	"8 bit (reduced 2x)1_U7 (reduced 2x)mixbit standardKMT XADPCM"
	"mixbit standard (reduced 2x)standard (mirrored, reduced 2x)KMT (reduced 2x)";
static const int	format_table[] = {0,1,2,3,64,66,128,130,192,384,386};

/* externs */

extern boolean		mySound_stop;

/* globals */

static ExtSoundHeader		sound_header = {NULL,1,65536000,0,0,-1,60,0,66535.0,NULL,};

static Handle		mixbit_snd = (Handle)NULL;
static KeyMap		keyMap;

/* privates */

static void dump_sound(Handle h, ufixed m, long z);

static OSErr encode_kmt_sfil(void);

static boolean parse_first_mixbit_snd(ExtSoundHeader *d, ufixed *m);
static UInt8 *build_ssnd_command_header(UInt8 *d, UInt8 f, int z, int i);

static void decode_mixbit_whole(mixbit_header *s, void *d, long z);

static void gar2_transmit(void);
static void gar2_unload(void);

static void stop_sound(void);

static int waitMouseDown(void);

/* code */

static void try_mixbit_snd()
{
	Handle			buffer = (Handle)NULL;
	mixbit_header	*h;
	long			length;
	ufixed			m;

	mySound_stop = FALSE;

	if (mixbit_snd==(Handle)NULL)
		return;
	stop_sound();
	sound_header.sampleSize = 16;

	if (!parse_first_mixbit_snd(&sound_header,&m))
		goto end;

	MoveHHi(mixbit_snd);
	HLock(mixbit_snd);

	h = (mixbit_header*)*mixbit_snd;

	length = ((sound_header.numChannels*sound_header.sampleSize+7)>>3)*sound_header.numFrames;

	FlushEvents(everyEvent,0);

	if ((buffer=new_platform_handle(length,TRUE))==(Handle)NULL)
	{
		SetOverheadBoxText(csprintf(temporary,"couldn't allocate data buffer, size %d",
			length));
		goto end;
	}

	sound_header.samplePtr = *buffer;
	decode_mixbit_whole(h,*buffer,length);
	GetKeys(keyMap);
	if (try_snd(&sound_header,m,keyMap[1] & 1)) // shift key
	{
		dump_sound(buffer,m,length);
		goto end;
	}

	if (sound_header.samplePtr!=(Ptr)NULL)
		{dispose_platform_handle(buffer); buffer = (Handle)NULL;}
end:
	sound_header.samplePtr = (Ptr)NULL;
	if (mixbit_snd!=(Handle)NULL)
		{DisposeHandle(mixbit_snd); mixbit_snd = (Handle)NULL;}
}

static boolean parse_first_mixbit_snd(ExtSoundHeader *d, ufixed *m)
{
	mixbit_header	*h;

	if (mixbit_snd==(Handle)NULL)
		{assert(0); return;}
	h = (mixbit_header*)*mixbit_snd;

	if (U64GtU32(h->projected_length,0x800000))
	{
		SetOverheadBoxText(csprintf(temporary,"bogus length: 0x%08X%08X",
			Int64Hi(h->projected_length),Int64Lo(h->projected_length)));
		return FALSE;
	}
	if (h->layer_info>7)
	{
		SetOverheadBoxText(csprintf(temporary,"bad or unsupported layering: 0x%02X",
			h->layer_info));
		return FALSE;
	}

	d->numChannels = h->layer_info+1, d->numFrames = Int64Lo(h->projected_length);

	d->sampleRate = (h->sample_rate+1000)<<15;
	if (!((*(UInt32*)(2+(byte*)&d->AIFFSampleRate)=d->sampleRate) & 0x80000000))
		d->sampleRate <<= 1, *m = 0x10000;
	else
		*m = 0x20000;

	return TRUE;
}

static void dump_sound(Handle h, ufixed m, long z)
{
	SInt32			*d, t = m==0x10000 ? 20 : 28;
	SInt16			refnum, curref = CurResFile();
	OSErr			err;

	SetHandleSize(h,z+t+64);
	if ((err=MemError())!=noErr)
	{
		handle_error(MemError());
		return;
	}
	BlockMove(sound_header.samplePtr,t+64+(byte*)*h,z);
	sound_header.samplePtr = (Ptr)NULL;
	d = (SInt32*)*h;
	memset(d,0,t+64);
	*d++ = 0x10001; *d++ = 0x50000; *d++ = 1;
	if (m!=0x10000)
	{
		*d++ = rateMultiplierCmd<<16; *d++ = m;
		++(**(SndListResource**)h).numCommands;
	}
	*d++ = (0x8000 | bufferCmd)<<16; *d++ = t;
	BlockMove(&sound_header,d,64); // sizeof(sound_header)-2

	if ((t=(int)memchr(&selectedFile.sfFile.name[1],'.',selectedFile.sfFile.name[0]))!=(int)NULL)
		selectedFile.sfFile.name[0] = t -= (int)selectedFile.sfFile.name;

	StandardPutFile(NULL,selectedFile.sfFile.name,&selectedFile);
	if (!selectedFile.sfGood)
		goto bail;
	HDelete(selectedFile.sfFile.vRefNum,selectedFile.sfFile.parID,selectedFile.sfFile.name);
	FSpCreateResFile(&selectedFile.sfFile,'MACS','sfil',0);
	if ((err=ResError())!=noErr)
	{
		handle_error(err);
		goto bail;
	}
	set_mouse_cursor(4,0);
	InitCursor();
	HUnlock(h);

	if ((refnum=FSpOpenResFile(&selectedFile.sfFile,fsRdWrPerm))==-1)
	{
		handle_error(4);
	bail:
		DisposeHandle(h);
		return;
	}
	UseResFile(refnum);
	AddResource(h,'snd ',128,selectedFile.sfFile.name);
	UpdateResFile(curref);
	UseResFile(curref);
	CloseResFile(refnum);
}

static OSErr encode_kmt_sfil()
{
	Handle			h = (Handle)NULL, mxbh = (Handle)NULL;
	SInt16			*s, v, c;
	ufixed			r;
	int				t, w, z, n = 0, i;
	SInt16			refnum, curref = CurResFile();
	int				f;
	OSErr			err = noErr;

	StandardGetFile(NULL,1,(OSType*)&"sfil",&selectedFile);
	if (!selectedFile.sfGood)
		goto end;
	LMSetResLoad(0);
	refnum = FSpOpenResFile(&selectedFile.sfFile,fsRdWrPerm);
	LMSetResLoad(1);
	if (refnum==-1)
		{err = 4; goto end;}

	UseResFile(refnum);
	if ((h=Get1IndResource('snd ',1))!=(Handle)NULL)
		DetachResource(h);
	UseResFile(curref);
	CloseResFile(refnum);
	if (h==(Handle)NULL)
		{err = 5; goto end;}
	MoveHHi(h);
	HLock(h);

	s = *(SInt16**)h;
	v = *s++; t = *s++*6;
	if (v==1)
		s = (SInt16*)(t+(byte*)s);
	else if (t!=2)
		{err = 7; goto end;}
	if ((c=*s++)<=0)
		{err = 6; goto end;}
	do
	{
		if ((UInt16)(*s & 0xFFFE)==0x8050U)
			goto sampled;
		s += 4;
	} while(--c);
	err = 3;
	goto end;
sampled:
	s = (SInt16*)(((SInt32*)s)[1]+*(byte**)h);
	r = ((SoundHeader*)s)->sampleRate, c = 1, w = 8;
	switch(((SoundHeader*)s)->encode)
	{
	case stdSH:
		n = ((SoundHeader*)s)->length;
		t = (long)((SoundHeader*)s)->sampleArea-(long)*h;
		break;
	case extSH:
		c = ((ExtSoundHeader*)s)->numChannels;
		n = ((ExtSoundHeader*)s)->numFrames;
		/* never has it been interpretted other than 8 and 16 */
		w = ((ExtSoundHeader*)s)->sampleSize<=8 ? 8 : 16;
		t = (long)((ExtSoundHeader*)s)->sampleArea-(long)*h;
		break;
	case cmpSH:
		w = ((ExtSoundHeader*)s)->sampleSize;
		n = ((CmpSoundHeader*)s)->numFrames;
		t = (long)((CmpSoundHeader*)s)->sampleArea-(long)*h;
		switch(((CmpSoundHeader*)s)->format)
		{
		case kIMACompression:
			if (w!=16)
				w = 0;
			break;
		case k8BitOffsetBinaryFormat:
			if (w!=8)
				w = 0;
			break;
		case k16BitBigEndianFormat:
			if (w!=16)
				w = 0;
			break;
		default:
			w = 0;
		}
		if (w!=0)
			break;
	default:
		err = 2;
		goto end;
	}
	assert(n>2 && t>=20 && r>=65536000);

	if ((ListDialog(NULL,psprintf(temporary,"%d frames; Select a encoding:",n),
			format_list,delimitedTextType,0,(list_spec_data*)";",(SInt16*)temporary,NULL)>>16)<=0)
		goto end;
//	Debugger();
	f = format_table[*(SInt16*)temporary];

	if ((i=VFix16Mul(c*n,get_ssnd_maximum_size(f)))<=0)
		{dprintf("format not ready!"); goto end;}
	if ((mxbh=NewHandle(64+i))==NULL)
		{err = MemError(); goto end;}
	HLock(mxbh);

	s = *(SInt16**)mxbh;
	/* build the header */
	*s++ = (_mixbit_version_one<<8) | --c;
	*s++ = (r-65503232)>>16;
	*s++ = 0; *s++ = 0; *s++ = n>>16; *s++ = n;

	set_mouse_cursor(4,0);
	if (w<=8)
		for(i=0; i<=c; ++i)
		{
			s = (SInt16*)build_ssnd_command_header((UInt8*)s,f,n,i);
			z = (int)mxb_encode_samples8u(i+(UInt8*)(t+*(byte**)h),(UInt8*)s,n,f,c);
			if (z==(int)NULL)
				break;
			if (f & 0x80)
				((UInt32*)s)[-1] |= z-(int)s; /* OR in the compressed size */
			s = (SInt16*)z;
		}
	else
		for(i=0; i<=c; ++i)
		{
			s = (SInt16*)build_ssnd_command_header((UInt8*)s,f,n,i);
			z = (int)mxb_encode_samples16(i+(SInt16*)(t+*(byte**)h),(UInt8*)s,n,f,c);
			if (z==(int)NULL)
				break;
			if (f & 0x80)
				((UInt32*)s)[-1] |= z-(int)s; /* OR in the compressed size */
			s = (SInt16*)z;
		}

	if (z==(int)NULL)
		{err = 1; goto end;}
	z -= (long)*mxbh;
	DisposeHandle(h); h = NULL;
	HUnlock(mxbh);

	InitCursor();
	SetHandleSize(mxbh,z+1);

	/* and terminate */
	(*mxbh)[z] = 0xFF;

	psprintf(temporary,"%.26P.ssnd", selectedFile.sfFile.name);
	StandardPutFile(NULL,(UInt8*)temporary,&selectedFile);
	if (!selectedFile.sfGood)
		goto end;
	HDelete(selectedFile.sfFile.vRefNum,selectedFile.sfFile.parID,selectedFile.sfFile.name);

	if ((err=FSpCreate(&selectedFile.sfFile,'AWd\267','ssnd',0))!=noErr)
		goto end;

	if ((err=FSpOpenDF(&selectedFile.sfFile,fsRdWrPerm,&refnum))==noErr)
	{
		long		count = z+1;

		HLock(mxbh);
		if ((err=FSWrite(refnum,&count,*mxbh))!=noErr)
			goto end;
		FSClose(refnum);
	}
end:
	if (h!=(Handle)NULL)
		DisposeHandle(h);
	if (mxbh!=(Handle)NULL)
		DisposeHandle(mxbh);

	return err;
}

static UInt8 *build_ssnd_command_header(UInt8 *d, UInt8 f, int z, int i)
{
	if (i>0)
		*d++ = 0xF7;

	if ((z-=17)>=0)
		{*d++ = 0x70 | (z>>16); *d++ = z>>8; *d++ = z;}
	else
		*d++ = z += 0x80;

	*d++ = f;
	if (f & 0x80)
		*d++ = 0, *d++ = 0, *d++ = 0; /* compressed size will come later */

	return d;
}

static void decode_mixbit_whole(mixbit_header *s, void *d, long z)
{
	long			r;

	/* clear it before decoding
	  (decode_mixbit_stream16 actually mixes in the samples!) */
	memset(d,0,z);

	if ((r=decode_mixbit_stream16(s,(void*)(s+1),d,z))<0)
	{
		if (r==-1)
			SetOverheadBoxText("decode_mixbit_stream: bad data");
		else if (r==-4)
			SetOverheadBoxText("decode_mixbit_stream: bad version");
		else
			SetOverheadBoxText(csprintf(temporary,"decode_mixbit_stream: #%d",r));
	}
}

/* system events */

boolean OpenDoc(boolean get)
{
	long			count;
	SInt16			refnum;
	OSErr			err = noErr;

	if (get)
	{
		StandardGetFile(NULL,1,(OSType*)&"ssnd",&selectedFile);
		if (!selectedFile.sfGood)
			return;
	}

	if ((err=FSpOpenDF(&selectedFile.sfFile,fsRdWrPerm,&refnum))!=noErr)
		goto end;
	if ((err=GetEOF(refnum,&count))!=noErr)
		goto close;
	if (count<12)
		count = 12;
	if ((mixbit_snd=NewHandle(count))==(Handle)NULL)
		{err = MemError(); goto close;}
	HLock(mixbit_snd);
	if ((err=FSRead(refnum,&count,*mixbit_snd))!=noErr)
		goto close;
	HUnlock(mixbit_snd);
close:
	FSClose(refnum);

	SetOverheadBoxText((char*)empty);
	try_mixbit_snd();
	SetCurrentSoundName(empty);
end:
	handle_error(err);
}

void startup(boolean display)
{
	static char		bs[] = "brain=5";

	platform_handle		parsed;

	char			*script = bs;
	OSType			tp;
	int				line = 0;
	SInt16			id;

	selectedFile.sfFile.name[0] = 0;

	assert(mixbit_snd==(Handle)NULL);

	if ((mixbit_snd=GetIndResource('mxsn',1))==(Handle)NULL)
		return;
	GetResInfo(mixbit_snd,&id,&tp,selectedFile.sfFile.name);
	DetachResource(mixbit_snd);

	try_mixbit_snd();
	SetCurrentSoundName(empty);

	InitCursor();
	FlushEvents(everyEvent,0);

	create_global_identifier("brain",_ias_type_double);
	parsed = new_platform_handle(0,FALSE);
	ias_parse_script(script,strlen(script),parsed,(const char*)NULL,(platform_handle)NULL,&line);
}

/* menus */

void ProgramMenuInit()
{
	MenuRef			mref;

	if ((mref=GetMenuHandle(MBFileID))==(MenuRef)NULL)
		return;

	MacInsertMenuItem(mref,"\p(-",0);
	MacInsertMenuItem(mref,"\pUnload/U",0);
	MacInsertMenuItem(mref,"\pTransmit/T",0);
	MacInsertMenuItem(mref,"\pEncode system sound/E",0);
	MacInsertMenuItem(mref,"\pPlay ssnd file/P",0);
}

void ProgramMenuChoice(long menuHit)
{
extern Boolean	gDone;
	KeyMap			keys;

	MenuRef			mref;
	WindowRef		win;
	short			menu = menuHit>>16, item = (SInt16)menuHit;
	Boolean			junk;

	GetKeys(keys);
	switch(menu)
	{
	case MBAppleID:
		if (item!=1)
		{
			if ((mref=GetMenuHandle(menu))!=(MenuRef)NULL)
			{
				Str255		astr;

				GetMenuItemText(mref,item,astr);
				junk = OpenDeskAcc(astr);
			} else
				SysBeep(20);
		} else
			Alert(500,NULL);
		break;
	case MBFileID:
		switch(item)
		{
		case 1:	/* play */
			OpenDoc(true);
			break;
		case 2:	/* encode */
			handle_error(encode_kmt_sfil());
			InitCursor();
			break;
		case 3:	/* transmit */
			gar2_transmit();
			break;
		case 4:	/* unload */
			gar2_unload();
			break;
		case 6: /* close */
		case 8: /* quit */
			gDone = TRUE; /* soft exit */
			break;
		}
		break;
	case MBEditID:
  		switch(item)
  		{
		case 1: /* undo */
			break;
		case 3: /* cut */
			break;
		case 4: /* copy */
			break;
		case 5: /* paste */
			break;
		case 6: /* clear */
			break;
		}
	}
}

/* newer gar sh** */

static void encode_gar2(void *data, long length);

#define PORTION_SIZE				(32*KILO)
#define PORTION_HEADING_TOTAL		(128)
#define HEADING_SIGNATURE_OFFSET	(64)
#define HEADING_STRING_OFFSET		(HEADING_SIGNATURE_OFFSET+sizeof(gar2_header))
#define MAXIMUM_PACKET_SIZE			(PORTION_SIZE+PORTION_HEADING_TOTAL)

static byte		gar2_header[] =
	{     0x01,  0x52,0x3A,  0x00,0x00,0x00,0x00,0x00,0x00,0x80,0x00,
	 0xF1,0x01,  0x00,0x00,  0xFF,};

static void gar2_transmit(void)
{
extern short	mySound_counter;

	byte			*buffer = NULL;
	long			count, eof;
	int				r = 0, n = 0;
	OSErr			err = noErr;
	short			refnum;

	StandardGetFile(NULL,-1,(OSType*)NULL,&selectedFile);
	if (!selectedFile.sfGood)
		return;
	stop_sound();

	wait(-1);
	if ((buffer=(byte*)malloc(MAXIMUM_PACKET_SIZE*16))==(byte*)NULL)
		{err = MemError(); goto close;}
	if ((err=FSpOpenDF(&selectedFile.sfFile,fsRdWrPerm,&refnum))!=noErr)
		goto end;
	if ((err=GetEOF(refnum,&eof))!=noErr)
		goto close;
//	if (eof<=0)
//		{err = 8; goto close;}

	sound_header.sampleSize = 8, sound_header.numChannels = 2,
		sound_header.sampleRate = 22050<<16;
	do
	{
		err = noErr;
		count = PORTION_SIZE;
		if ((err=FSRead(refnum,&count,PORTION_HEADING_TOTAL+buffer))!=noErr
				&& err!=eofErr || (count<=0 && sound_header.samplePtr))
			goto close;
		if (err==noErr)
			err = -128;
		if (r==2)
			SetOverheadBoxText("Transmitting all remaining data... press '.' to stop.");
		else
		{
			wait(-1);
			SetOverheadBoxText("Click or press return to play next packet...\015Press 'D' to play all remaining.");
			wait(-1);
			if ((r=waitMouseDown())==0)
				break;
			while(Button())
				;
			SetOverheadBoxText("Transmitting... press '.' to stop.\015Press 'D' to skip and play all remaining.");
		}

		/* Warning: this is speaker audio! */

		wait(-1);
		memset(buffer,0x15,HEADING_SIGNATURE_OFFSET);
		memcpy(HEADING_SIGNATURE_OFFSET+buffer,gar2_header,sizeof(gar2_header));
		memcpy(HEADING_STRING_OFFSET+buffer,selectedFile.sfFile.name,32);
		*(-4+(UInt32*)(PORTION_HEADING_TOTAL+buffer)) = NATIVE_TO_BEU32(n);
		*(-3+(UInt32*)(PORTION_HEADING_TOTAL+buffer)) = NATIVE_TO_BEU32(count);
		*(-2+(UInt32*)(PORTION_HEADING_TOTAL+buffer)) = 0;
		*(-1+(UInt32*)(PORTION_HEADING_TOTAL+buffer)) = NATIVE_TO_BEU32(eof);
		encode_gar2(buffer,PORTION_HEADING_TOTAL+count);
		sound_header.numFrames = (count+PORTION_HEADING_TOTAL)<<3;
		sound_header.samplePtr = (Ptr)buffer;
		if (try_snd(&sound_header,0x10000,FALSE))
			r = 2;
		++n;
	} while(mySound_counter);
close:
	FSClose(refnum);
	/* zero out the sound output */
	memset(buffer,0x80,128);
	sound_header.samplePtr = (Ptr)buffer;
	sound_header.numFrames = 64;
	try_snd(&sound_header,0x10000,FALSE);
end:
	SetCurrentSoundName(empty);
	sound_header.samplePtr = (Ptr)NULL;
	if (buffer!=NULL)
		free(buffer);
	if (err==eofErr)
		err = noErr;
	if (err==noErr)
		SetOverheadBoxText("Done.");
	else
		handle_error(err);
}

static UInt32	gar2_table[4] =
{
	0x111111EE, 0x1111EEEE, 0xEE1111EE, 0xEE11EEEE,
};

static void encode_gar2(void *_d, long c)
{
	UInt8		*s;
	UInt32		*d;
	int			v;

	if ((s=(UInt8*)_d)==NULL && c<=0)
		return;
	d = (UInt32*)(c*16+s); s += c;
	do
	{
		v = *--s;
		*--d = gar2_table[v & 0x3];
		*--d = gar2_table[(v>>2) & 0x3];
		*--d = gar2_table[(v>>4) & 0x3];
		*--d = gar2_table[v>>6];
	} while(--c);
}

static void gar2_unload(void)
{
}

/* be sure you clear the `samplePtr' after playing a sound with try_snd,
  or else you may hang on the next call to this */
static void stop_sound()
{
	if (sound_header.samplePtr!=(Ptr)NULL)
	{
		mySound_stop = TRUE;
		while(mySound_stop)
			wait(-1);
	}
}

static int waitMouseDown()
{
	EventRecord		myEvent;

	while(!Button())
		if (GetNextEvent(keyDownMask | autoKeyMask,&myEvent))
			switch((char)myEvent.message)
			{
			case 13:
				return 1;
			case '.':
				return 0;
			case 'd':
			case 'D':
				return 2;
			}

	return 1;
}
