
#include "macintosh_cseries.h"

#include "interface.h"
#include "shell.h"

#ifndef __MACMEMORY__
#	include <MacMemory.h>
#endif
#ifndef __STANDARDFILE__
#	include <StandardFile.h>
#endif
#ifndef __ENDIAN__
#	include <Endian.h>
#endif
#ifndef __SOUND__
#	include <Sound.h>
#endif

typedef void (*PercentCompletionProc)(fract);

OSErr translate_wave_to_file(short in_file, short out_file, PercentCompletionProc);
OSErr translate_file_to_wave(ConstStr255Param name, short in_file, short out_file, UInt16 rate, PercentCompletionProc);

#define MAXIMUM_BLOCK_SIZE			32*KILO		/* WARNING: output is 8 times this! */

static struct
{
	OSType		dataType;
	UInt32		fileLength;		/* past starting data (8 bytes) */

	/* wave header   = WAVE ckid */
	OSType		waveType;

	/* PCM format header = fmt  ckid, 16, 1 */
	OSType		formatType;
	UInt32		formatLength;
	SInt16		format;

	SInt16		channelCount;
	UInt16		sampleRate, filler1, averageBytesPerSec, filler2;
	SInt16		blockAlignSize, nBitsPerSample;

	OSType		dataStartCkID;
	UInt32		dataLength;
} wave_header_template =
{
	'RIFF', -8, 'WAVE',
	'fmt ', Endian32_Swap(16), Endian16_Swap(1),
	Endian16_Swap(1),
	Endian16_Swap(rate11025hz),-1,Endian16_Swap(rate11025hz),-1,
	Endian16_Swap(1),
	Endian16_Swap(8),
	'data', 0
};

/* go for inline instead for the code section */
#ifdef env68k
#	undef Endian16_Swap
#	undef Endian32_Swap
#endif

OSErr translate_wave_to_file(short src, short dst, PercentCompletionProc completion)
{
	ParamBlockRec	src_file, dst_file;

	void		*buffer = malloc(MAXIMUM_BLOCK_SIZE);
	UInt32		previous_ticks = TickCount(), position = 0, eof;
	OSErr		err = 0;

	if (!buffer) return MemError();
	src_file.ioParam.ioRefNum = src, dst_file.ioParam.ioRefNum = dst;
	if (err=PBGetEOFSync(&src_file)) goto end;
	eof = (UInt32)src_file.ioParam.ioMisc;
	while(position<eof)
	{
		if (((*(UInt32*)0x16A)-previous_ticks) & 16)	/* Ticks -- ~1/4 second*/
			completion(NFix30Div(position,eof));
		if (err=PBReadSync(&src_file)) goto end;

		position += MAXIMUM_BLOCK_SIZE;
	}
end:
	return err;
}

/*
	each two bits represent
		0:		+0
		1:		++
		2:		-0
		3:		-+
	so...
		$63	=	++-0+0-+
		$AF	=	-0-0-+-+
		...
*/
static const UInt32		hex_samples[16] =
{
	0xAA80AA80,		/* 0 */
	0xAA80AACC,
	0xAA805580,
	0xAA8055CC,

	0xAACCAA80,		/* 4 */
	0xAACCAACC,
	0xAACC5580,
	0xAACC55CC,

	0x5580AA80,		/* 8 */
	0x5580AACC,
	0x55805580,
	0x558055CC,

	0x55CCAA80,		/* C */
	0x55CCAACC,
	0x55CC5580,
	0x55CC55CC,		/* F */
};

#pragma parameter translate_bytes_8b4(__A0,__A1,__D2)
extern void translate_bytes_8b4(byte*,const UInt32*,SInt32) =
	{0x2F0A, 0x45F0, 0x2600, 0xD1C2, 0x1020, 0xE9C0, 0x1604, 0x240, 15, 0x2531, 0x400,
		0x2531, 0x1400, 0x5382, 0x6EEA, 0x245F};
/*
		move.l		a2,-(sp)
		lea			(a0,d2.w*8),a2
		adda.l		d2,a0

@loop	move.b		-(a0),d0
		bfextu		d0{24:4},d1
		andi.w		#0xF,d0
		move.l		(a1,d0.w*4),-(a2)
		move.l		(a1,d1.w*4),-(a2)
		subq.l		#1,d2
		bgt.s		@loop

		move.l		(sp)+,a2
*/

OSErr translate_file_to_wave(ConstStr255Param name,
	short src, short dst,
	UInt16 rate,
	register PercentCompletionProc completion)
{
	ParamBlockRec	src_file, dst_file;

	UInt32			length;

	register byte		*buffer = malloc(MAXIMUM_BLOCK_SIZE*8);
	register UInt32		previous_ticks = TickCount(), position = 0, eof, size;
	register OSErr		err = 0;

	if (!buffer) return MemError();

	src_file.ioParam.ioRefNum = src, dst_file.ioParam.ioRefNum = dst;
	if (err=PBGetEOFSync(&src_file)) goto end;
	eof = ((length=(UInt32)src_file.ioParam.ioMisc)<<3)+2176+sizeof(wave_header_template);
	dst_file.ioParam.ioMisc = (Ptr)eof;
	if (err=PBSetEOFSync(&dst_file)) goto end;

	wave_header_template.averageBytesPerSec = wave_header_template.sampleRate =
		Endian16_Swap(rate);
	wave_header_template.fileLength = Endian32_Swap(eof-8);
	wave_header_template.dataLength = Endian32_Swap(eof-sizeof(wave_header_template));

	dst_file.ioParam.ioBuffer = (Ptr)&wave_header_template;
	dst_file.ioParam.ioReqCount = sizeof(wave_header_template);
	dst_file.ioParam.ioPosMode = fsFromStart;
	dst_file.ioParam.ioPosOffset = 0;
	if (err=PBWriteSync(&dst_file)) goto end;

	memset(buffer,0,272);
	((SInt32*)buffer)[0] = ((SInt32*)buffer)[1] = 0x1BE41BE4;
	/* 32 bit addressing for now - first four bytes are left zero */
	*(3+(UInt32*)buffer) = length;

	BlockMove(name,buffer+16,name[0]+1);
	translate_bytes_8b4(buffer,hex_samples,272);

	dst_file.ioParam.ioBuffer = (Ptr)buffer;
	dst_file.ioParam.ioReqCount = 2176;
	dst_file.ioParam.ioPosMode = 0;
	dst_file.ioParam.ioPosOffset = 0;
	if (err=PBWriteSync(&dst_file)) goto end;

	while(length>position)
	{
		if (((*(UInt32*)0x16A)-previous_ticks)>5)	/* Ticks -- 1/12 second*/
		{
			previous_ticks = TickCount();
			completion(NFix30Div(position,length));
		}
		size = length-position;
		src_file.ioParam.ioBuffer = (Ptr)buffer;
		src_file.ioParam.ioReqCount = MIN(size,MAXIMUM_BLOCK_SIZE);
		src_file.ioParam.ioPosMode = fsFromStart;
		src_file.ioParam.ioPosOffset = position;
		if (err=PBReadSync(&src_file)) goto end;

		position += MAXIMUM_BLOCK_SIZE;

		translate_bytes_8b4(buffer,hex_samples,src_file.ioParam.ioReqCount);

		dst_file.ioParam.ioBuffer = (Ptr)buffer;
		dst_file.ioParam.ioReqCount = src_file.ioParam.ioReqCount<<3;
		dst_file.ioParam.ioPosMode = 0;
		dst_file.ioParam.ioPosOffset = 0;
		if (err=PBWriteSync(&dst_file)) goto end;
	}
end:
	return err;
}

static void myCompletionRoutine(fract percent)
{
	static short	run = 0;

	if (++run & 3) return;
	if (percent>0)
	{
		csprintf(temporary,"%d%% Completed", NFix30Mul(percent,100));
		SetCrapBoxText(temporary);
	} else SetCrapBoxText((run & 4) ? "--" : "||");
	HandleEvent(NULL,NULL,0,0);
	if (run & 4) HandleEvent(NULL,NULL,0,0);
	if (run & 8) HandleEvent(NULL,NULL,0,0);
}

enum
{
	_user_canceled,
	_translate_from_audio,
	_translate_to_audio
};

static const char	user_cancel_text[] = "- User Canceled -";

void startup(boolean display);
void startup(boolean display)
{
	StandardFileReply	source_file, destination_file;

	OSType		file_type, creator_type;
	UInt16		rate = 11025;
	short		command;

	if (display)
		SetCrapBoxText("Press F to translate from wave, and T -> wave");
	while(1)
	{
		EventRecord		event;

		if (!WaitNextEvent(everyEvent,&event,30,NULL)) continue;
		if (event.what==keyDown || event.what==autoKey)
		{
			switch((Byte)event.message)
			{
				case '.':
				case kEscapeCharCode:
					command = _user_canceled;
					SetCrapBoxText(user_cancel_text);
					return;
				case 'f':
				case 'F':
					command = _translate_from_audio;
					break;
				case 't':
				case 'T':
					command = _translate_to_audio;
					break;
			}
			break;
		}
		HandleEvent(NULL,&event,true,0);
	}
	if (command==_user_canceled) return;

	if (command==_translate_to_audio)
	{
		file_type = 'WAVE', creator_type = 'SCPL';

		SetCrapBoxText("Press 1-4 for the sample rate: (r * 11025)");
		while(1)
		{
			EventRecord		event;
	
			WaitNextEvent(everyEvent,&event,30,NULL);
			if (event.what==keyDown || event.what==autoKey)
			{
				switch((Byte)event.message)
				{
					case '.':
					case kEscapeCharCode:
						command = _user_canceled;
						SetCrapBoxText(user_cancel_text);
						return;
					case '1':
					case '2':
					case '3':
					case '4':
						rate *= event.message & 7;
						break;
				}
				break;
			}
			HandleEvent(NULL,&event,true,0);
		}
		StandardGetFile(NULL,-1,NULL,&source_file);
		SetCrapBoxText((char*)empty);
		memcpy(destination_file.sfFile.name,source_file.sfFile.name,sizeof(StrFileName));
		if (destination_file.sfFile.name[0]>27)
		{
			destination_file.sfFile.name[0] = 27;
			destination_file.sfFile.name[27] = '';
		}
	} else
	{
		file_type = 'TEXT', creator_type = 'SITx';

Debugger();
		SetCrapBoxText((char*)empty);
		StandardGetFile(NULL,2,(OSType*)"WAVE.WAV",&source_file);
	}
	HandleEvent(NULL,NULL,0,0);
	HandleEvent(NULL,NULL,0,0);
	HandleEvent(NULL,NULL,0,0);

	if (!source_file.sfGood) return;
	psprintf(temporary,"%P.gar", destination_file.sfFile.name);
	StandardPutFile(NULL,command==_translate_to_audio ? (UInt8*)temporary : empty,&destination_file);
	if (!destination_file.sfGood) return;

	{
		short		source = 0, destination = 0;
		OSErr		err;

		FSpDelete((const FSSpec*)&destination_file.sfFile);
		if (err=FSpCreate((const FSSpec*)&destination_file.sfFile,creator_type,file_type,smSystemScript))
			goto close;
		if (err=FSpOpenDF((const FSSpec*)&source_file.sfFile,fsRdPerm,&source))
			goto close;
		if (err=FSpOpenDF((const FSSpec*)&destination_file.sfFile,fsWrPerm,&destination))
			goto close;

		switch(command)
		{
			case _translate_from_audio:
				err = translate_wave_to_file(source,destination,myCompletionRoutine);
				break;
			case _translate_to_audio:
				err = translate_file_to_wave(source_file.sfFile.name,
					source,destination,rate,myCompletionRoutine);
		}
	close:
		if (source>2)
			FSClose(source);
		if (destination>2)
			FSClose(destination);
		if (err)
		{
			csprintf(temporary,"Error: %d", err);
			SetCrapBoxText(temporary);
		} else SetCrapBoxText("- Done -");
	}
}
