/*
 tapedata2.c
  source file for my second version of transmitting binary data over audio

A.P.:
2005
	May 11
		Creation - or duplication/variant from version one.
	May 14
		Added short_int to reduce the number of instructions generated on M68K.
			-- it still doesnt work.
		Finally wrote source code for the program to be able to run on any platform.
	11AM
		The interpretation code works! - partly (there is junk at the beginning and
			end).
		Junk at end was due to the length not being counted down when the scan for
			significant change left.  Fixed by subtracting afterward.
		Changed it to save as directory; new file gets put in directory with name
			found in the wave.
		The junk at the beginning was due to the heading counter being in the wrong
			place.
	12PM
		Success!
	1PM
		Replaced division with conditionals using precalculated values.  Now all of
			the (short_int) casts can be removed.
		It still hasnt been tested for loss and noise.
	2PM
		CRs and LFs wont be replaced with spaces anymore.
		Instead of just one, the first two samples are now ignored.
		Fixed a bug generated by the last thing I just said.
		Fixed a problem where the adjustment to avoid same next sample cases was
			interpreted as a next byte.
	3PM
		Fixed a bug where a folder was being created for _every_ file.
		I made it a little less sensitive to sharp deltas.
	4PM
		Turns out it was reading past the buffer while the remaining byte count was
			still positive.  I think its how the while loops were constructed.
			Once it comes to the end of the buffer, it may bail before the part to
			subtract from the remaining byte count.
		It worked on the other two files, but not on one (and would mean others as
			well).
	May 15
	1PM
		Fixed a mistake where temporary was used as a parameter.
		Fixed a mistake regarding the process not updating the remaining byte count.
	2PM-4PM
		Success!  I could find no differences for all the files tested (plus more
			later on).
		I modified the table to have more visible seperation between sample chunks.
	May 19
		After adding the debug log, Ive found that things have gone bad here...
			and its because of the changes made.
	May 20
		After comparing with the backup code, it appears to be just one line that
			was missing after some changes were reverted.
2006
	February 16
		Cleaned this up.
*/

#include "macintosh_cseries.h"

#include "interface.h"
#include "shell.h"

#include <MacMemory.h>
#include <StandardFile.h>
#include <Endian.h>
#include <Sound.h>

#include "myDialogs.h"
#include "format_recognition.h"

typedef void (*PercentCompletionProc)(sfract);

OSErr translate_wave_to_file(Str255, short in_file, short out_file, PercentCompletionProc);
OSErr translate_file_to_wave(ConstStr255Param name, short in_file, short out_file, UInt16 rate, PercentCompletionProc);
Boolean filter_name(Str255 name);

#define MAXIMUM_BLOCK_SIZE			32*KILO		/* WARNING: output is 5 times this! */

static struct
{
	OSType			dataType;
	UInt32			fileLength;		/* past starting data (8 bytes) */

	/* wave header   = WAVE ckid */
	OSType			waveType;

	/* PCM format header = fmt  ckid, 16, 1 */
	OSType			formatType;
	UInt32			formatLength;
	SInt16			format;

	SInt16			channelCount;
	UInt16			sampleRate, filler1, averageBytesPerSec, filler2;
	SInt16			blockAlignSize, nBitsPerSample;

	OSType			dataStartCkID;
	UInt32			dataLength;
} wave_header_template =
{
	'RIFF', -8, 'WAVE',
	'fmt ', Endian32_Swap(16), Endian16_Swap(1),
	Endian16_Swap(1),
	Endian16_Swap(rate11025hz),-1,Endian16_Swap(rate11025hz),-1,
	Endian16_Swap(1),
	Endian16_Swap(8),
	'data', 0
};

/* go for inline instead for the code section */
#ifdef env68k
#	undef Endian16_Swap
#	undef Endian32_Swap
#endif

static const char	pStreamDump[] = "+stream aborted @ $%X";
static const char	nStreamDump[] = "-stream aborted @ $%X";

#define SCAN_TO_MINIMUM(a,m,l)		\
	do{ c = *a++;	d = c-p;	p = c;	\
		if (d>m || d<-m) break; }while(--l)

OSErr translate_wave_to_file(Str255 name,
	short src, short dst,
	PercentCompletionProc completion)
{
	ParamBlockRec	src_file, dst_file;

	register int	r;

	void			*buffer = malloc(MAXIMUM_BLOCK_SIZE);
	UInt64			total;
	UInt32			previous_ticks = TickCount();
	long			length, position, eof, size, count = 0;
	short_int		m, p, ph, nh, pm, nm, px, nx, i = 2, j = 0;
	OSErr			err = 0, prb = 0;

	if (!buffer) return MemError();
	src_file.ioParam.ioRefNum = src, dst_file.ioParam.ioRefNum = dst;
	src_file.ioParam.ioBuffer = buffer;
	src_file.ioParam.ioReqCount = 256,
		src_file.ioParam.ioPosMode = 0, src_file.ioParam.ioPosOffset = 0;
	if (err=PBReadSync(&src_file)) goto end;

	{
		register SInt32		*aptr = buffer;

		if (*aptr++!='RIFF') return -1;
		if ((eof=Endian32_Swap(*aptr++))<16) goto end;
		if (*aptr++!='WAVE') return -1;

		eof += 8;

		while(*aptr++!='data')
		{
			length = (1+Endian32_Swap(*aptr++))&~1;
			if ((unsigned)length>0x40000000) return -1;
			aptr = (SInt32*)(length+(Byte*)aptr);
			if ((long)aptr-(long)buffer>255)
			{
dprintf("header is too large or has no data tag");
				return 0;
			}
		}
		eof = Endian32_Swap(*aptr++);
		if (eof>0x40000000) return -1;
		eof += position = (long)aptr-(long)buffer;
	}
	while(position<eof)
	{
		register UInt8		*aptr = buffer, *bptr = aptr;
		register short_int	c, d = 0;

		if (((*(UInt32*)0x16A)-previous_ticks) & 16)	/* Ticks -- ~1/4 second*/
			completion(NFix30Div(position,eof));
		size = eof-position;

		src_file.ioParam.ioBuffer = buffer;
		src_file.ioParam.ioReqCount = length = MIN(size,MAXIMUM_BLOCK_SIZE),
			src_file.ioParam.ioPosMode = fsFromStart,
			src_file.ioParam.ioPosOffset = position;
		if (err=PBReadSync(&src_file)) goto end;

		if (!count)
		{
			p = *aptr++;
			if (--length<=0) goto more;
			/* get the magnitude by searching for a minimum delta (noise gate)
				but skip the first two (theyre kind of likely to be off) */
			do
			{
				SCAN_TO_MINIMUM(aptr,3,length);
				m = d, ++j;
				assert(d);
				if (--length<=0) goto more;
			} while(--i);
//--aptr;
			if (d<0) pm = -d; else pm = d;
			assert(pm<=17);					/* be sure that clipping isnt imminent */

			nm = -pm, ph = (pm+1)>>1;	nh = -ph, r = 0;
			pm += ph, nm += nh;				/* inward threshold */
			px = pm*2+ph, nx = nm*2+nh;		/* outward threshold */

			/* the length and name string come first */
			while(--length>=0)
			{
				c = *aptr++;	d = c-p;
				p = c;
				if (d>=ph || d<=nh)
				{
					r <<= 2;
					if (m<0) d = -d;
					if (d>=pm)
					{
						r += 1;
						if (d>px)
						{
							/* delta is way too large (on the positive side);
								it is likely to go out of order */
	dprintf(pStreamDump, position+(long)aptr-(long)buffer);
							aptr -= 2;	p = *aptr, length += 2, j = 3, prb |= 2;
						}
					} else if (d<=nm)
					{
						r += 3;
						if (d<nx)
						{
							/* delta is way too large (on the negative side)... */
	dprintf(nStreamDump, position+(long)aptr-(long)buffer);
							aptr -= 2;	p = *aptr, length += 2, j = 3, prb |= 2;
						}
					} else if (d<0)
						r += 2;
					if (++j>=4)
					{
						*bptr++ = r, j = 0;	r = 0;
						if (++i>255) break;
						SCAN_TO_MINIMUM(aptr,3,length);
						if (--length<=0) break;
					}
				}
			}
			total = ((UInt64*)buffer)[1], bptr = buffer;
			BlockMoveData(2+(UInt64*)buffer,name,256);
			bptr = buffer;
			if (length<=0) goto more;
			SCAN_TO_MINIMUM(aptr,3,length);
			if (--length<=0) goto more;
		}
		if (j<0)
		{
			SCAN_TO_MINIMUM(aptr,3,length);
			j = 0;
			if (--length<=0) goto more;
		}
		while(--length>=0)
		{
			c = *aptr++;	d = c-p;
			p = c;
			if (d>=ph || d<=nh)
			{
				r <<= 2;
				if (m<0) d = -d;
				if (d>=pm)
				{
					r += 1;
					if (d>px)
					{
dprintf(pStreamDump, position+(long)aptr-(long)buffer);
						aptr -= 2;	p = *aptr, length += 2, j = 3, prb |= 2;
					}
				} else if (d<=nm)
				{
					r += 3;
					if (d<nx)
					{
dprintf(nStreamDump, position+(long)aptr-(long)buffer);
						aptr -= 2;	p = *aptr, length += 2, j = 3, prb |= 2;
					}
				} else if (d<0)
					r += 2;
				if (++j>=4)
				{
					*bptr++ = r; r = 0;
					if (length<=0)
					{
						j = -1; break;
					}
					j = 0;
					SCAN_TO_MINIMUM(aptr,3,length);
					if (--length<=0) break;
				}
			}
		}
	more:
		position += MAXIMUM_BLOCK_SIZE, ++count;

		dst_file.ioParam.ioBuffer = (Ptr)buffer;
		dst_file.ioParam.ioReqCount = (long)bptr-(long)buffer,
			dst_file.ioParam.ioPosMode = 0, dst_file.ioParam.ioPosOffset = 0;
		if (err=PBWriteSync(&dst_file)) goto end;
	}
end:
	if (!err) err = prb;
	return err;
}

/*
	difference between each sample per byte
		+1/N:	0
		+2/B:	1
		-1/A:	2
		-2/P:	3
	so...
		$63	=	1203
		$AF	=	1133
		...
*/
static const UInt32		byte_samples[256] =
{
	0x90A0B0C0, 0x90A0B0D0, 0x90A0B0A0, 0x90A0B090,			/* 00 */
	0x90A0C0D0, 0x90A0C0E0, 0x90A0C0B0, 0x90A0C0A0,
	0x90A090A0, 0x90A090B0, 0x8F9E8D79, 0x90A09070,
	0x90A08090, 0x90A080A0, 0x90A08070, 0x90A08060,

	0x90B0C0D0, 0x90B0C0E0, 0x90B0C0B0, 0x90B0C0A0,			/* 10 */
	0x90B0D0E0, 0x90B0D0F0, 0x90B0D0C0, 0x90B0D0B0,
	0x90B0A0B0, 0x90B0A0C0, 0x90B0A090, 0x90B0A068,
	0x90B090A0, 0x90B090B0, 0x90B08C78, 0x90B09070,

	0x908090A0, 0x908090B0, 0x8F7E8D79, 0x90809070,			/* 20 */
	0x9080A0B0, 0x9080A0C0, 0x9080A090, 0x9080A068,
	0x91827387, 0x90807090, 0x90807060, 0x90807050,
	0x90806070, 0x90806088, 0x90806050, 0x90806040,

	0x90708090, 0x907080A0, 0x90708070, 0x90708060,			/* 30 */
	0x907090A0, 0x907090B0, 0x906C8C78, 0x90709070,
	0x90706070, 0x90706088, 0x90706050, 0x90706040,
	0x90705060, 0x90705070, 0x90705040, 0x90705030,

	0xA0B0C0D0, 0xA0B0C0E0, 0xA0B0C0B0, 0xA0B0C0A0,			/* 40 */
	0xA0B0D0E0, 0xA0B0D0F0, 0xA0B0D0C0, 0xA0B0D0B0,
	0xA0B0A0B0, 0xA0B0A0C0, 0xA0B0A090, 0xA0B0A078,
	0xA0B090A0, 0xA0B090B0, 0xA0B08C78, 0xA0B09070,

	0xA0C0D0E0, 0xA0C0D0F0, 0xA0C0D0C0, 0xA0C0D0B0,			/* 50 */
	0xA0C0E0F0, 0xA0C0E0FF, 0xA0C0E0D0, 0xA0C0E0C0,
	0xA0C0B0C0, 0xA0C0B0D0, 0xA0C0B0A0, 0xA0C0B090,
	0xA0C0A0B0, 0xA0C0A0C0, 0xA0C0A090, 0xA0C0A078,

	0xA090A0B0, 0xA090A0C0, 0xA090A090, 0xA090A078,			/* 60 */
	0xA090B0C0, 0xA090B0D0, 0xA090B0A0, 0xA090B090,
	0xA0908090, 0xA09080A0, 0xA0908070, 0xA0908060,
	0xA2927286, 0xA0907090, 0xA0907060, 0xA0907050,

	0xA08090A0, 0xA08090B0, 0xA07C8C78, 0xA0809070,			/* 70 */
	0xA080A0B0, 0xA080A0C0, 0xA080A090, 0xA080A078,
	0xA2827286, 0xA0807090, 0xA0807060, 0xA0807050,
	0xA0806070, 0xA0806088, 0xA0806050, 0xA0806040,

	0x708090A0, 0x708090B0, 0x6E7D8C78, 0x70809070,			/* 80 */
	0x7080A0B0, 0x7080A0C0, 0x7080A090, 0x7080A078,
	0x70817286, 0x70807090, 0x70807060, 0x70807050,
	0x70806070, 0x70806088, 0x70806050, 0x70806040,

	0x7090A0B0, 0x7090A0C0, 0x7090A090, 0x6F8F9F78,			/* 90 */
	0x7090B0C0, 0x7090B0D0, 0x7090B0A0, 0x7090B090,
	0x70908090, 0x709080A0, 0x70908070, 0x70908060,
	0x70927286, 0x70907090, 0x70907060, 0x70907050,

	0x70617286, 0x70607090, 0x70607060, 0x70607050,			/* A0 */
	0x70608090, 0x706080A0, 0x70608070, 0x70608060,
	0x70605060, 0x70605070, 0x70605040, 0x70605030,
	0x70604050, 0x70604060, 0x70604030, 0x70604020,

	0x70506070, 0x70506088, 0x70506050, 0x70506040,			/* B0 */
	0x70507286, 0x70507090, 0x70507060, 0x70507050,
	0x70504050, 0x70504060, 0x70504030, 0x70504020,
	0x70503040, 0x70503050, 0x70503020, 0x70503010,

	0x60708090, 0x607080A0, 0x60708070, 0x60708060,			/* C0 */
	0x607090A0, 0x607090B0, 0x5D6C8C78, 0x60709070,
	0x60706070, 0x60706088, 0x60706050, 0x60706040,
	0x60705060, 0x60705070, 0x60705040, 0x60705030,

	0x608090A0, 0x608090B0, 0x5D7C8C78, 0x60809070,			/* D0 */
	0x6080A0B0, 0x6080A0C0, 0x6080A090, 0x6080A078,
	0x60827286, 0x60807090, 0x60807060, 0x60807050,
	0x60806070, 0x60806088, 0x60806050, 0x60806040,

	0x60506070, 0x60506088, 0x60506050, 0x60506040,			/* E0 */
	0x60507286, 0x60507090, 0x60507060, 0x60507050,
	0x60504050, 0x60504060, 0x60504030, 0x60504020,
	0x60503040, 0x60503050, 0x60503020, 0x60503010,

	0x60405060, 0x60405070, 0x60405040, 0x60405030,			/* F0 */
	0x60406070, 0x60406088, 0x60406050, 0x60406040,
	0x60403040, 0x60403050, 0x60403020, 0x60403010,
	0x60402030, 0x60402040, 0x60402010, 0x60402000
};

#ifdef env68k
# pragma parameter translate_bytes_5r2(__A0,__A1,__D2)
	extern void translate_bytes_5r2(byte*,const UInt32*,SInt32) =
		{0x2F0A, 0x2002, 0xD080, 0xD080, 0xD082, 0x45F0, 0x800, 0x14BC, 128, 0xD1C2,
		0x7000, 0x1020, 0x2531, 0x400, 0x153C, 128, 0x5382, 0x6EF2, 0x245F};
	/*
			move.l		a2,-(sp)
			move.l		d2,d0
			add.l		d0,d0
			add.l		d0,d0
			add.l		d2,d0
			lea			(a0,d2.l),a2
			move.b		#128,(a2)
			adda.l		d2,a0
			moveq		#0,d0

		@1	move.b		-(a0),d0
			move.l		(a1,d0.w*4),-(a2)
			move.b		#128,-(a2)
			subq.l		#1,d2
			bgt.s		@1

			move.l		(sp)+,a2
	*/
#else
	_inline translate_bytes_5r2(byte *s, const UInt32 *t, SInt32 l)
	{
		UInt32		*d = (UInt32*)((l*5)+s);

		do
		{
			*--d = t[*--s];
			*--((byte*)d) = 128;
		} while(--l);
	}
#endif

OSErr translate_file_to_wave(ConstStr255Param name,
	short src, short dst,
	UInt16 rate,
	register PercentCompletionProc completion)
{
	ParamBlockRec	src_file, dst_file;

	UInt32			length;

	register byte		*buffer = malloc(MAXIMUM_BLOCK_SIZE*5+1);
	register UInt32		previous_ticks = TickCount()+60, position = 0, eof, size;
	register OSErr		err = 0;

	if (!buffer) return MemError();

	src_file.ioParam.ioRefNum = src, dst_file.ioParam.ioRefNum = dst;
	if (err=PBGetEOFSync(&src_file)) goto end;
	eof = ((length=(UInt32)src_file.ioParam.ioMisc)*5)+1280+sizeof(wave_header_template);
	dst_file.ioParam.ioMisc = (Ptr)eof;
	if (err=PBSetEOFSync(&dst_file)) goto end;

	wave_header_template.averageBytesPerSec = wave_header_template.sampleRate =
		Endian16_Swap(rate);
	wave_header_template.fileLength = Endian32_Swap(eof-7);
	wave_header_template.dataLength = Endian32_Swap(eof-(sizeof(wave_header_template)-1));

	dst_file.ioParam.ioBuffer = (Ptr)&wave_header_template;
	dst_file.ioParam.ioReqCount = sizeof(wave_header_template),
		dst_file.ioParam.ioPosMode = 0, dst_file.ioParam.ioPosOffset = 0;
	if (err=PBWriteSync(&dst_file)) goto end;

	memset(buffer,0,256);
	/* 32 bit addressing for now - first four bytes or the length are left zero */
	*(3+(UInt32*)buffer) = length;

	BlockMoveData(name,buffer+16,name[0]+1);
	if (buffer[16]>239) buffer[16] = 239, buffer[255] = '';
	translate_bytes_5r2(buffer,byte_samples,256);

	dst_file.ioParam.ioBuffer = (Ptr)buffer;
	dst_file.ioParam.ioReqCount = 1280;
	dst_file.ioParam.ioPosMode = 0, dst_file.ioParam.ioPosOffset = 0;
	if (err=PBWriteSync(&dst_file)) goto end;

	while(length>position)
	{
		if ((*(UInt32*)0x16A-previous_ticks) & 16)		/* Ticks -- ~1/4 second */
			completion(NFix30Div(position,length));

		size = length-position;
		src_file.ioParam.ioBuffer = (Ptr)buffer;
		src_file.ioParam.ioReqCount = MIN(size,MAXIMUM_BLOCK_SIZE),
			src_file.ioParam.ioPosMode = 0, src_file.ioParam.ioPosOffset = 0;
		if (err=PBReadSync(&src_file)) goto end;

		position += MAXIMUM_BLOCK_SIZE;

		if (!(dst_file.ioParam.ioReqCount=src_file.ioParam.ioReqCount*5))
			return 0;
		translate_bytes_5r2(buffer,byte_samples,src_file.ioParam.ioActCount);
		if (position>=length) ++dst_file.ioParam.ioReqCount;

		dst_file.ioParam.ioBuffer = (Ptr)buffer;
		dst_file.ioParam.ioPosMode = 0, dst_file.ioParam.ioPosOffset = 0;
		if (err=PBWriteSync(&dst_file)) goto end;
	}
end:
	return err;
}

static UInt8 name_remap[256] =
" \1\2\3\4\5\6\7\10\11\12\13\14\15\16\17\20\21\22\23\24\25\26\27\30\31\32\33\34\35\36\37\
 !\"#$%&'()*+,-./\
0123456789-;<=>?\
@ABCDEFGHIJKLMNO\
PQRSTUVWXYZ[\\]^_\
`abcdefghijklmno\
pqrstuvwxyz{|}~\
\
\
\
\
\
\
\
";

Boolean filter_name(Str255 name)
{
	register UInt8		*aptr = name, c, len;

	if (!(len=*aptr++)) return false;
	if (len>31)
	{
		aptr[-1] = 31, aptr[len-1] = '';
		len = 31;
	}
	do *aptr++ = name_remap[*aptr]; while(--len);
	return true;
}

/* interface code */

static void myCompletionRoutine(sfract percent)
{
	static short	run = 0;

	if (++run & 3) return;
	if (percent>0)
	{
		csprintf(temporary,"%d%% Completed", NFix30Mul(percent,100));
		SetOverheadBoxText(temporary);
	} else SetOverheadBoxText((run & 4) ? "--" : "||");
	HandleEvent(NULL,NULL,0,0);
	HandleEvent(NULL,NULL,0,0);
	if (run & 4) HandleEvent(NULL,NULL,0,0);
}

enum
{
	_user_canceled,
	_translate_from_audio,
	_translate_to_audio,
	_extract_wad_data,
	_extract_shapes_data
};

static const char	user_cancel_text[] = "- User Canceled -";
static const UInt8	nt[] = "\p<no title>";

static void perform_request(short command);

void ProgramMenuChoice(long menuHit)
{
extern Boolean	gDone;
	register short		menu = ((Point*)&menuHit)->v, item = (SInt16)menuHit;
	register MenuRef	mref;
	register WindowRef	win;
	register Boolean	junk;

	switch(menu)
	{
		case MBAppleID:
			if (item!=1) if (mref=GetMenuHandle(menu))
				{
					Str255		astr;

					GetMenuItemText(mref,item,astr);
					junk = OpenDeskAcc(astr);
				} else SysBeep(20); else Alert(500,NULL);
			break;
		case MBFileID:
			switch(item)
			{
				case 1:	perform_request(_translate_from_audio);	break;
				case 2:	perform_request(_translate_to_audio);	break;
				case 3:	perform_request(_extract_wad_data);		break;
				case 4:	perform_request(_extract_shapes_data);	break;
				case 6:	/* close */
					break;
				case 8: /* quit */
					gDone = TRUE;	/* soft exit */
					break;
			}
			break;
		case MBEditID:
	  		switch(item)
	  		{
				case 1: /* undo */
					break;
				case 3: /* cut */
					break;
				case 4: /* copy */
					break;
				case 5: /* paste */
					break;
				case 6: /* clear */
					break;
			}
	}
}

void UpdateMenus()
{
}

void startup(boolean display);
void startup(boolean display)
{
}

void ProgramMenuInit(void)
{
	MenuRef			mref;

	if ((mref=GetMenuHandle(MBFileID))==(MenuRef)NULL)
		return;

	MacInsertMenuItem(mref,"\pGAR->data/F",0);
	MacInsertMenuItem(mref,"\pdata->GAR/T",1);
	MacInsertMenuItem(mref,"\pExtract WAD Data/E",2);
	MacInsertMenuItem(mref,"\pExtract Shapes Data/H",3);
	MacInsertMenuItem(mref,"\p(-",4);
}

static UInt16	rate_list[] = {11025, 12000, 22050, 24000, 44100};

/* thats the bottom line - a line of bums */
static void perform_request(short command)
{
	StandardFileReply	src_file, dst_file;

	union spec_data		data_spec;

	Str255		name;
	OSType		file_type, creator_type;
	UInt16		rate;
	short		source = 0, destination = 0, temp;
	OSErr		err, derr;

	if (command==_user_canceled) return;

	if (command==_translate_to_audio)
	{
		file_type = 'WAVE', creator_type = 'SCPL';

		data_spec.adata.count = sizeof(rate_list)/2, data_spec.adata.form = "%hu";
		if (!ListDialog(NULL,"\pChoose the sample rate:",rate_list,
				dataArrayType,showSaveListButton,&data_spec,&temp,NULL))
		{
			rate = 0;
			return;
		}
		rate = rate_list[temp];
		StandardGetFile(NULL,-1,NULL,&src_file);
		SetOverheadBoxText((char*)empty);
		memcpy(dst_file.sfFile.name,src_file.sfFile.name,sizeof(StrFileName));
		if (dst_file.sfFile.name[0]>27)
		{
			dst_file.sfFile.name[0] = 27;
			dst_file.sfFile.name[27] = '';
		}
		psprintf(temporary,"%P.gar", dst_file.sfFile.name);
	} else
	{
		short		count;
		OSType		*types = NULL;

		SetOverheadBoxText((char*)empty);
		switch(command)
		{
		case _translate_from_audio:
			file_type = 'TEXT', creator_type = 'SITx';
			count = 2, types = (OSType*)"WAVE.WAV";
			break;
		case _extract_wad_data:
			file_type = 'BINA', creator_type = 'hDmp';
			getpstr(temporary,129,1);
			types = (OSType*)(1+temporary), count = ((UInt8)temporary[0])>>2;
			break;
		case _extract_shapes_data:
			file_type = 'col', creator_type = '52';
			count = 4, types = (OSType*)"shp2shp";
			break;
		}
		StandardGetFile(NULL,count,types,&src_file);
	}

	if (!src_file.sfGood) return;
	if (command>_translate_to_audio)
	{
		if (err=HOpenDF(src_file.sfFile.vRefNum,src_file.sfFile.parID,src_file.sfFile.name,fsRdPerm,&source))
			goto close;
		ListDialog(NULL,"\pSelect the portion to extract:",rate_list,
			dataArrayType,showSaveListButton,&data_spec,&temp,NULL);
	}
	StandardPutFile(creator_type=='SITx' ? "\pSave directory as:" : NULL,
		command==_translate_to_audio ? (UInt8*)temporary : empty,&dst_file);
	if (!dst_file.sfGood) return;

	initialize_debugger(FALSE);		/* start logging */

	memcpy(temporary+1,dst_file.sfFile.name+1,destination=dst_file.sfFile.name[0]);
/*
	temporary[0] = '', temporary[destination+1] = '';
	debugger_log(temporary,2+destination);
*/
	dprintf("%P", dst_file.sfFile.name);

	HDelete(dst_file.sfFile.vRefNum,dst_file.sfFile.parID,dst_file.sfFile.name);
	source = 0, destination = 0;
	if (creator_type=='SITx')
	{
		if (err=DirCreate(dst_file.sfFile.vRefNum,dst_file.sfFile.parID,dst_file.sfFile.name,&dst_file.sfFile.parID))
			goto close;
		if (err=HCreate(dst_file.sfFile.vRefNum,dst_file.sfFile.parID,nt,creator_type,file_type))
			goto close;
		if (err=HOpenDF(src_file.sfFile.vRefNum,src_file.sfFile.parID,src_file.sfFile.name,fsRdPerm,&source))
			goto close;
		if (err=HOpenDF(dst_file.sfFile.vRefNum,dst_file.sfFile.parID,nt,fsWrPerm,&destination))
			goto close;
	} else
	{
		if (err=HCreate(dst_file.sfFile.vRefNum,dst_file.sfFile.parID,dst_file.sfFile.name,creator_type,file_type))
			goto close;
		if (source>0)
			if (err=HOpenDF(src_file.sfFile.vRefNum,src_file.sfFile.parID,src_file.sfFile.name,fsRdPerm,&source))
				goto close;
		if (err=HOpenDF(dst_file.sfFile.vRefNum,dst_file.sfFile.parID,dst_file.sfFile.name,fsWrPerm,&destination))
			goto close;
	}

	switch(command)
	{
	case _translate_from_audio:
		name[0] = 0;
		err = translate_wave_to_file(name,source,destination,myCompletionRoutine);
		if (filter_name(name))
			if (!(derr=HRename(dst_file.sfFile.vRefNum,dst_file.sfFile.parID,nt,name)))
			{
				dprintf("Renamed to found p-string %P;g", name);
			}
		break;
	case _translate_to_audio:
		err = translate_file_to_wave(src_file.sfFile.name,
			source,destination,rate,myCompletionRoutine);
		break;
	case _extract_wad_data:
	case _extract_shapes_data:
		break;
	}
close:
	if (source>0)
		FSClose(source);
	if (destination>0)
		FSClose(destination);
	if (err)
	{
		if (err<0) csprintf(temporary,"Error: %d", err);
		else csprintf(temporary,"There may be corruption due to problems interpreting.\xd(%d) -- see the DebugLog for details", err);
		SetOverheadBoxText(temporary);
	} else SetOverheadBoxText("- Done -");
}

void global_idle_proc(void)
{
}
