
#include "macintosh_cseries.h"

#include "interface.h"
#include "shell.h"

#ifndef __FILETYPESANDCREATORS__
#	include <FileTypesAndCreators.h>
#endif
#ifndef __MACMEMORY__
#	include <MacMemory.h>
#endif
#ifndef __QUICKDRAW__
#	include <QuickDraw.h>
#endif
#ifndef __RESOURCES__
#	include <Resources.h>
#endif
#ifndef __STANDARDFILE__
#	include <StandardFile.h>
#endif

#ifndef __MACERRORS__
#	include <MacErrors.h>
#endif

#include "my32bqd.h"
#include "mydialogs.h"

static void filter_pstr(UInt8*);

static const char	rlist_form[] = "\xd%d\t%P";

static Handle GetResourceList(OSType resType, boolean currentOnly, short *_count)
{
	Str255		name;
	Handle		text_list = (Handle)NULL, h;
	OSType		type;
	long		length;
	short		x, count, id, err;

	if (currentOnly) count = Count1Resources(resType);
	else count = CountResources(resType);
	if (_count) *_count = count;
	if (count>0)
	{
		SetResLoad(0);
		if (currentOnly) h = Get1IndResource(resType,1);
		else h = GetIndResource(resType,1);
		SetResLoad(1);
		assert(h);
		GetResInfo(h,&id,&type,name);
		filter_pstr(name);
		csprintf(temporary,rlist_form+1, id,name);
		if (err=PtrToHand(temporary,&text_list,length=strlen(temporary)))
		{
			dprintf("PtrToHand(%p,%p,%d) gave %d", temporary,text_list,length,err);
			goto end;
		}
	}
	for(x=2; x<=count; ++x)
	{
		Handle		h;

		SetResLoad(0);
		if (currentOnly) h = Get1IndResource(resType,x);
		else h = GetIndResource(resType,x);
		SetResLoad(1);
		assert(h);
		GetResInfo(h,&id,&type,name);
		filter_pstr(name);
		csprintf(temporary,rlist_form, id,name);
		if (err=PtrAndHand(temporary,text_list,length=strlen(temporary)))
		{
			dprintf("PtrAndHand(%p,%p,%d) gave %d", temporary,text_list,length,err);
			goto end;
		}
	}
	PtrAndHand("\0",text_list,1);
end:
	return text_list;
}

/* filter_pstr, with the remapping table, deletes NULLs and CRs, and replaces
	tabs with the space character */

static UInt8 remapping_table[256] =
"\0\1\2\3\4\5\6\7\10 \12\13\14\0\16\17\20\21\22\23\24\25\26\27\30\31\32\33\34\35\36\37\
 !\"#$%&'()*+,-./\
0123456789:;<=>?\
@ABCDEFGHIJKLMNO\
PQRSTUVWXYZ[\\]^_\
`abcdefghijklmno\
pqrstuvwxyz{|}~\
\
\
\
\
\
\
\
";

static void filter_pstr(register UInt8 *str)
{
	register UInt8		length, x, *sstr = str, *dstr;

	if (x=*sstr++)
	{
		length = x, dstr = sstr;
		do
		{
			if (!(*dstr++=remapping_table[*sstr++]))
				--dstr, --length;
		} while(--x);
		str[0] = length;
	}
}

static OSErr NewOffscreenPort(GrafPtr port, Rect *bounds)
{
	GrafPtr		old_port;
	short		width, height, err = 0;

	if (!(width=bounds->right-bounds->left) || (unsigned)width>16383 ||
			!(height=bounds->bottom-bounds->top) || (unsigned)height>16383)
		return paramErr;

	GetPort(&old_port);
	OpenPort(port);
	if (err=QDError()) goto end;
	SetPort(port);	/* just to be sure */

	*(SInt32*)&port->portBits.bounds = 0;
	port->portBits.bounds.right = width, port->portBits.bounds.bottom = height;
	port->portBits.baseAddr = NewPtr((port->portBits.rowBytes=((width+15)>>4)<<1)*height);
	if (err=MemError())
	{
	close:
		ClosePort(port);
		goto end;
	}
	MovePortTo(bounds->left,bounds->top);
	PortSize(width,height);
	ClipRect(bounds);
	CopyRgn(port->clipRgn,port->visRgn);
end:
	SetPort(old_port);
	return err;
}

static void DisposeOffscreenPort(GrafPtr port)
{
	if (*(UInt16*)&port->portBits.baseAddr && *(UInt8*)&port->portBits.baseAddr<0x40)
	{
		DisposePtr(port->portBits.baseAddr);
		port->portBits.baseAddr = NULL;
	}
	ClosePort(port);
}

void startup(boolean display);
void startup(boolean display)
{
	StandardFileReply	selected_file;

	GrafPort	port;
	Rect		bounds = {0,0,240,320};
	UInt64		last_ticks, ticks;

	register SInt16		*each;

	PicHandle	oldcrap, newcrap = (PicHandle)NULL;
	Handle		rlist = (Handle)NULL, *saved = NULL, *saved_each;
	GrafPtr		oldport;

	UInt32		setup = 1;
	short		curRF = CurResFile(), srcRefNum = 0, dstRefNum = 0, attrs = 0;
	short		x, count, index, saved_count = 0, error;
	SInt16		*batch = NULL;

	Debugger();
	set_cursor(0,false);
	set_cursor(511,false);

	*(Float64*)&last_ticks = 0;
	GetPort(&oldport);
	StandardGetFile(NULL,-1,NULL,&selected_file);
	if (!selected_file.sfGood) goto end;

	SetResLoad(0);
	srcRefNum = FSpOpenResFile((FSSpec*)&selected_file.sfFile,fsRdPerm);
	SetResLoad(1);
	if (srcRefNum<=1)
	{
		dprintf("FSpOpenResFile(%p,fsRdPerm) gave %d, ResError: %d", &selected_file.sfFile,srcRefNum,ResError());
		goto close;
	}
	UseResFile(srcRefNum);
	if (!(rlist=GetResourceList('PICT',TRUE,&count)))
		goto close;
	UseResFile(curRF);
	if (count<=0)
		goto close; /* should not happen; rlist would be NULL in that case */

	batch = malloc(2*count);
	assert(batch);
	HLock(rlist);
	count = SelectTextLineDlg(NULL,psprintf(temporary,"(%04X) Select any picture(s):",
		GetResFileAttrs(srcRefNum)),*rlist,0,'\xd',0,batch,NULL)>>16;
	DisposeHandle(rlist);
	if (count<=0) goto close;
	saved = malloc(sizeof(Handle)*count);
	assert(saved);

	StandardPutFile(NULL,"\pPICT resources",&selected_file);
	if (!selected_file.sfGood) goto close;
	/*FSpDelete((FSSpec*)&selected_file.sfFile);*/
	FSpCreateResFile((FSSpec*)&selected_file.sfFile,sigResEdit,ftResEditResourceFile,0);
	if ((error=ResError()) && error!=dupFNErr)
	{
		dprintf("FSpCreateResFile(%p) gave %d (ResError)", &selected_file.sfFile,error);
		goto close;
	}
	if ((dstRefNum=FSpOpenResFile((FSSpec*)&selected_file.sfFile,fsWrPerm))<=1)
	{
		dprintf("FSpOpenResFile(%p,fsWrPerm) gave %d, ResError: %d", &selected_file.sfFile,dstRefNum,ResError());
		goto close;
	}
	each = batch, saved_each = saved;
	do
	{
		Str255		name;

		UseResFile(srcRefNum);
		if ((index=*each++)<0) goto close;
		if (!(oldcrap=(PicHandle)Get1IndResource('PICT',++index)))
		{
			dprintf("Get1IndResource('PICT',%d) gave %d (ResError)", index,ResError());
			goto close;
		}
		{
			OSType		type;

			GetResInfo((Handle)oldcrap,&index,&type,name);
		}
		attrs = GetResAttrs((Handle)oldcrap);
		DetachResource((Handle)oldcrap);

		bounds = (**oldcrap).picFrame;
		bounds.right = RECT_WIDTH(&bounds), bounds.bottom = RECT_HEIGHT(&bounds);
		*(SInt32*)&bounds = 0;

		if (error=NewOffscreenPort(&port,&bounds))
		{
			dprintf("NewOffscreenPort(%p,%d) gave %d", &bounds,error);
			goto skip;
		}
		SetPort(&port);
		EraseRect(&bounds);
		DrawPicture(oldcrap,&bounds);
	skip:
		DisposeHandle((Handle)oldcrap);
		oldcrap = NULL;

		if (!port.portBits.baseAddr) continue;
		if (!(newcrap=OpenPicture(&bounds)))
		{
			dprintf("OpenPicture(%p) gave NULL, QDError: %d", &bounds,QDError());
			goto close;
		}
		CopyBits(&port.portBits,&port.portBits,&bounds,&bounds,srcCopy,(RgnHandle)NULL);
		ClosePicture();

		UseResFile(dstRefNum);
		AddResource((Handle)newcrap,'PICT',index,name);
		if (error=ResError())
			dprintf("AddResource(%p,'PICT',%d,%p) gave %d (ResError)", newcrap,index,name,error);
		else
		{
			UpdateResFile(dstRefNum);
			if (error=ResError())
				dprintf("UpdateResFile(%d) gave %d (ResError)", dstRefNum,error);
			SetResAttrs((Handle)newcrap,attrs);
			DetachResource((Handle)newcrap);
			DisposeHandle((Handle)newcrap);
		}
		if (port.portBits.baseAddr) DisposeOffscreenPort(&port);
		SetPort(oldport);
		if (display)
		{
			++setup;
			sprintf(temporary,"Pictures remaining: %d", count);
			SetCrapBoxText(temporary);
		}
		ticks = machine_tick_count();
		if (U64Gt(&ticks,&last_ticks))
		{
			last_ticks = ticks;
			Add64U32(&last_ticks,30);				// half a second
			HandleEvent(NULL,NULL,0,0);
		// call a second time for background updates
			if (!(setup&3)) HandleEvent(NULL,NULL,0,0);
		}
	} while(--count);
close:
	if (batch) free(batch);
	if (saved) free(saved);
	UseResFile(curRF);
	if (srcRefNum>0) CloseResFile(srcRefNum);
	if (dstRefNum>0)
	{
		/*SetResFileAttrs(dstRefNum,GetResFileAttrs(dstRefNum)|mapCompact);*/
		UpdateResFile(dstRefNum);
		CloseResFile(dstRefNum);
	}
	SetCrapBoxText("-Done-");
end:
	SetPort(oldport);
}
